/*
 * linux/arch/arm/mach-pxa/leds-xyz3.c
 *
 * Author: Ryoichi KATO <Ryoichi.Kato@jp.sony.com>
 * Last Change: 2006/08/15 13:04:14.
 *
 * Based On: linux/arch/arm/mach-pxa/leds-mainstone.c
 *
 */
/*
 * linux/arch/arm/mach-pxa/leds-mainstone.c
 *
 * Author:     Nicolas Pitre
 * Created:    Nov 05, 2002
 * Copyright:  MontaVista Software Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/config.h>
#include <linux/init.h>
#include <linux/module.h>

#include <asm/hardware.h>
#include <asm/leds.h>
#include <asm/system.h>

#include <asm/arch/pxa-regs.h>
#include <asm/arch/xyz3.h>
#include <asm/arch/leds-xyz3.h>

#include "leds.h"


unsigned int __hw_led_state;

void xyz3_leds_event(led_event_t evt)
{
    unsigned long flags;

    local_irq_save(flags);

    switch (evt) {
    case led_start:
    case led_stop:
    case led_claim:
    case led_release:
        break;

#ifdef CONFIG_LEDS_TIMER
    case led_timer:
        xyz3_led_ctrl(xyz3_led_toggle, XYZ3_DEBUG_LED_1);
        break;
#endif

#ifdef CONFIG_LEDS_CPU
    case led_idle_start:
        xyz3_led_ctrl(xyz3_led_off, XYZ3_DEBUG_LED_2);
        break;

    case led_idle_end:
        xyz3_led_ctrl(xyz3_led_on, XYZ3_DEBUG_LED_2);
        break;
#endif

    case led_halted:
    case led_green_on:
    case led_green_off:
    case led_amber_on:
    case led_amber_off:
    case led_red_on:
    case led_red_off:
    case led_blue_on:
    case led_blue_off:
    default:
        break;
    }

    local_irq_restore(flags);
}


static int __init
xyz3_led_ctrl_init(void)
{
    unsigned long flags;
    local_irq_save(flags);
    XYZ3_SET_LED_VALUE( 0 );
    local_irq_restore(flags);
    return 0;
}

EXPORT_SYMBOL(__hw_led_state);
core_initcall(xyz3_led_ctrl_init);

