/*
 * navu-gpio.c - GPIO driver for "nav-u" NAVU/4 series
 *
 * Copyright (c) 2007 Sony Corporation.
 */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/major.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <linux/fs.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/cdev.h>
#include <linux/platform_device.h>
#include <linux/interrupt.h>
#include <linux/list.h>

#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <asm/semaphore.h>
#include <asm/mach/irq.h>
#include <asm/arch/system.h>
#include <asm/arch/pxa-regs.h>
#include <asm/arch/xyz3.h>
#include <asm/arch/xyz3-devnum.h>

#include <asm/arch/navu-gpio.h>



int navuGpioGetLevel(unsigned long port)
{
    unsigned long bit;
    int stat = 0;

    if( ( port & NAVU_GP_HIGH_WORD ) == NAVU_GP_HIGH_WORD )
    {
        bit = (port & NAVU_GPIO_MASK) << 16;
    }
    else
    {
        bit = port & NAVU_GPIO_MASK;
    }

    if( ( port & NAVU_GPREG_MASK ) == NAVU_GPxR0 )
    {
        if( ( GPLR0 & bit ) == bit )
        {
            stat = 1;
        }
    }
    else if( ( port & NAVU_GPREG_MASK ) == NAVU_GPxR1 )
    {
        if( ( GPLR1 & bit ) == bit )
        {
            stat = 1;
        }
    }
    else  if( ( port & NAVU_GPREG_MASK ) == NAVU_GPxR2 )
    {
        if( ( GPLR2 & bit ) == bit )
        {
            stat = 1;
        }
    }
    else  if( ( port & NAVU_GPREG_MASK ) == NAVU_GPxR3 )
    {
        if( ( GPLR3 & bit ) == bit )
        {
            stat = 1;
        }
    }
    else
    {
        stat = -EINVAL;
    }

    return stat;
}

int navuGpioSetLevel(unsigned long port, int level)
{
    unsigned long bit;
    int stat = 0;

    if( ( port & NAVU_GP_HIGH_WORD ) == NAVU_GP_HIGH_WORD )
    {
        bit = (port & NAVU_GPIO_MASK) << 16;
    }
    else
    {
        bit = port & NAVU_GPIO_MASK;
    }

    if( level == 0 )
    {
        if( ( port & NAVU_GPREG_MASK ) == NAVU_GPxR0 )
        {
            GPCR0 |=  bit;
        }
        else if( ( port & NAVU_GPREG_MASK ) == NAVU_GPxR1 )
        {
            GPCR1 |=  bit;
        }
        else  if( ( port & NAVU_GPREG_MASK ) == NAVU_GPxR2 )
        {
            GPCR2 |=  bit;
        }
        else  if( ( port & NAVU_GPREG_MASK ) == NAVU_GPxR3 )
        {
            GPCR3 |=  bit;
        }
        else
        {
            stat = -EINVAL;
        }
    }
    else
    {
        if( ( port & NAVU_GPREG_MASK ) == NAVU_GPxR0 )
        {
            GPSR0 |=  bit;
        }
        else if( ( port & NAVU_GPREG_MASK ) == NAVU_GPxR1 )
        {
            GPSR1 |=  bit;
        }
        else  if( ( port & NAVU_GPREG_MASK ) == NAVU_GPxR2 )
        {
            GPSR2 |=  bit;
        }
        else  if( ( port & NAVU_GPREG_MASK ) == NAVU_GPxR3 )
        {
            GPSR3 |=  bit;
        }
        else
        {
            stat = -EINVAL;
        }
    }

    return stat;
}


EXPORT_SYMBOL(navuGpioGetLevel);
EXPORT_SYMBOL(navuGpioSetLevel);
