/*
 * linux/arch/arm/mach-pxa/xyz3-ac97ad.c
 *    -- AC97(WM9705) A/D driver on Sony XYZ3
 *        
 * Author:     Hirokazu ABE and Ryoichi KATO
 * Created:    Apr 05, 2006
 * Copyright:  Sony Corp.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/major.h>
#include <linux/sched.h>
/* #include <linux/malloc.h> */
#include <linux/slab.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <linux/fs.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/cdev.h>

#include <asm/uaccess.h>

#include <linux/wm97xx.h>

#include <asm/arch/xyz3-devnum.h>

#define AC97AD_CH_MIN 4
#define AC97AD_CH_MAX 7

static struct wm97xx* ac97ad_wm;

/* +++++++++++++ File operations ++++++++++++++ */

static ssize_t ac97ad_read(struct file* filp, char __user * buff,
			    size_t count, loff_t*  f_pos)
{
    ssize_t readsize = 0;
    uint16_t readdata;
    int ch = (int)(filp->private_data);

    if(ac97ad_wm == NULL){ return -EPERM; }

    readdata = wm97xx_read_aux_adc(ac97ad_wm, (ch & 0x7) << 12);
    readsize = sizeof(readdata);

    if((count < readsize) ||
       (0 != copy_to_user(buff, &readdata, sizeof(readdata))) ){
	readsize = 0;
    }

    return readsize;
}

static int ac97ad_open(struct inode* inode, struct file* filp)
{
    if((imajor(inode) != XYZ3_MAJOR_AC97AD) ||
       ((iminor(inode) < XYZ3_MINOR_AC97AD_MIN) || (XYZ3_MINOR_AC97AD_MAX < iminor(inode))) ){
	return -ENODEV;
    }
  
    filp->private_data = (void*)(iminor(inode) - XYZ3_MINOR_AC97AD_MIN + AC97AD_CH_MIN);

    return 0;
}

static int ac97ad_release(struct inode* inode, struct file* filp)
{
    /* Nothing process */

    return 0;
}

static struct file_operations ac97ad_fops = {
    owner: THIS_MODULE,
    read: ac97ad_read,
    open: ac97ad_open,
    release: ac97ad_release,
};
static struct cdev ac97ad_cdev = {
    owner: THIS_MODULE,
    ops: &ac97ad_fops,
};

/* +++++++++++++ End File operations ++++++++++++++ */


static int ac97ad_probe(struct wm97xx* wm)
{
    ac97ad_wm = wm;
    return 0;
}

static void ac97ad_remove(struct wm97xx* wm)
{
    ac97ad_wm = NULL;
}

static struct wm97xx_misc_dev ac97ad_wm97xx_misc_dev = {
    .probe = ac97ad_probe,
    .remove = ac97ad_remove,
};

static int driver_registered = -1;
static int devno_registered = -1;
static int cdev_registered = -1;

#define XYZ3_MINOR_AC97AD_NOOF (XYZ3_MINOR_AC97AD_MAX - XYZ3_MINOR_AC97AD_MIN +1)

static void xyz3_ac97ad_exit( void )
{
    if(cdev_registered == 0){
	cdev_del(&ac97ad_cdev);
	cdev_registered = -1;
    }
    if(devno_registered == 0){
	unregister_chrdev_region(MKDEV(XYZ3_MAJOR_AC97AD, XYZ3_MINOR_AC97AD_MIN), XYZ3_MINOR_AC97AD_NOOF);
	devno_registered = -1;
    }
    if(driver_registered == 0){
	wm97xx_unregister_misc_dev(&ac97ad_wm97xx_misc_dev);
	driver_registered = -1;
    }
}

static int __init xyz3_ac97ad_init( void )
{
    INIT_LIST_HEAD(&(ac97ad_wm97xx_misc_dev.list));

    /* regist wm97xx driver */
    driver_registered = wm97xx_register_misc_dev(&ac97ad_wm97xx_misc_dev);
    if(driver_registered != 0){
	xyz3_ac97ad_exit();
	return -1;
    }

    /* regist DEVNO */
    devno_registered = register_chrdev_region(MKDEV(XYZ3_MAJOR_AC97AD, XYZ3_MINOR_AC97AD_MIN), XYZ3_MINOR_AC97AD_NOOF, "ac97ad");
    if(devno_registered != 0){
	xyz3_ac97ad_exit();
	return -1;
    }
  
    /* regist CDEV */
    cdev_init(&ac97ad_cdev, &ac97ad_fops);
    ac97ad_cdev.owner = THIS_MODULE;
    ac97ad_cdev.ops = &ac97ad_fops;
    cdev_registered = cdev_add(&ac97ad_cdev, MKDEV(XYZ3_MAJOR_AC97AD, XYZ3_MINOR_AC97AD_MIN), XYZ3_MINOR_AC97AD_NOOF);
    if(cdev_registered != 0){
	xyz3_ac97ad_exit();
	return -1;
    }

    return 0;
}

module_init(xyz3_ac97ad_init);
module_exit(xyz3_ac97ad_exit);

MODULE_AUTHOR("");
MODULE_DESCRIPTION("AC97(WM9705) A/D driver on Sony XYZ3");
MODULE_LICENSE("GPL");
