/*
 *  linux/arch/arm/mach-pxa/xyz3.c
 *
 *  Support for the Intel HCDDBBVA0 Development Platform.
 *  (go figure how they came up with such name...)
 *
 *  Author:	Nicolas Pitre
 *  Created:	Nov 05, 2002
 *  Copyright:	MontaVista Software Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 */

#include <linux/init.h>
#include <linux/platform_device.h>
#include <linux/sysdev.h>
#include <linux/interrupt.h>
#include <linux/sched.h>
#include <linux/bitops.h>
#include <linux/fb.h>

#include <asm/types.h>
#include <asm/setup.h>
#include <asm/memory.h>
#include <asm/mach-types.h>
#include <asm/hardware.h>
#include <asm/irq.h>

#include <asm/mach/arch.h>
#include <asm/mach/map.h>
#include <asm/mach/irq.h>

#include <asm/arch/pxa-regs.h>
#include <asm/arch/xyz3.h>
#include <asm/arch/audio.h>
#include <asm/arch/pxafb.h>
#include <asm/arch/mmc.h>

#include <linux/proc_fs.h>	//m33

#include "generic.h"

//m33
#define PROC_CPUFREQ
#ifdef PROC_CPUFREQ
#define BASE_CLK	13000000
static struct proc_dir_entry* clock_proc;

static int clock_proc_read(char *page, char **start, off_t off,
                           int count, int *eof, void *data)
{
	unsigned long	ccsr;
	unsigned long	clkcfg;
	unsigned int 	l, L, m, M, n2, N, S;
	int          	cccr_a, t, ht, b;
	int          	len;

	ccsr = CCSR;
	cccr_a = CCCR & (1 << 25);

	/* Read clkcfg register: it has turbo, b, half-turbo (and f) */
	asm( "mrc\tp14, 0, %0, c6, c0, 0" : "=r" (clkcfg) );
	t  = clkcfg & (1 << 1);
	ht = clkcfg & (1 << 2);
	b  = clkcfg & (1 << 3);

	l  = ccsr & 0x1f;
	n2 = (ccsr>>7) & 0xf;
	m  = (l <= 10) ? 1 : (l <= 20) ? 2 : 4;

	L  = l * BASE_CLK;
	N  = (L * n2) / 2;
	M  = (!cccr_a) ? (L/m) : ((b) ? L : (L/2));
	S  = (b) ? L : (L/2);

	len  = sprintf( page, "Run Mode clock: %d.%02dMHz (*%d)\n",
			L / 1000000, (L % 1000000) / 10000, l );
	len += sprintf( page+len, "Turbo Mode clock: %d.%02dMHz (*%d.%d, %sactive)\n",
			N / 1000000, (N % 1000000)/10000, n2 / 2, (n2 % 2)*5,
			(t) ? "" : "in" );
	len += sprintf( page+len, "Memory clock: %d.%02dMHz (/%d)\n",
			M / 1000000, (M % 1000000) / 10000, m );
	len += sprintf( page+len, "System bus clock: %d.%02dMHz \n",
			S / 1000000, (S % 1000000) / 10000 );

	return len;
}

static int clock_proc_write(struct file *file, const char __user *buffer,
                            unsigned long count, void *data )
{
	char buf[128];
	unsigned long mhz;
	unsigned long cccr;
	unsigned long clkcfg;
	int n2;

	memset( buf, 0, 128 );
	if( count < 128 )
	{
		copy_from_user( buf, buffer, count );
	}
	else
	{
		printk(KERN_ERR "%s:string length so long.\n", __FUNCTION__ );
		return -EFAULT;
	}

	mhz = simple_strtoul( buf, NULL, 0 );

	printk(KERN_INFO "request CPU Frequency:%ldMHz\n", mhz );

	switch( mhz )
	{
		case 208:
			n2 = 2;
			break;
		case 312:
			n2 = 3;
			break;
		case 416:
			n2 = 4;
			break;
		case 520:
			n2 = 5;
			break;
		case 624:
			n2 = 6;
			break;
		default:
			printk(KERN_ERR "%s:Invalid Frequency.\n", __FUNCTION__ );
			return -EFAULT;
	}

	cccr = CCCR;
	printk(KERN_INFO "%s:org CCCR=0x%08lx\n", __FUNCTION__, cccr );

	asm( "mrc\tp14, 0, %0, c6, c0, 0" : "=r" (clkcfg) );
	printk(KERN_INFO "%s:org CLKCFG=0x%08lx\n", __FUNCTION__, clkcfg );

	cccr = (cccr & 0x1f) + (n2 << 7);
	CCCR = cccr;
	printk(KERN_INFO "%s:set CCCR=0x%08lx\n", __FUNCTION__, cccr );

	clkcfg |= 0x02;
	asm( "mcr\tp14, 0, %0, c6, c0, 0" : : "r" (clkcfg) );

	return count;
}
#endif   //PROC_CPUFREQ

static void __init xyz3_init_irq(void)
{
	pxa_init_irq();
}

static struct resource smc91x_resources[] = {
	[0] = {
		.start	= (XYZ3_ETH_PHYS + 0x300),
		.end	= (XYZ3_ETH_PHYS + 0xfffff),
		.flags	= IORESOURCE_MEM,
	},
	[1] = {
		.start  = IRQ_GPIO(10),
		.end    = IRQ_GPIO(10),
		.flags	= IORESOURCE_IRQ,
	}
};

static struct platform_device smc91x_device = {
	.name		= "smc91x",
	.id		= 0,
	.num_resources	= ARRAY_SIZE(smc91x_resources),
	.resource	= smc91x_resources,
};


static struct platform_device xyz3_audio_device = {
	.name		= "pxa2xx-ac97",
	.id		= -1,
//	.dev		= { .platform_data = &xyz3_audio_ops },
};

static struct platform_device xyz3ostimer_device = {
	.name		= "xyz3-ostimer",
	.id		= -1,
};

static struct platform_device xyz3adc_device = {
	.name		= "xyz3-adc",
	.id		= -1,
	.dev	= { .parent = &xyz3ostimer_device.dev },
};

static struct platform_device xyz3pwm_device = {
	.name		= "xyz3-pwm",
	.id		= -1,
};

static struct platform_device xyz3gpio_device = {
	.name		= "xyz3-gpio",
	.id		= -1,
};

static struct platform_device xyz3doc_device = {
	.name		= "xyz3-doc",
	.id		= -1,
};

static void __init xyz3_init(void)
{

	platform_device_register(&smc91x_device);
	platform_device_register(&xyz3_audio_device);
	platform_device_register(&xyz3ostimer_device);
	platform_device_register(&xyz3adc_device);
	platform_device_register(&xyz3pwm_device);
	platform_device_register(&xyz3gpio_device);
	platform_device_register(&xyz3doc_device);

#ifdef PROC_CPUFREQ
	clock_proc = create_proc_entry("cpufreq", 0644, NULL);
        if(clock_proc != NULL) {
            clock_proc->read_proc  = clock_proc_read;
            clock_proc->write_proc = clock_proc_write;
        }
#endif   //PROC_CPUFREQ
}


static struct map_desc xyz3_io_desc[] __initdata = {
  {   /* CPLD */
        .virtual    =  XYZ3_FPGA_VIRT,
        .pfn        = __phys_to_pfn(XYZ3_FPGA_PHYS),
        .length     = 0x00100000,
        .type       = MT_DEVICE
  },
};

static void __init xyz3_map_io(void)
{
	pxa_map_io();
	iotable_init(xyz3_io_desc, ARRAY_SIZE(xyz3_io_desc));

	/* initialize sleep mode regs (wake-up sources, etc) */
	PGSR0 = 0x00008018; /* nCS1, PI2C_SDA, PI2C_SCL */
	PGSR1 = 0x00000002; /* nCS5 */
	PGSR2 = 0x0001C000; /* nCS4, nCS3, nCS2 */
	PGSR3 = 0x00000000;
	PWER  = 0x00000003; // GPIO 0,1 wakeup; POWER BTN and ACC_ON
	PRER  = 0x00000000; // ditto (disable raising edge detect)
	PFER  = 0x00000003; // ditto (enable falling edge detect)
	PSLR |= 0x00400000; // Ignore VDD Fault in deep sleep mode.
}

MACHINE_START(MAINSTONE, "Sony PXA270 Development Platform (aka XYZ3)")
	.phys_ram		= 0xa0000000,
	.phys_io		= 0x40000000,
	.io_pg_offst	= (io_p2v(0x40000000) >> 18) & 0xfffc,
	.boot_params	= 0xa0000100,
	.map_io			= xyz3_map_io,
	.init_irq		= xyz3_init_irq,
	.timer			= &pxa_timer,
	.init_machine	= xyz3_init,
MACHINE_END
