
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/major.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <linux/fs.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/cdev.h>
#include <linux/platform_device.h>

#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>
#include <asm/uaccess.h>
#include <asm/semaphore.h>


#include <asm/arch/system.h>
#include <asm/arch/pxa-regs.h>
#include <asm/arch/xyz3.h>
#include <asm/arch/xyz3-devnum.h>

// xyz3
#include <asm/arch/xyz3ad7795.h>
#include <asm/arch/xyz3ssp.h>
#include <asm/arch/xyz3ringbuf.h>
#include <asm/arch/xyz3-ostimer.h>

// Timer  Nom
#define AD_CYCTIMER		5
#define AD_TICK			6

// Power Mode
#define ADC_POWER_DOWN 	MODEMD_POWERDOWN
#define ADC_IDLE		MODEMD_IDLE

// SSP port for AD7795
#define SSP_PORT_ADC	1

//#define DEBUG
#ifdef DEBUG
#define DPRN(x)		x
#else
#define DPRN(x)
#endif


#if 0
static int DevMajorNumber = 0;			// major number
#else
static int DevMajorNumber = XYZ3_MAJOR_ADC;		// major number
#endif
static int DevMinorNumber = XYZ3_MINOR_ADC;		// minor number
static int DevCount       = 1;			// 
static char *DevName      = "adc";		// device driver name


typedef struct {
	int used;
	Xyz3SspDev		ssp;
	struct semaphore sem;
	struct cdev cdev;
} AdcDevSt;

AdcDevSt Adc_devst;

static int RdCh;

Xyz3RingBuf *DataBuf;
PosSensor	AdData;

static Ad7795RegArg AdSaveReg;

static Xyz3SspDev SspSaveReg;

// Function pointer of callback.
void (*fall)(u16 x, u16 y, u16 z);


// Timer call-back routine
void adcTimerService(void)
{
	u32 data;

	// Read Data
	switch( RdCh )
	{
		case 0:
			AdData.tick = xyz3_ostimer_get_count( AD_TICK );
			data = COM_REG( COMREG_RD | COMRS_DATA );
			AdData.g_x  = xyz3Ad7795FrameTrans( data, 8, 16);

			// Next:Y ch
			data = ( COM_REG( COMREG_WR | COMRS_CONFIG ) << 16 )
			       + CONF_REG( CONFREG_UNI | CONG_REF_REF1 | CONFREG_CH(ADC_CH_G_Y) );
			xyz3Ad7795FrameTrans( data, 24, 0 );
			break;
		case 1:
			data = COM_REG( COMREG_RD | COMRS_DATA );
			AdData.g_y  = xyz3Ad7795FrameTrans( data, 8, 16);

			// Next:Z ch
			data = ( COM_REG( COMREG_WR | COMRS_CONFIG ) << 16 )
			       + CONF_REG( CONFREG_UNI | CONG_REF_REF1 | CONFREG_CH(ADC_CH_G_Z) );
			xyz3Ad7795FrameTrans( data, 24, 0 );
			break;
		case 2:
			data = COM_REG( COMREG_RD | COMRS_DATA );
			AdData.g_z  = xyz3Ad7795FrameTrans( data, 8, 16);

			// Next:Temp ch
			data = ( COM_REG( COMREG_WR | COMRS_CONFIG ) << 16 )
			       + CONF_REG( CONFREG_UNI | CONG_REF_REF2 | CONFREG_CH(ADC_CH_TEMP) );
			xyz3Ad7795FrameTrans( data, 24, 0 );
			break;
		case 3:
			data = COM_REG( COMREG_RD | COMRS_DATA );
			AdData.temp  = xyz3Ad7795FrameTrans( data, 8, 16);
			// Next:Atmo ch
			data = ( COM_REG( COMREG_WR | COMRS_CONFIG ) << 16 )
			       + CONF_REG( CONFREG_UNI | CONG_REF_REF2 | CONFREG_CH(ADC_CH_ATMO) );
			xyz3Ad7795FrameTrans( data, 24, 0 );
			break;
		case 4:
			data = COM_REG( COMREG_RD | COMRS_DATA );
			AdData.atmo  = xyz3Ad7795FrameTrans( data, 8, 16);

			// Next:X ch
			data = ( COM_REG( COMREG_WR | COMRS_CONFIG ) << 16 )
			       + CONF_REG( CONFREG_UNI | CONG_REF_REF1 | CONFREG_CH(ADC_CH_G_X) );
			xyz3Ad7795FrameTrans( data, 24, 0 );
			break;
		default:
			// Init state.(RdCh is -1)
			// Next:X ch
			data = ( COM_REG( COMREG_WR | COMRS_CONFIG ) << 16 )
			       + CONF_REG( CONFREG_UNI | CONG_REF_REF1 | CONFREG_CH(ADC_CH_G_X) );
			xyz3Ad7795FrameTrans( data, 24, 0 );
			break;
	}

	// Next channel
	RdCh++;

	// save
	if( RdCh > 4 )
	{
		xyz3RingBufWrite( DataBuf, &AdData );
		RdCh = 0;

		if( fall != NULL )
		{
			fall( AdData.g_x, AdData.g_y, AdData.g_z);
		}

	}
}

// callback routine registry
int AdCallbackRegistry( void (*func)(u16, u16, u16) )
{
	if( fall == NULL )
	{
		fall = func;
		return 0;
	}
	printk( KERN_ERR "%s:It is already registered!!\n", __FUNCTION__ );
	return -1;
}
// callback unregistry
void AdCallbackUnregistry( void )
{
	fall = NULL;
}


// Timer Initialize routine.
static int adcTimerInit( void )
{
	DPRN( printk( KERN_DEBUG "%s() called.\n", __FUNCTION__ ) );

	// Read channel Init state.
	RdCh = -1;

	// Timer registory
	xyz3_ostimer_setting( AD_TICK, 0, 0, 0, OSTIMER_RESOLUTION_1ms );
	if( xyz3_ostimer_register_callback( AD_CYCTIMER, AD_SAMPL_TIME, OSTIMER_RESOLUTION_1ms,  adcTimerService ) == 0 )
	{
		printk( KERN_ERR "%s:xyz3_ostimer_register_callback(AD_CYCTIMER) failed!!\n", __FUNCTION__ );
		return -EBUSY;
	}
	return 0;
}

// Timer cleanup.
static void adcTimerClean( void )
{
	DPRN( printk( KERN_DEBUG "%s() called.\n", __FUNCTION__ ) );

	if( xyz3_ostimer_remove_callback( AD_CYCTIMER, adcTimerService ) == 0 )
	{
		printk( KERN_ERR "%s:xyz3_ostimer_remove_callback(AD_CYCTIMER) failed!!\n", __FUNCTION__ );
	}
	xyz3_ostimer_setting( AD_CYCTIMER, 0, 0, 0, OSTIMER_RESOLUTION_STOP );
	xyz3_ostimer_setting( AD_TICK, 0, 0, 0, OSTIMER_RESOLUTION_STOP );
}


// AD7795 Initialize.
static int adcInit( void )
{
	u32		data;
	Xyz3SspDev *ssp = &Adc_devst.ssp;

	DPRN( printk( KERN_DEBUG "%s() called.\n", __FUNCTION__ ) );

	ssp->sscr0 = SSCR0_TIM | SSCR0_RIM | SSCR0_SSE |SSCR0_SerClkDiv(26) | SSCR0_EDSS | SSCR0_DataSize(16);	// 32bit width
	ssp->sscr1 = SSCR1_SPH | SSCR1_SPO;
	ssp->sssr  = 0;
	ssp->ssto  = 0;
	ssp->sspsp = 0;

	// SSP port initialize
	xyz3Ad7795SspInit( ssp );
	udelay(500);

	//reset
	xyz3Ad7795FrameTrans( 0xffffffff, 32, 0 );
	udelay(500);

	// Init configration register
	data = ( COM_REG( COMREG_WR | COMRS_CONFIG ) << 16 )
	       + CONF_REG( CONFREG_UNI );
	xyz3Ad7795FrameTrans( data, 24, 0 );
	udelay(500);

	// Zero calibration
	data = ( COM_REG( COMREG_WR | COMRS_MODE ) << 16 )
	       + MODE_REG( MODEMD_IN_ZERO_CAL | MODEREG_AMP_CM | MODEREG_CHOP_DIS | MODEREG_FS(1) );
	xyz3Ad7795FrameTrans( data, 24, 0 );
	udelay(500);

	// Full calibration
	data = ( COM_REG( COMREG_WR | COMRS_MODE ) << 16 )
	       + MODE_REG( MODEMD_IN_FULL_CAL | MODEREG_AMP_CM | MODEREG_CHOP_DIS | MODEREG_FS(1) );
	xyz3Ad7795FrameTrans( data, 24, 0 );
	udelay(500);

	// Continuous Mode
	data = ( COM_REG( COMREG_WR | COMRS_MODE ) << 16 )
	       + MODE_REG( MODEMD_CONTINUE | MODEREG_AMP_CM | MODEREG_CHOP_DIS | MODEREG_FS(1) );
	xyz3Ad7795FrameTrans( data, 24, 0 );
	udelay(500);

	// Timer Initilize
	return adcTimerInit();
}


// Timer stop.
static int adcTimerStop( void )
{
	DPRN( printk( KERN_DEBUG "%s() called.\n", __FUNCTION__ ) );

	xyz3_ostimer_remove_callback( AD_CYCTIMER, adcTimerService );
	return 0;
}

// Timer restart.
static int adcTimerRestart( void )
{
	DPRN( printk( KERN_DEBUG "%s() called.\n", __FUNCTION__ ) );

	if( xyz3_ostimer_register_callback( AD_CYCTIMER, AD_SAMPL_TIME, OSTIMER_RESOLUTION_1ms,  adcTimerService ) == 0 )
	{
		// NG
		return -EBUSY;
	}
	return 0;
}


// ADC Power mode setting
static int adcSetPowerMode( int mode )
{
	u32 data;
	DPRN( printk( KERN_DEBUG "%s() called.\n", __FUNCTION__ ) );

	switch( mode )
	{
		case ADC_POWER_DOWN:
		case ADC_IDLE:
			break;
		default:
			printk( KERN_ERR "%s():Invalid argument!(%d)\n", __FUNCTION__, mode );
			return -EINVAL;
	}

	data = ( COM_REG( COMREG_WR | COMRS_MODE ) << 16 )
	       + MODE_REG( mode | MODEREG_AMP_CM | MODEREG_CHOP_DIS | MODEREG_FS(1) );
	xyz3Ad7795FrameTrans( data, 24, 0 );
	udelay(500);
	return 0;
}

static int ad7795RestoreState( Ad7795RegArg *parg )
{
	u32 data;

	DPRN( printk( KERN_DEBUG "%s() called.\n", __FUNCTION__ ) );

	//reset
	xyz3Ad7795FrameTrans( 0xffffffff, 32, 0 );
	udelay(500);

	// Idle mode
	adcSetPowerMode( ADC_IDLE );

	// 1 ch: Offset / Full Scale register restore
	data = ( COM_REG( COMREG_WR | COMRS_CONFIG ) << 16 )
	       + CONF_REG( CONFREG_UNI | CONG_REF_REF1 | CONFREG_CH(ADC_CH_G_X) );
	xyz3Ad7795FrameTrans( data, 24, 0 );

	data = ( COM_REG( COMREG_RD | COMRS_OFFSET ) << 16 )
	       + (parg->offset[0] & 0xffff);
	xyz3Ad7795FrameTrans( data, 24, 0);

	data = ( COM_REG( COMREG_RD | COMRS_OFFSET ) << 16 )
	       + (parg->fullscale[0] & 0xffff);
	xyz3Ad7795FrameTrans( data, 24, 0);

	// 2 ch: Offset / Full Scale register restore
	data = ( COM_REG( COMREG_WR | COMRS_CONFIG ) << 16 )
	       + CONF_REG( CONFREG_UNI | CONG_REF_REF1 | CONFREG_CH(ADC_CH_G_Y) );
	xyz3Ad7795FrameTrans( data, 24, 0 );

	data = ( COM_REG( COMREG_RD | COMRS_OFFSET ) << 16 )
	       + (parg->offset[1] & 0xffff);
	xyz3Ad7795FrameTrans( data, 24, 0);

	data = ( COM_REG( COMREG_RD | COMRS_OFFSET ) << 16 )
	       + (parg->fullscale[1] & 0xffff);
	xyz3Ad7795FrameTrans( data, 24, 0);

	// 3 ch: Offset / Full Scale register restore
	data = ( COM_REG( COMREG_WR | COMRS_CONFIG ) << 16 )
	       + CONF_REG( CONFREG_UNI | CONG_REF_REF1 | CONFREG_CH(ADC_CH_G_Z) );
	xyz3Ad7795FrameTrans( data, 24, 0 );

	data = ( COM_REG( COMREG_RD | COMRS_OFFSET ) << 16 )
	       + (parg->offset[2] & 0xffff);
	xyz3Ad7795FrameTrans( data, 24, 0);

	data = ( COM_REG( COMREG_RD | COMRS_OFFSET ) << 16 )
	       + (parg->fullscale[2] & 0xffff);
	xyz3Ad7795FrameTrans( data, 24, 0);

	// 4-6 ch: Offset / Full Scale register restore
	data = ( COM_REG( COMREG_WR | COMRS_CONFIG ) << 16 )
	       + CONF_REG( CONFREG_UNI | CONG_REF_REF1 | CONFREG_CH(ADC_CH_TEMP) );
	xyz3Ad7795FrameTrans( data, 24, 0 );

	data = ( COM_REG( COMREG_RD | COMRS_OFFSET ) << 16 )
	       + (parg->offset[3] & 0xffff);
	xyz3Ad7795FrameTrans( data, 24, 0);

	data = ( COM_REG( COMREG_RD | COMRS_OFFSET ) << 16 )
	       + (parg->fullscale[3] & 0xffff);
	xyz3Ad7795FrameTrans( data, 24, 0);

	// Mode register restore
	data = ( COM_REG( COMREG_WR | COMRS_MODE ) << 16 ) + (parg->mode & 0xffff);
	xyz3Ad7795FrameTrans( data, 24, 0 );
	udelay(500);

	// Configuratiorn register restore
	data = ( COM_REG( COMREG_WR | COMRS_CONFIG ) << 16 ) + (parg->config & 0xffff);
	xyz3Ad7795FrameTrans( data, 24, 0 );
	udelay(500);

	return 0;
}

// ADC internal register read.
static int adcReadReg( Ad7795RegArg *parg )
{
	u32 data;

	DPRN( printk( KERN_DEBUG "%s() called.\n", __FUNCTION__ ) );

	// Status Reg
	data = COM_REG( COMREG_RD | COMRS_STATUS );
	parg->status = xyz3Ad7795FrameTrans( data, 8, 8);

	// Mode Reg
	data = COM_REG( COMREG_RD | COMRS_MODE );
	parg->mode = xyz3Ad7795FrameTrans( data, 8, 16);

	// Configration Reg
	data = COM_REG( COMREG_RD | COMRS_CONFIG );
	parg->config = xyz3Ad7795FrameTrans( data, 8, 16);

	// ID Reg
	data = COM_REG( COMREG_RD | COMRS_ID );
	parg->id = xyz3Ad7795FrameTrans( data, 8, 8);

	// IO Reg
	data = COM_REG( COMREG_RD | COMRS_IO );
	parg->io = xyz3Ad7795FrameTrans( data, 8, 8);

	// 1 ch: Offset / Full Scale
	data = ( COM_REG( COMREG_WR | COMRS_CONFIG ) << 16 )
	       + CONF_REG( CONFREG_UNI | CONG_REF_REF1 | CONFREG_CH(ADC_CH_G_X) );
	xyz3Ad7795FrameTrans( data, 24, 0 );

	data = COM_REG( COMREG_RD | COMRS_OFFSET );
	parg->offset[0] = xyz3Ad7795FrameTrans( data, 8, 16);

	data = COM_REG( COMREG_RD | COMRS_F_SCALE );
	parg->fullscale[0] = xyz3Ad7795FrameTrans( data, 8, 16);

	// 2 ch: Offset / Full Scale
	data = ( COM_REG( COMREG_WR | COMRS_CONFIG ) << 16 )
	       + CONF_REG( CONFREG_UNI | CONG_REF_REF1 | CONFREG_CH(ADC_CH_G_Y) );
	xyz3Ad7795FrameTrans( data, 24, 0 );

	data = COM_REG( COMREG_RD | COMRS_OFFSET );
	parg->offset[1] = xyz3Ad7795FrameTrans( data, 8, 16);

	data = COM_REG( COMREG_RD | COMRS_F_SCALE );
	parg->fullscale[1] = xyz3Ad7795FrameTrans( data, 8, 16);

	// 3 ch: Offset / Full Scale
	data = ( COM_REG( COMREG_WR | COMRS_CONFIG ) << 16 )
	       + CONF_REG( CONFREG_UNI | CONG_REF_REF1 | CONFREG_CH(ADC_CH_G_Z) );
	xyz3Ad7795FrameTrans( data, 24, 0 );

	data = COM_REG( COMREG_RD | COMRS_OFFSET );
	parg->offset[2] = xyz3Ad7795FrameTrans( data, 8, 16);

	data = COM_REG( COMREG_RD | COMRS_F_SCALE );
	parg->fullscale[2] = xyz3Ad7795FrameTrans( data, 8, 16);

	// 4-6 ch: Offset / Full Scale
	data = ( COM_REG( COMREG_WR | COMRS_CONFIG ) << 16 )
	       + CONF_REG( CONFREG_UNI | CONG_REF_REF2 | CONFREG_CH(ADC_CH_TEMP) );
	xyz3Ad7795FrameTrans( data, 24, 0 );

	data = COM_REG( COMREG_RD | COMRS_OFFSET );
	parg->offset[3] = xyz3Ad7795FrameTrans( data, 8, 16);

	data = COM_REG( COMREG_RD | COMRS_F_SCALE );
	parg->fullscale[3] = xyz3Ad7795FrameTrans( data, 8, 16);

	// Config Reg is returned. 
	data = ( COM_REG( COMREG_WR | COMRS_CONFIG ) << 16 )
	       + CONF_REG( parg->config );
	xyz3Ad7795FrameTrans( data, 24, 0 );

	return 0;
}

// Get AD7795 configuration
static int adcGetConfig( AdcDevSt *pdev, Ad7795RegArg *parg )
{
	return adcReadReg( parg );
}

// Get Sensor data.
static int adcGetData( AdcDevSt *pdev, AdReadArg *parg )
{
	int i;
	int readcnt;
	int maxcnt;
	PosSensor *rdbuf;
	unsigned long ret;

	DPRN( printk( KERN_DEBUG "%s() called.\n", __FUNCTION__ ) );

	maxcnt = parg->maxReadCount;
	if( maxcnt <= 0 )
	{
		return 0;
	}

	if( maxcnt > POS_BUF_NUM )
	{
		maxcnt = POS_BUF_NUM;
	}

	rdbuf = kmalloc( sizeof(PosSensor) * maxcnt, GFP_KERNEL );
	if( rdbuf == NULL )
	{
		return -ENOMEM;
	}
	memset( rdbuf, 0, sizeof(PosSensor) * maxcnt );

	readcnt = 0;
	for( i=0; i<maxcnt; i++ )
	{
		if( xyz3RingBufRead( DataBuf, &rdbuf[i] ) == 0 )
		{
			break;
		}
		readcnt++;
	}

	ret = copy_to_user( parg->buf, rdbuf, sizeof(PosSensor) * readcnt );
	kfree( rdbuf );
	if( ret != 0 )
	{
		return ret;
	}

	return readcnt;
}



static int adc_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
	int stat = 0;
	AdcDevSt	*pdev = (AdcDevSt *)filp->private_data;

	DPRN( printk( KERN_DEBUG "%s() called.\n", __FUNCTION__ ) );

	switch( cmd )
	{
		case XYZ3ADC_GET_CONFIG:
			stat = adcGetConfig( pdev, (Ad7795RegArg*)arg );
			break;

		case XYZ3ADC_READ:
			stat = adcGetData( pdev, (AdReadArg *)arg );
			break;

		case XYZ3ADC_STOP:
			stat = adcTimerStop();
			break;

		case XYZ3ADC_RESTART:
			stat = adcTimerRestart();
			break;

		default:
			stat = -EINVAL;
			break;
	}
	return stat;
}

static int adc_open(struct inode *inode, struct file *filp)
{
	AdcDevSt	*pdev;

	DPRN( printk( KERN_DEBUG "%s() called.\n", __FUNCTION__ ) );

	pdev = container_of(inode->i_cdev, AdcDevSt, cdev);
	filp->private_data = pdev;

	if( pdev->used )
	{
		printk( KERN_ERR "%s:AD7795 using.\n", __FUNCTION__ );
		return -EBUSY;
	}

	if( down_interruptible( &pdev->sem ) )
	{
		printk( KERN_ERR "%s:down_interruptible failed!!\n", __FUNCTION__ );
		return -ERESTARTSYS;
	}

	pdev->used = 1;
	up( &pdev->sem );

//	return adc_init( &pdev->channel, port );
	return 0;
}


static int adc_release( struct inode *inode, struct file *filp )
{
	AdcDevSt	*pdev;

	DPRN( printk( KERN_DEBUG "%s() called.\n", __FUNCTION__ ) );

	pdev = container_of(inode->i_cdev, AdcDevSt, cdev);

	if( down_interruptible( &pdev->sem ) )
	{
		printk( KERN_ERR "%s:down_interruptible failed!!\n", __FUNCTION__ );
		return -ERESTARTSYS;
	}

	pdev->used = 0;
	up( &pdev->sem );

	return 0;
}


static struct file_operations adc_fops = {
	owner:	THIS_MODULE,
	ioctl:	adc_ioctl,
	open:	adc_open,
	release: adc_release,
};


/*
 *  Power manager Support Routines
 */


int __init xyz3_adc_probe(struct platform_device *dev)
{
	DPRN( printk( KERN_DEBUG "%s() called.\n", __FUNCTION__ ) );
	return 0;
}

static int xyz3_adc_remove(struct platform_device *dev)
{
	DPRN( printk( KERN_DEBUG "%s() called.\n", __FUNCTION__ ) );
	return 0;
}

static int xyz3_adc_suspend(struct platform_device *dev, pm_message_t state)
{
	printk( KERN_DEBUG "%s() called.\n", __FUNCTION__ );

	// TimeŕAOS-timerZ[u

	adcTimerStop(); 

	adcReadReg( &AdSaveReg );

	xyz3SspSaveState( &SspSaveReg, SSP_PORT_ADC );

	return 0;
}

static int xyz3_adc_resume(struct platform_device *dev)
{
	int rc = 0;

	printk( KERN_DEBUG "%s() called.\n", __FUNCTION__ );

	xyz3SspRestoreState( &SspSaveReg, SSP_PORT_ADC );

	ad7795RestoreState( &AdSaveReg );

	rc = adcTimerRestart();

	return rc;
}

static struct platform_driver xyz3adc_driver = {
	.probe		= xyz3_adc_probe,
	.remove		= xyz3_adc_remove,
	.suspend	= xyz3_adc_suspend,
	.resume		= xyz3_adc_resume,
	.driver		= {
		.name	= "xyz3-adc",
	},
};


/*
 * Driver cleanup routine.
 */
static void xyz3_adc_cleanup(void)
{
	dev_t devno = MKDEV(DevMajorNumber, DevMinorNumber);

	DPRN( printk(KERN_DEBUG "%s called.\n", __FUNCTION__) );

	adcTimerClean();

	xyz3RingBufFree( DataBuf );

	platform_driver_unregister(&xyz3adc_driver);

	cdev_del(&Adc_devst.cdev);

	unregister_chrdev_region(devno, DevCount);
}

/*
 * Driver initalize routine.
 */
static int __init xyz3_adc_init( void )
{
	int	result;
	int	err;
	int	devno;
	dev_t dev = 0;

	DPRN( printk(KERN_DEBUG "%s called.\n", __FUNCTION__) );

	// Append this device driver to system. 
	if (DevMajorNumber)
	{
		dev = MKDEV(DevMajorNumber, DevMinorNumber);
		result = register_chrdev_region(dev, DevCount, DevName);
	}
	else
	{
		result = alloc_chrdev_region(&dev, DevMinorNumber, DevCount, DevName);
		DevMajorNumber = MAJOR(dev);
	}

	if (result < 0) 
	{
		printk( KERN_ERR "%s:can't get major %d\n", __FUNCTION__, DevMinorNumber);
		return result;
	}

	memset( &Adc_devst, 0, sizeof(AdcDevSt) );

	init_MUTEX(&Adc_devst.sem);

	devno = MKDEV(DevMajorNumber, DevMinorNumber);

	cdev_init(&Adc_devst.cdev, &adc_fops);
	Adc_devst.cdev.owner = THIS_MODULE;
	Adc_devst.cdev.ops = &adc_fops;

	err = cdev_add (&Adc_devst.cdev, devno, 1);
	/* Fail gracefully if need be */
	if (err)
	{
		printk(KERN_NOTICE "%s:Error %d adding adc0\n", __FUNCTION__, err);
	}

	result = platform_driver_register(&xyz3adc_driver);
	if( result != 0 )
	{
		printk(KERN_ERR "%s:driver_register() failed. (%d)\n", __FUNCTION__, result);
	}

	// data buffer.
	DataBuf = xyz3RingBufInit( POS_BUF_NUM, sizeof(PosSensor) );
	if( DataBuf < 0 )
	{
		printk(KERN_ERR "%s:xyz3RingBufInit() failed!\n", __FUNCTION__);
		goto error;
	}

	// AD7795 Initialize
	adcInit();
	return result;

error:
	xyz3_adc_cleanup();
	return result;
}

module_init(xyz3_adc_init);
module_exit(xyz3_adc_cleanup);

EXPORT_SYMBOL(AdCallbackRegistry);
EXPORT_SYMBOL(AdCallbackUnregistry);

MODULE_LICENSE("GPL");

