#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/pm.h>

#include <asm/arch/pxa-regs.h>
#include <asm/arch/xyz3.h>

static void deepsleep( void )
{
    unsigned long pmnc;

    printk( KERN_DEBUG "%s: %s() called.\n", __FILE__, __FUNCTION__ );
    mdelay(10);

    /*
     * Power-off from kernel means serious error and
     * ACC-ON (GPIO-1) should be ignored for the next boot-up.
     */
    PSLR  |=  0x00400000; // Ignore VDD Fault in deep sleep mode.
    PWER  &= ~0x02; // Disable wake-up due to GPIO1 edge detect.
    PWER  |=  0x01; // Enable wake-up due to GPIO0 edge detect.
    PRER  &= ~0x03; // Disable wake-up due to GPIO1,0 rising edge detect.
    PFER  &= ~0x02; // Disable wake-up due to GPIO1 falling edge detect.
    PFER  |=  0x01; // Enable wake-up due to GPIO0 falling edge detect.

    // going to deep sleep mode.
    pmnc = 0x0f;
    asm ("mcr p14, 0, %0, c7, c0, 0" : : "r" (pmnc));
}

static void backlightoff( void )
{
    printk( KERN_DEBUG "%s: %s() called.\n", __FILE__, __FUNCTION__ );

    GAFR0_U &= ~(0x00000003);
    GPDR0   &= ( 1 << 16 );
    mdelay(10);
    GPCR0   &= ( 1 << 16 );
}

static void xyz3_power_off(void)
{
    printk( KERN_DEBUG "%s: %s() called.\n", __FILE__, __FUNCTION__ );
    mdelay(100);
    local_irq_disable();
    backlightoff();
    /*
     * set scratchpad to zero, just in case it is used as a
     * restart address by the bootloader.
     */
    PSPR = 0;
    /* enter sleep mode */
    deepsleep();
}

static void __init xyz3_power_off_init(void)
{
    pm_power_off = xyz3_power_off;
}
arch_initcall(xyz3_power_off_init);
