
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/major.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <linux/fs.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/cdev.h>
#include <linux/platform_device.h>

#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>
#include <asm/uaccess.h>
#include <asm/semaphore.h>


#include <asm/arch/system.h>
#include <asm/arch/pxa-regs.h>
#include <asm/arch/xyz3.h>
#include <asm/arch/xyz3-devnum.h>

#include <asm/arch/xyz3pwm.h>


/* Memory define */
#define PWM_BASE_ADDR		0x40B00000	// 
#define PWM_SIZE			0x00200000	// 
#define PWM_MAX_CHANNEL		4

//#define DEBUG
#ifdef DEBUG
#define DPRN(x)		x
#else
#define DPRN(x)
#endif


#if 0
static int DevMajorNumber = 0;		// major number
#else
static int DevMajorNumber = XYZ3_MAJOR_PWM;		// major number
#endif
static int DevMinorNumber = XYZ3_MINOR_PWM;		// minor number
static int DevCount = PWM_MAX_CHANNEL;			// 
static char *DevName = "pwm";		// device driver name


#define PWM_CTRL_SD	 		0x0040
#define PWM_PRESCALE_MASK	0x003F
#define PWM_DUTY_FD			0x0400
#define PWM_DUTY_MASK		0x03FF
#define PWM_PERIOD_MASK		0x03FF

typedef struct {
	unsigned long	ctrlReg;
	unsigned long	dutyReg;
	unsigned long	periodReg;
} pwm_reg;


typedef struct {
	int			port;
	int			used;
	pwm_reg		*preg;
	struct		semaphore sem;
	struct		cdev cdev;
} PwmDevSt;

PwmDevSt	*pwm_devst;

// buffer for Suspend/Resume
Xyz3PwmArg pwm_state[PWM_MAX_CHANNEL];



static int pwmInit( void )
{
	if( (CKEN & CKEN0_PWM0) == 0 )
	{
		DPRN( printk( KERN_DEBUG "%s:CKEN1_PWM0 enable.\n", __FUNCTION__ ) );
		CKEN |= CKEN0_PWM0;
	}
	return 0;
}

static int pwmGetCondition( PwmDevSt *pdev, Xyz3PwmArg *parg )
{
	pwm_reg *preg = pdev->preg;

	DPRN( printk( KERN_DEBUG "%s:preg=%lx\n", __FUNCTION__, (long)preg) );

	parg->ctrl = 0x00;
	if( preg->ctrlReg & PWM_CTRL_SD )
	{
		parg->ctrl |= XYZ3_PWM_SDM_ABRUPT;
	}
	if( preg->dutyReg & PWM_DUTY_FD )
	{
		parg->ctrl |= XYZ3_PWM_FULL_DUTY;
	}

	DPRN( printk( KERN_DEBUG "%s:prescale=%lx\n", __FUNCTION__, preg->ctrlReg) );
	DPRN( printk( KERN_DEBUG "%s:duty=%lx\n",     __FUNCTION__, preg->dutyReg) );
	DPRN( printk( KERN_DEBUG "%s:period=%lx\n",   __FUNCTION__, preg->periodReg) );

	parg->prescale = preg->ctrlReg     & PWM_PRESCALE_MASK;
	parg->duty     = preg->dutyReg     & PWM_DUTY_MASK;
	parg->period   = preg->periodReg   & PWM_PERIOD_MASK;
	return 0;
}


static int pwmSetCondition( PwmDevSt *pdev, Xyz3PwmArg *parg )
{
	pwm_reg *preg = pdev->preg;

	unsigned long  prescale = parg->prescale & PWM_PRESCALE_MASK;
	unsigned long  duty     = parg->duty & PWM_DUTY_MASK;
	unsigned long  period   = parg->period & PWM_PERIOD_MASK;

	DPRN( printk( KERN_DEBUG "%s:prescale=%ld\n", __FUNCTION__, prescale) );
	DPRN( printk( KERN_DEBUG "%s:duty=%ld\n",     __FUNCTION__, duty) );
	DPRN( printk( KERN_DEBUG "%s:period=%ld\n",   __FUNCTION__, period) );

	if( parg->ctrl & XYZ3_PWM_SDM_ABRUPT )
	{
		prescale |= PWM_CTRL_SD;
	}
	if( parg->ctrl & XYZ3_PWM_FULL_DUTY )
	{
		duty |= PWM_DUTY_FD;
	}

	preg->ctrlReg   = prescale;
	preg->dutyReg   = duty;
	preg->periodReg = period;

	return 0;
}


static int pwm_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
	int stat = 0;
	PwmDevSt	*pdev = (PwmDevSt *)filp->private_data;

	DPRN( printk( KERN_DEBUG "%s() called.\n", __FUNCTION__ ) );

	switch( cmd )
	{
		case XYZ3PWM_GET_CONDITION:
			stat = pwmGetCondition( pdev, (Xyz3PwmArg*)arg );
			break;

		case XYZ3PWM_SET_CONDITION:
			stat = pwmSetCondition( pdev, (Xyz3PwmArg*)arg );
			break;

		default:
			stat = -EINVAL;
			break;
	}
	return stat;
}


static int pwm_open(struct inode *inode, struct file *filp)
{
	PwmDevSt	*pdev;
	int			port;

	DPRN( printk( KERN_DEBUG "%s() called.\n", __FUNCTION__ ) );
	port = iminor( inode ) - XYZ3_MINOR_PWM;
	DPRN( printk( KERN_DEBUG "%s:port=%d\n", __FUNCTION__, port) );

	pdev = container_of(inode->i_cdev, PwmDevSt, cdev);
	filp->private_data = pdev;

	if( pdev->used )
	{
		printk( KERN_ERR "%s:port %d is using.\n", __FUNCTION__, port);
		return -EBUSY;
	}

	if( down_interruptible( &pdev->sem ) )
	{
		printk( KERN_ERR "%s:down_interruptible failed!!\n", __FUNCTION__ );
		return -ERESTARTSYS;
	}

	pdev->used = 1;
	pdev->port = port;
	up( &pdev->sem );

	return 0;
}


static int pwm_release( struct inode *inode, struct file *filp )
{
	PwmDevSt	*pdev;

	DPRN( printk( KERN_DEBUG "%s() called.\n", __FUNCTION__ ) );

	pdev = container_of(inode->i_cdev, PwmDevSt, cdev);

	if( down_interruptible( &pdev->sem ) )
	{
		printk( KERN_ERR "%s:down_interruptible failed!!\n", __FUNCTION__ );
		return -ERESTARTSYS;
	}

	pdev->used = 0;
	up( &pdev->sem );

	return 0;
}


static struct file_operations pwm_fops = {
	owner:	THIS_MODULE,
	ioctl:	pwm_ioctl,
	open:	pwm_open,
	release: pwm_release,
};

/*
 *  Power manager Support Routines
 */
int __init pwm_probe(struct platform_device *dev)
{
	DPRN( printk( KERN_DEBUG "%s() called.\n", __FUNCTION__ ) );
	return 0;
}

static int pwm_remove(struct platform_device *dev)
{
	DPRN( printk( KERN_DEBUG "%s() called.\n", __FUNCTION__ ) );
	return 0;
}

static int pwm_suspend(struct platform_device *dev, pm_message_t state)
{
	int i;
	printk( KERN_DEBUG "%s() called.\n", __FUNCTION__ );

	for (i = 0; i < DevCount; i++)
	{
		pwmGetCondition( &pwm_devst[i], &pwm_state[i]);
	}

	for (i = 0; i < DevCount; i++)
	{
		pwm_devst[i].preg->ctrlReg = PWM_CTRL_SD;
	}

	return 0;
}

static int pwm_resume(struct platform_device *dev)
{
	int i;
	printk( KERN_DEBUG "%s() called.\n", __FUNCTION__ );

	for (i = 0; i < DevCount; i++)
	{
		pwmSetCondition( &pwm_devst[i], &pwm_state[i]);
	}

	return 0;
}

static struct platform_driver xyz3pwm_driver = {
	.probe		= pwm_probe,
	.remove		= pwm_remove,
	.suspend	= pwm_suspend,
	.resume		= pwm_resume,
	.driver		= {
		.name	= "xyz3-pwm",
	},
};


/*
 * Driver cleanup routine.
 */
static void pwm_cleanup(void)
{
	int	i;
	dev_t devno = MKDEV(DevMajorNumber, DevMinorNumber);

	platform_driver_unregister(&xyz3pwm_driver);

	if( pwm_devst )
	{
		for (i = 0; i < DevCount; i++)
		{
			cdev_del(&pwm_devst[i].cdev);
		}
		kfree(pwm_devst);
		pwm_devst = NULL;
	}

	release_mem_region(PWM_BASE_ADDR, PWM_SIZE);
	unregister_chrdev_region(devno, DevCount);
}

/*
 * Driver initalize routine.
 */
static int __init pwm_init( void )
{
	int	i;
	int	result;
	int	err;
	int	devno;
	dev_t dev = 0;

	// Actually allocate io-port region. Always success.
	if( request_mem_region(PWM_BASE_ADDR, PWM_SIZE, DevName) == NULL )
	{
		printk( KERN_ERR "%s:request_mem_region() error!!\n", __FUNCTION__ );
		return -1;
	}

	// Append this device driver to system. 
	if (DevMajorNumber)
	{
		dev = MKDEV(DevMajorNumber, DevMinorNumber);
		result = register_chrdev_region(dev, DevCount, DevName);
	}
	else
	{
		result = alloc_chrdev_region(&dev, DevMinorNumber, DevCount, DevName);
		DevMajorNumber = MAJOR(dev);
	}

	if (result < 0) 
	{
		printk( KERN_ERR "%s:can't get major %d\n", __FUNCTION__, DevMinorNumber);
		return result;
	}

	pwm_devst = kmalloc(DevCount * sizeof(PwmDevSt), GFP_KERNEL);
	if (!pwm_devst) {
		result = -ENOMEM;
		goto error;
	}
	memset(pwm_devst, 0, DevCount * sizeof(PwmDevSt));

	for (i = 0; i < DevCount; i++)
	{
		init_MUTEX(&pwm_devst[i].sem);

		devno = MKDEV(DevMajorNumber, DevMinorNumber + i);

		cdev_init(&pwm_devst[i].cdev, &pwm_fops);
		pwm_devst[i].cdev.owner = THIS_MODULE;
		pwm_devst[i].cdev.ops = &pwm_fops;
		err = cdev_add (&pwm_devst[i].cdev, devno, 1);
		/* Fail gracefully if need be */
		if (err)
		{
			printk(KERN_NOTICE "%s:Error %d adding pwm%d", __FUNCTION__, err, i);
		}

		switch( i )
		{
			case 0:
//				pwm_devst[i].preg = (pwm_reg *)PWM_CTRL0;
				pwm_devst[i].preg = (pwm_reg *)io_p2v(0x40B00000);
				break;
			case 1:
//				pwm_devst[i].preg = (pwm_reg *)PWM_CTRL1;
				pwm_devst[i].preg = (pwm_reg *)io_p2v(0x40C00000);
				break;
			case 2:
//				pwm_devst[i].preg = (pwm_reg *)__REG(0x40B00010);
				pwm_devst[i].preg = (pwm_reg *)io_p2v(0x40B00010);
				break;
			case 3:
//				pwm_devst[i].preg = (pwm_reg *)__REG(0x40C00010);
				pwm_devst[i].preg = (pwm_reg *)io_p2v(0x40C00010);
				break;
		default:
				printk( KERN_ERR "%s:Invalid channel\n", __FUNCTION__ );
				break;
		}
	}

	pwmInit();

	result = platform_driver_register(&xyz3pwm_driver);
	if( result != 0 )
	{
		printk(KERN_ERR "%s:driver_register() failed. (%d)", __FUNCTION__, result);
	}

	return result;

error:
	pwm_cleanup();
	return result;
}


module_init(pwm_init);
module_exit(pwm_cleanup);

/*
 * ƁAinsmodƂɁA
 *  xyz3_pwm: Unknown symbol driver_unregister
 *  xyz3_pwm: Unknown symbol platform_bus_type
 *  xyz3_pwm: Unknown symbol driver_register
 * ƂȂ
 */
MODULE_LICENSE("GPL");

