
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/major.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <linux/fs.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/cdev.h>
#include <linux/platform_device.h>

#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>
#include <asm/uaccess.h>
#include <asm/semaphore.h>


#include <asm/arch/system.h>
#include <asm/arch/pxa-regs.h>
#include <asm/arch/xyz3.h>

#include <asm/arch/xyz3ssp.h>

//#define DEBUG
#ifdef DEBUG
#define DPRN(x)		x
#else
#define DPRN(x)
#endif


int PortBitWidth;
Xyz3SspDev Ad7795SspState;

int xyz3Ad7795SspInit( Xyz3SspDev *ssp )
{
	// AD7795 use SSP ch1.
	PortBitWidth = 0;

	// CKEN
	CKEN |= CKEN23_SSP1;

	// Memory
	memcpy( &Ad7795SspState, ssp, sizeof(Xyz3SspDev) );

	// SSP ch1 reg
	SSTO_P1  = Ad7795SspState.ssto;
	SSPSP_P1 = Ad7795SspState.sspsp;
	SSCR1_P1 = Ad7795SspState.sscr1;
	SSCR0_P1 = Ad7795SspState.sscr0;

	// Data size
	PortBitWidth = ( ( Ad7795SspState.sscr0 & SSCR0_EDSS ) >> 16 )
	               + ( Ad7795SspState.sscr0 & SSCR0_DSS ) - 1;

	return 0;
}


#define SSSR_TFL	(0xf << 8)	/* Transmit FIFO Level */

// Read/Write for AD7795
u32 xyz3Ad7795FrameTrans( u32 wrData, int wrBit, int rdBit )
{
	int timeout;
	u32 rdData;


	// Changeover of transration bit width.
	if( PortBitWidth != (wrBit + rdBit) )
	{
		PortBitWidth = wrBit + rdBit;
		SSCR0_P1 = 0;

		Ad7795SspState.sscr0 &= ~(SSCR0_EDSS | SSCR0_DSS);
		if( PortBitWidth > 17 )
		{
			Ad7795SspState.sscr0 |= SSCR0_EDSS;
		}
		Ad7795SspState.sscr0 |= ( (PortBitWidth-1) & SSCR0_DSS );

		SSTO_P1  = Ad7795SspState.ssto;
		SSPSP_P1 = Ad7795SspState.sspsp;
		SSCR1_P1 = Ad7795SspState.sscr1;
		SSCR0_P1 = Ad7795SspState.sscr0;
	}

	timeout = 0;
	SSDR_P1 = wrData << rdBit;
	while( (SSSR_P1 & (SSSR_TFL | SSSR_BSY)) != 0 )
	{
		if( timeout ++ > 1000 )
		{
			printk( KERN_ERR "%s():Transmit busy.(0x%08x) \n", __FUNCTION__,  SSSR_P1);
			return -EBUSY;
		}
	}

	timeout = 0;
	while( (SSSR_P1 & SSSR_RNE) == 0 )
	{
		if( timeout ++ > 1000 )
		{
			printk( KERN_ERR "%s():Data not receive.(0x%08x) \n", __FUNCTION__,  SSSR_P1);
			return -EBUSY;
		}
	}

	rdData = SSDR_P1 & ( (1 << rdBit) -1 );
	return rdData;
}


/*
 *  Power manager Support Routines
 */
int xyz3SspSaveState(Xyz3SspDev *pdev, int port)
{
	DPRN( printk( KERN_DEBUG "%s() called.\n", __FUNCTION__ ) );

	pdev->sscr0 = SSCR0_P(port);
	pdev->sscr1 = SSCR1_P(port);
	pdev->sssr  = SSSR_P(port);
	pdev->ssto  = SSTO_P(port);
	pdev->sspsp = SSPSP_P(port);

	// CKEN
	switch( port )
	{
		case 1:
			CKEN &= ~CKEN23_SSP1;
			break;
		case 2:
			CKEN &= ~CKEN3_SSP2;
			break;
		case 3:
			CKEN &= ~CKEN4_SSP3;
			break;
		default:
			return -EINVAL;
	}
	udelay(10);

	return 0;
}

int xyz3SspRestoreState(Xyz3SspDev *pdev, int port)
{
	DPRN( printk( KERN_DEBUG "%s() called.\n", __FUNCTION__ ) );

	// CKEN
	switch( port )
	{
		case 1:
			CKEN |= CKEN23_SSP1;
			break;
		case 2:
			CKEN |= CKEN3_SSP2;
			break;
		case 3:
			CKEN |= CKEN4_SSP3;
			break;
		default:
			return -EINVAL;
	}

	udelay(10);

	SSCR1_P(port) = pdev->sscr1;
	SSSR_P(port)  = pdev->sssr;
	SSTO_P(port)  = pdev->ssto;
	SSPSP_P(port) = pdev->sspsp;
	SSCR0_P(port) = pdev->sscr0 & ~SSCR0_SSE;

	return 0;
}


EXPORT_SYMBOL(xyz3Ad7795SspInit);
EXPORT_SYMBOL(xyz3Ad7795FrameTrans);
EXPORT_SYMBOL(xyz3SspSaveState);
EXPORT_SYMBOL(xyz3SspRestoreState);
