/*
 * File:         drivers/char/tea5764hn.c
 * Based on:	 drivers/
 * Author:	 takeshi oosumimoto
 *
 * Created:	 June 8th, 2007
 * Description:	
 * Rev:
 *
 * Modified:
 *               Copyright 2007 Sony Co.
 *
 * Bugs:
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.
 * If not, write to the Free Software Foundation,
 * 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <linux/types.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/init.h>
#include <linux/input.h>
#include <linux/interrupt.h>
#include <linux/i2c.h>
#include <linux/delay.h>
#include <linux/kthread.h>
#include <linux/sched.h>

#include <asm/uaccess.h>
#include <asm/irq.h>
#include <asm/semaphore.h>

#include <asm/io.h>
#include <asm/arch/pxa-regs.h>
#include <asm/arch/xyz3-devnum.h>
#include <asm/arch/navu-gpio.h>
#include <asm/arch/navu-gpio-kernel.h>
#include "linux/tea5764.h"

//#include "i2cformlinux.h"
#include "I2CForm.h"
#include "IRDSListener.h"
#include "CTEA5764Driver.h"


#define TE5764_XINTPORT NAVU_IN_RDS_xIRQ

/* SlaveAddress 7bit ȤȤ1bitեȤR/WORƤ */
#define TEA5764_TUNE_RDS_MODE_SHIFT 0x22
#define TEA5764_TUNE_FM_MODE_SHIFT 0x20

#define TEA5764_TUNE_RDS_READMODE (TEA5764_TUNE_RDS_MODE_SHIFT | 0x01)
#define TEA5764_TUNE_FM_READMODE (TEA5764_TUNE_FM_MODE_SHIFT | 0x01)
#define TEA5764_TUNE_RDS_WRITEMODE TEA5764_TUNE_FM_MODE_SHIFT
#define TEA5764_TUNE_FM_WRITEMODE TEA5764_TUNE_FM_MODE_SHIFT

static int I2CFormLinux_i2c_send(unsigned char bSlaveAddress,char *pBuffer,unsigned long size,long *pExecWriteSize);
static int I2CFormLinux_i2c_recv(unsigned char bSlaveAddress,char *pBuffer,unsigned long size,long *pExecReadSize);
static int I2CFormLinux_init(void);
static void I2CFormLinux_PowerUp(void);
static void I2CFormLinux_PowerDown(void);
#if 0
static int I2CFormLinux_i2c_sendPowerUp(unsigned char bSlaveAddress,char *pBuffer,unsigned long size,long *pExecReadSize);
static int I2CFormLinux_Deinit(void);
#endif

static BOOL TMC_RDSReadData(unsigned char * pBufIn,unsigned long dwLenIn,unsigned char * pBufOut,unsigned long dwLenOut,unsigned long *pdwActualOut);
static BOOL TMC_RDSPreset(char * pBufIn,unsigned long dwLenIn,char * pBufOut,unsigned long dwLenOut,unsigned long *pdwActualOut);
static BOOL TMC_RDSSeekFreq(char * pBufIn,unsigned long dwLenIn,char * pBufOut,unsigned long dwLenOut,unsigned long *pdwActualOut);
static BOOL TMC_RDSGetLevel(char * pBufIn,unsigned long dwLenIn,char * pBufOut,unsigned long dwLenOut,unsigned long *pdwActualOut);
static void TMC_ClearRingCount(void);
static void TMC_RDSAllCancel(void);
static BOOL TMC_RDSNaviAtt(char * pBufIn,unsigned long dwLenIn,char * pBufOut,unsigned long dwLenOut,unsigned long *pdwActualOut);
static BOOL TMC_RDSGetAttSts(char * pBufIn,unsigned long dwLenIn,char * pBufOut,unsigned long dwLenOut,unsigned long *pdwActualOut);
static void TMC_RegLock(unsigned long *a);
static void TMC_RegUnLock(unsigned long *b);
static void TMC_ReadDataLock(void);
static void TMC_ReadDataUnLock(void);
static void TMC_ChangeMode(int iMode);
static void TMC_RDSInit(void);
static BOOL TMC_ReInitRds(void);
static BOOL TMC_MakeNGData(char * pBufOut,unsigned long dwLenOut);
static int TMC_ReadModeCheck(void);
static bool TMC_CheckCradle(void);

#if 0
//static int TEASWPortThread(void *data);
static void TEASWPortThread(unsigned long data);
#endif

static void tea5764_update_rds_data(int Event,unsigned char *pBuffer,int size);

#define TMC_SEEK_TIMEOUT 5000
#define RDS_CANCEL_TIMEOUT 1000

/* exec mode */
#define RDS_TMC_MODE_IDLE 0
#define RDS_TMC_MODE_SEEK 1
#define RDS_TMC_MODE_READ 2
#define RDS_TMC_MODE_CANCEL_WAIT 3

/* flag init*/
#define TMC_WAIT_INIT 0

/* Preset flag */
#define TMC_WAIT_PRESET_END 1

#define WAIT_FAILED 6

/* Read data flag */
#define TMC_WAIT_RDS_UPDATE 1
#define TMC_WAIT_RDS_DATAREAD_CANCEL 2

/* seek flag */
#define TMC_WAIT_SEEK_END 1
#define TMC_WAIT_SEEKLIMIT 2
#define TMC_WAIT_SEEK_CANCEL 3

#define TMC_RDS_SEEK 0
#define TMC_RDS_SEEKLIMIT 1
#define TMC_RDS_DATA_UPDATE 2
#define TMC_RDS_PRESET_END 3
#define RDS_READ_CANCEL 4

/* dataХåեȤꤢ */
#define RDS_READ_BUFFER_MAX 50
#define RDS_READ_BUFFER_SIZE 0x13

/* NGǡ */
#define RDS_DATA_FLAG_LSYN 0xA0
#define RDS_DATA_STAT1_NG 0x73
#define RDS_DATA_STAT2_NG 0xF8

/*쥸κ祵*/
#define TEA5764_READ_MAX 28
#define TEA5764_WRITE_MAX 11

/* WaitFlag define */
#define WAIT_TIMEOUT 20
#define WAIT_ABANDONED_0 10
#define WAIT_OBJECT_0 30
#define RDS_SEEK_TYPE_SEEKEND 0 
#define RDS_SEEK_TYPE_LIMITEND 1 
#define RDS_SEEK_TYPE_SEEKERROR 2 
#define RDS_SEEKLOOP_MAX 2

static IRDSListener *pte5764_ListenerPtr = NULL;
static II2CForm *pte5764_FormPtr = NULL;

static wait_queue_head_t tmc_preset_ack;
static wait_queue_head_t tmc_tune_ack;
static wait_queue_head_t tmc_dataread_ack;
static int tmc_preset_flag;
static int tmc_tune_flag;
static int tmc_dataread_flag;

static wait_queue_head_t tmc_exec_cancel;
static int tmc_cancel_flag = 0;

static char RdsBufferReadCnt = 0;
static char RdsBufferWriteCnt = 0;
static unsigned char RDSReadData[RDS_READ_BUFFER_MAX][RDS_READ_BUFFER_SIZE];

#define TMC_CRADLE_ON 0
#define TMC_CRADLE_OFF 1

static int CradleFlag = TMC_CRADLE_ON;

//static wait_queue_head_t tmc_swp_timer;

//#define CONFIG_SAA6588

#undef CONFIG_SAA6588

#undef CONFIG_TIME_MEASURE

#ifdef CONFIG_TIME_MEASURE
#define TIME_MEASURE_GPIO 11
#endif

//#define TEA5764_MAJOR 245		//experimental

#define TEA5764_DRV_NAME         "tea5764"
#define TEA5764_FM_DRV_NAME      "tea5764 FM"
#define TEA5764_RDS_DRV_NAME     "tea5764 RDS"

#define TEA5764_FM_I2C_ID		0xE623
#define TEA5764_RDS_I2C_ID		0xE624

#define TEA5764_FM_I2C_ADDR		0x10
#define TEA5764_RDS_I2C_ADDR		0x11



#define RDS_BUFFER 256

#define POLL_INTERVAL (HZ/25)	/*40 ms*/


static int tea5764_suspend(struct device *dev, pm_message_t state);
static int tea5764_resume(struct device *dev);
static int tea5764_IntThread(void *data);
static int NavuCradleThread(void *data);
static int tea5764_fm_attach (struct i2c_adapter *adap);
static int tea5764_rds_attach (struct i2c_adapter *adap);
static int detach_client (struct i2c_client *client);


//static DECLARE_WAIT_QUEUE_HEAD(tea5764_wait);
//static __u8 rdsin=0,rdsout=0,rdsstat=0;
//static int users=0;
//static struct timer_list readtimer;
//static unsigned char rdsbuf[RDS_BUFFER];
//static unsigned char prev_blid;

//static spinlock_t tea5764_io_lock;
//static wait_queue_head_t read_queue;

//static void tea5764_defer_work(void *arg);

//static struct workqueue_struct *tea5764_workqueue;
//static DECLARE_WORK(tea5764_work, tea5764_defer_work,0);

static struct i2c_driver tea5764_fm_driver;
static struct i2c_client *tea5764_fm_client;

static struct i2c_driver tea5764_rds_driver;
static struct i2c_client *tea5764_rds_client;

static DEFINE_SPINLOCK(tea5764_lock);
//static DEFINE_SPINLOCK(tea5764_ReadDatalock);
//static DEFINE_SPINLOCK(tea5764_Modelock);

//static int tmc_opencnt = 0;
struct task_struct *tea5764_intx_th;
struct task_struct *tea5764_Cradle_th;
struct tasklet_struct tmc_att_tasklet;
//DECLARE_TASKLET(tmc_att_tasklet,TEASWPortThread,0);

static unsigned short ignore[] = { I2C_CLIENT_END };
static unsigned short normal_addr_tea5764_fm[] = { TEA5764_FM_I2C_ADDR, I2C_CLIENT_END };
static unsigned short normal_addr_tea5764_rds[] = { TEA5764_RDS_I2C_ADDR, I2C_CLIENT_END };

static struct i2c_driver tea5764_fm_driver = {
	.owner		= THIS_MODULE,
// .driver = {
	.name = TEA5764_FM_DRV_NAME,
 // },
//  .id = TEA5764_FM_I2C_ID,
	.flags		= I2C_DF_NOTIFY,
	.attach_adapter = tea5764_fm_attach,
	.detach_client = detach_client,
	.command	= 0,
	.driver = {
		.suspend	= tea5764_suspend,
		.resume		= tea5764_resume,
	}
};
static struct i2c_driver tea5764_rds_driver = {
	.owner		= THIS_MODULE,
//  .driver = {
  .name = TEA5764_RDS_DRV_NAME,
//  },
	.flags		= I2C_DF_NOTIFY,
//  .id = TEA5764_RDS_I2C_ID,
	.attach_adapter = tea5764_rds_attach,
	.detach_client = detach_client,
	.command	= 0,
};

static struct i2c_client_address_data addr_data_tea5764_fm = {
  .normal_i2c = normal_addr_tea5764_fm,
  .probe = ignore,
  .ignore = ignore,
};

static struct i2c_client_address_data addr_data_tea5764_rds = {
  .normal_i2c = normal_addr_tea5764_rds,
  .probe = ignore,
  .ignore = ignore,
};

/*-----------------------------------------------------------------------------------------------*/

/* tea5764_fm_probe (struct i2c_adapter *adap, int addr, int kind) */


/*-----------------------------------------------------------------------------------------------*/
static int tea5764_fm_probe (struct i2c_adapter *adap, int addr, int kind)
{
	struct i2c_client *client;
	int rc;

	printk(KERN_INFO "tea5764_fm_probe\n");
	client = kmalloc (sizeof (struct i2c_client), GFP_KERNEL);
	if (!client) {
		return -ENOMEM;
	}


	memset (client, 0, sizeof (struct i2c_client));
	strncpy (client->name, TEA5764_FM_DRV_NAME, I2C_NAME_SIZE);
	client->addr = addr;
	client->adapter = adap;
	client->driver = &tea5764_fm_driver;

	if ((rc = i2c_attach_client (client)) != 0) {
		kfree (client);
		printk (KERN_ERR "i2c_attach_client fail: %d\n", rc);
		return rc;
	}

	tea5764_fm_client = client;
	printk(KERN_INFO "%s_attach: at 0x%02x\n",client->name, client->addr);
	return 0;
}

/*-----------------------------------------------------------------------------------------------*/

/* tea5764_rds_probe (struct i2c_adapter *adap, int addr, int kind) */


/*-----------------------------------------------------------------------------------------------*/
static int tea5764_rds_probe (struct i2c_adapter *adap, int addr, int kind)
{
	struct i2c_client *client;
	int rc;

	client = kmalloc (sizeof (struct i2c_client), GFP_KERNEL);
	if (!client)
		return -ENOMEM;

	memset (client, 0, sizeof (struct i2c_client));
	strncpy (client->name, TEA5764_RDS_DRV_NAME, I2C_NAME_SIZE);
	client->addr = addr;
	client->adapter = adap;
	client->driver = &tea5764_rds_driver;

	if ((rc = i2c_attach_client (client)) != 0) {
		kfree (client);
		printk (KERN_ERR "i2c_attach_client fail: %d\n", rc);
		return rc;
	}

	tea5764_rds_client = client;
	printk(KERN_INFO "%s_attach: at 0x%02x\n",client->name, client->addr);
	return 0;
}

/*-----------------------------------------------------------------------------------------------*/

/* tea5764_fm_attach (struct i2c_adapter *adap) */


/*-----------------------------------------------------------------------------------------------*/
static int tea5764_fm_attach (struct i2c_adapter *adap)
{
	return i2c_probe(adap, &addr_data_tea5764_fm, &tea5764_fm_probe);
}
/*-----------------------------------------------------------------------------------------------*/

/* tea5764_rds_attach (struct i2c_adapter *adap) */


/*-----------------------------------------------------------------------------------------------*/
static int tea5764_rds_attach (struct i2c_adapter *adap)
{
	return i2c_probe(adap, &addr_data_tea5764_rds, &tea5764_rds_probe);
}

/*-----------------------------------------------------------------------------------------------*/

/* detach_client (struct i2c_client *client) */


/*-----------------------------------------------------------------------------------------------*/
static int detach_client (struct i2c_client *client)
{
	int rc;
	if ((rc = i2c_detach_client (client)) == 0) {
		kfree (i2c_get_clientdata (client));
	}
	return rc;
}

/*-----------------------------------------------------------------------------------------------*/

/* tea5764_i2c_read(struct i2c_client *client,u8 *buf,u16 len) */


/*-----------------------------------------------------------------------------------------------*/

static int tea5764_i2c_read(struct i2c_client *client,u8 *buf,u16 len)
{
	int ret = -1;
		
	if (i2c_check_functionality(client->adapter, I2C_FUNC_I2C)) {
		if((ret = i2c_master_recv(client, buf, len)) < 0){
			printk(KERN_ERR "TEA5764: I2C transfer error\n");
			return ret;
		}
	} else {
		printk(KERN_EMERG "AD7142: i2c bus doesn't support raw I2C operation\n");
	}
	return ret;	
}

/*-----------------------------------------------------------------------------------------------*/

/* tea5764_i2c_send(struct i2c_client *client,u8 *buf,u16 len) */


/*-----------------------------------------------------------------------------------------------*/
static int tea5764_i2c_send(struct i2c_client *client,u8 *buf,u16 len)
{
	int ret = -1;
	
	if (i2c_check_functionality(client->adapter, I2C_FUNC_I2C)) {
		if((ret = i2c_master_send(client, buf, len)) < 0){
			printk(KERN_EMERG "TEA5764: I2C transfer error\n");
			return ret;
		}
	} else {
		printk(KERN_EMERG "AD7142: i2c bus doesn't support raw I2C operation\n");
	}
	return ret;	
}


#define DEBMSG_INIT_HANDLE(a)
static int tea5764_IntThread(void *data)
{
	int ret;
	navu_irq_ioctl_t tmcgpioint;
	unsigned long flag;
	/* portȥꥬ */
	tmcgpioint.port = TE5764_XINTPORT;
	tmcgpioint.trigger = NAVU_IRQ_FALL_EDGE;

	DEBMSG_INIT_HANDLE(printk(KERN_INFO "tea5764_IntThread start\n");)
	while(!kthread_should_stop()){
		DEBMSG_INIT_HANDLE(printk(KERN_INFO "------------------- tea5764_IntThread 1 ----------------------\n");)
		if(navuGpioIrqWait( &tmcgpioint ) == 0) {
			DEBMSG_INIT_HANDLE(printk(KERN_INFO "------------------- tea5764_IntThread 2 ----------------------\n");)
			/* 褿ɤ߽Ф */
			TMC_RegLock(&flag);
			if(CradleFlag == TMC_CRADLE_ON) {
				ret = Ctea5764Driver_int_handler();
			}
			TMC_RegUnLock(&flag);
		}
		try_to_freeze();
	}
	DEBMSG_INIT_HANDLE(printk(KERN_INFO "tea5764_IntThread end\n");)
	return 0;
}
#define DEBMSG_CRADLE(a)
static int NavuCradleThread(void *data)
{
	int ret;
	unsigned long flag;
	navu_irq_ioctl_t tmcgpioint;
	/* portȥꥬ */
	tmcgpioint.port = NAVU_IN_ON_CRADLE;
	tmcgpioint.trigger = NAVU_IRQ_DOUBLE_EDGE;

	while(!kthread_should_stop()){
	DEBMSG_CRADLE(printk(KERN_INFO "NavuCradleThread\n");)
		ret = navuGpioIrqWait( &tmcgpioint );
		try_to_freeze();
		/* 褿ɤ߽Ф */
		TMC_RegLock(&flag);
		if (ret == 1) {
			DEBMSG_CRADLE(printk(KERN_INFO "Cradle GPIO41 1\n");)
			CradleFlag = TMC_CRADLE_OFF;
		} else if (ret == 0) {
			DEBMSG_CRADLE(printk(KERN_INFO "Cradle GPIO41 0\n");)
			TMC_ReInitRds();
			CradleFlag = TMC_CRADLE_ON;
		} else {
			DEBMSG_CRADLE(printk(KERN_INFO "Cradle GPIO41 Error\n");)
		}
		TMC_RegUnLock(&flag);
	}
	DEBMSG_CRADLE(printk(KERN_EMERG "NavuCradleThread end\n");)
	return 0;
}
/***********************************************************
*
* FUNCTION NAME :tea5764_open
*
* INPUTS/OUTPUTS:
* in_inode - Description of openned file.
* in_filp - Description of openned file.
*
* RETURN
* 0: Open ok.
* -ENXIO  No such device
*
* FUNCTION(S) CALLED:
*
* GLOBAL VARIABLES REFERENCED:
*
* GLOBAL VARIABLES MODIFIED: NIL
*
* DESCRIPTION: It is invoked when user call 'open' system call
*              to open spi device.
*
* CAUTION:
*************************************************************
* MODIFICATION HISTORY :
**************************************************************/
#define DEBMSG_OPEN(a)
static int tea5764_open(struct inode *inode, struct file *filp)
{
//	unsigned long flags;
	unsigned long minor_no;
#ifdef CPU_REG
	u32 ckenreg = 0;
#endif
	int datadata;

	CradleFlag = TMC_CRADLE_ON;

//	spin_lock_irqsave(&tea5764_lock, flags);
//	users++;
//	if (1 == users) 
//		init_waitqueue_head(&read_queue);
		

#ifdef CONFIG_TIME_MEASURE
    if(gpio_request(TIME_MEASURE_GPIO, NULL)){
		DEBMSG_OPEN(printk(KERN_ERR "Failed ro request GPIO_%d\n",TIME_MEASURE_GPIO);)
		return -EBUSY;
    }
    gpio_direction_output(TIME_MEASURE_GPIO);
#endif

	//ԵؿΥϥɥ
	init_waitqueue_head(&tmc_preset_ack);
	init_waitqueue_head(&tmc_tune_ack);
	init_waitqueue_head(&tmc_dataread_ack);
	init_waitqueue_head(&tmc_exec_cancel);
	tmc_preset_flag =0;
	tmc_tune_flag =0;
	tmc_dataread_flag =0;
#ifdef CPU_REG
	ckenreg = CKEN;
	printk(KERN_EMERG "CKEN REG 0x%08X\n",ckenreg);
	if ((ckenreg & CKEN14_I2C) != 0) {
		printk(KERN_EMERG "CKEN14_I2C Enable\n");
	} else {
		printk(KERN_EMERG "CKEN14_I2C disable\n");
	}
#endif
//	spin_unlock_irqrestore(&tea5764_lock, flags);
	if (navuGpioGetLevel(NAVU_IN_ON_CRADLE) == 0) {
		DEBMSG_OPEN(printk(KERN_EMERG "NAVU_IN_ON_CRADLE 0\n");)
		CradleFlag = TMC_CRADLE_ON;
		//ǥХν
		minor_no = iminor(inode);
		if (minor_no == XYZ3_MINOR_TMC) {
			navuGpioSetLevel(NAVU_OUT_RDS_POWER, 0);
//			mdelay(20);
			navuGpioSetLevel(NAVU_OUT_RDS_POWER, 1);
			tea5764_intx_th = kthread_run(tea5764_IntThread,(void *)&datadata,"tea5764_int");
			if(IS_ERR(tea5764_intx_th)) {
				printk(KERN_ERR "tmc make kthread error\n");
				return -EBUSY;
			}

			tea5764_Cradle_th = kthread_run(NavuCradleThread,(void *)&datadata,"tea5764Cradle_int");
			if(IS_ERR(tea5764_Cradle_th)) {
				printk(KERN_ERR "tmc make kthread cradle error\n");
				return -EBUSY;
			}
//			Ctea5764Driver_init(pte5764_ListenerPtr,pte5764_FormPtr);
			Ctea5764Driver_init2nd(pte5764_ListenerPtr,pte5764_FormPtr);
			TMC_ChangeMode(RDS_TMC_MODE_IDLE);
		} else {
			navuGpioSetLevel(NAVU_OUT_RDS_POWER, 1);
		}
	} else{
		DEBMSG_OPEN(printk(KERN_EMERG "NAVU_IN_ON_CRADLE 1\n");)
		CradleFlag = TMC_CRADLE_OFF;
		minor_no = iminor(inode);
		if (minor_no == XYZ3_MINOR_TMC) {
			navuGpioSetLevel(NAVU_OUT_RDS_POWER, 0);
			mdelay(20);
			navuGpioSetLevel(NAVU_OUT_RDS_POWER, 1);
			tea5764_intx_th = kthread_run(tea5764_IntThread,(void *)&datadata,"tea5764_int");
			if(IS_ERR(tea5764_intx_th)) {
				printk(KERN_ERR "tmc make kthread error\n");
				return -EBUSY;
			}

			tea5764_Cradle_th = kthread_run(NavuCradleThread,(void *)&datadata,"tea5764Cradle_int");
			if(IS_ERR(tea5764_Cradle_th)) {
				printk(KERN_ERR "tmc make kthread cradle error\n");
				return -EBUSY;
			}
			TMC_ChangeMode(RDS_TMC_MODE_IDLE);
		} else {
		}
	}

#if 0
	/* å */
	if (tmc_opencnt == 0) {
		navuGpioSetLevel(NAVU_OUT_RDS_POWER, 0);
		mdelay(20);
		navuGpioSetLevel(NAVU_OUT_RDS_POWER, 1);
		tea5764_intx_th = kthread_run(tea5764_IntThread,(void *)&datadata,"tea5764_int");
		if(IS_ERR(tea5764_intx_th)) {
			printk(KERN_ERR "tmc make kthread error\n");
			return -EBUSY;
		}

		tea5764_Cradle_th = kthread_run(NavuCradleThread,(void *)&datadata,"tea5764Cradle_int");
		if(IS_ERR(tea5764_Cradle_th)) {
			printk(KERN_ERR "tmc make kthread cradle error\n");
			return -EBUSY;
		}

		/* 񻺤ͭ */
		pte5764_FormPtr = kmalloc(sizeof(II2CForm),GFP_KERNEL);
		if (pte5764_FormPtr == NULL) {
			return -EBUSY;
		}
		pte5764_FormPtr->i2c_send = I2CFormLinux_i2c_send;
		pte5764_FormPtr->i2c_recv = I2CFormLinux_i2c_recv;
		pte5764_FormPtr->i2c_init = I2CFormLinux_init;
		pte5764_FormPtr->i2c_PwSend = I2CFormLinux_i2c_send;
		pte5764_FormPtr->i2c_PwUp = I2CFormLinux_PowerUp;
		pte5764_FormPtr->i2c_PwDown = I2CFormLinux_PowerDown;

		pte5764_ListenerPtr = kmalloc(sizeof(IRDSListener),GFP_KERNEL);
		if (pte5764_ListenerPtr == NULL) {
			kfree(pte5764_FormPtr);
			pte5764_FormPtr = NULL;
			return -EBUSY;
		}
		pte5764_ListenerPtr->update_rds_data = tea5764_update_rds_data;
		TMC_ClearRingCount();
		if (navuGpioGetLevel(NAVU_IN_ON_CRADLE) == 0) {
			printk(KERN_EMERG "NAVU_IN_ON_CRADLE 0\n");
			CradleFlag = TMC_CRADLE_ON;
			//ǥХν
			minor_no = iminor(inode);
			if (minor_no == XYZ3_MINOR_TMC) {
				navuGpioSetLevel(NAVU_OUT_RDS_POWER, 0);
				mdelay(20);
				navuGpioSetLevel(NAVU_OUT_RDS_POWER, 1);
				tea5764_intx_th = kthread_run(tea5764_IntThread,(void *)&datadata,"tea5764_int");
				if(IS_ERR(tea5764_intx_th)) {
					printk(KERN_ERR "tmc make kthread error\n");
					return -EBUSY;
				}

				tea5764_Cradle_th = kthread_run(NavuCradleThread,(void *)&datadata,"tea5764Cradle_int");
				if(IS_ERR(tea5764_Cradle_th)) {
					printk(KERN_ERR "tmc make kthread cradle error\n");
					return -EBUSY;
				}
				Ctea5764Driver_init(pte5764_ListenerPtr,pte5764_FormPtr);
			} else {
				navuGpioSetLevel(NAVU_OUT_RDS_POWER, 1);
			}
		} else{
			printk(KERN_EMERG "NAVU_IN_ON_CRADLE 1\n");
			CradleFlag = TMC_CRADLE_OFF;
		}
		TMC_ChangeMode(RDS_TMC_MODE_IDLE);
		tmc_opencnt++;
	} else {
		minor_no = iminor(inode);
		if (minor_no == XYZ3_MINOR_TMC) {
			Ctea5764Driver_init(pte5764_ListenerPtr,pte5764_FormPtr);
		} else {
		}
		tmc_opencnt++;
	}
#endif
	return 0;
}

/*-----------------------------------------------------------------------------------------------*/

/* tea5764_release(struct inode *inode, struct file *filp) */


/*-----------------------------------------------------------------------------------------------*/
static int tea5764_release(struct inode *inode, struct file *filp)
{
//	unsigned long flags;
	unsigned long minor_no;

//	spin_lock_irqsave(&tea5764_lock, flags);
//	users--;
//	if (0 == users){
//		del_timer_sync(&readtimer);
//		rdsstat=0;
//	}
#ifdef CONFIG_TIME_MEASURE
    gpio_free(TIME_MEASURE_GPIO);
#endif

	minor_no = iminor(inode);
	if (minor_no == XYZ3_MINOR_TMC) {
		if (tea5764_intx_th != NULL) {
			navuGpioIrqCancel(TE5764_XINTPORT);
			kthread_stop(tea5764_intx_th);
			tea5764_intx_th = NULL;
		}
		if (tea5764_Cradle_th != NULL) {
			navuGpioIrqCancel(NAVU_IN_ON_CRADLE);
			kthread_stop(tea5764_Cradle_th);
			tea5764_Cradle_th = NULL;
		}
	}
//	spin_unlock_irqrestore(&tea5764_lock, flags);
//		if (tmc_att_tasklet != NULL) {
//			tasklet_kill(&tmc_att_tasklet);
//			tmc_att_tasklet = NULL;
//		}
	return 0;
}

/***********************************************************
*
* FUNCTION NAME :tea5764_read
*
* INPUTS/OUTPUTS:
* in_filp - Description of openned file.
* in_count - how many bytes user wants to get.
* out_buf - data would be write to this address.
*
* RETURN
* positive number: bytes read back
* -ENODEV When minor not available.
* -EMSGSIZE When size more than a single ASCII digit followed by /n.
*
* FUNCTION(S) CALLED:
*
* GLOBAL VARIABLES REFERENCED:
*
* GLOBAL VARIABLES MODIFIED: NIL
*
* DESCRIPTION: It is invoked when user call 'read' system call
*              to read from system.
*
* CAUTION:
*************************************************************
* MODIFICATION HISTORY :
**************************************************************/
static ssize_t tea5764_read(struct file *file, char __user *data, size_t count, loff_t *ppos)
{
#if 0
	int i=0;
	unsigned char readbuf[RDS_BUFFER];

	if(rdsstat==0) {
		spin_lock(&tea5764_io_lock);
		rdsstat=1;
		spin_unlock(&tea5764_io_lock);
		init_timer(&readtimer);
		readtimer.function=tea5764_handler;
		readtimer.data=(unsigned long)0;
		readtimer.expires=jiffies+POLL_INTERVAL;
		add_timer(&readtimer);
	}

	if(rdsin==rdsout) {
		if (file->f_flags & O_NONBLOCK)
			return -EWOULDBLOCK;
		interruptible_sleep_on(&read_queue);
	}

	while( i<count && rdsin!=rdsout)
		readbuf[i++]=rdsbuf[rdsout++];

	if (copy_to_user(data,readbuf,i))
		return -EFAULT;
	return i;
#endif
	return 0;
}


/***********************************************************
*
* FUNCTION NAME :tea5764_ioctl
*
* INPUTS/OUTPUTS:
* in_inode - Description of openned file.
* in_filp - Description of openned file.
* in_cmd - Command passed into ioctl system call.
* in/out_arg - It is parameters which is specified by last command
*
* RETURN:
* 0 OK
* -EINVAL
*
* FUNCTION(S) CALLED:
*
* GLOBAL VARIABLES REFERENCED:
*
* GLOBAL VARIABLES MODIFIED: NIL
*
* DESCRIPTION:
*
* CAUTION:
*************************************************************
* MODIFICATION HISTORY :
**************************************************************/
#define DEBMSG_IOCTL(a)
static int tea5764_ioctl(struct inode *inode, struct file *filp, uint cmd, unsigned long arg)
{

	int iFreqency = 0;
	i2c_message msg;
//	unsigned long flags;
	long dwActualOut;
	bool ret;
	TMC_RDS_SEEKIN *seekin;
	TMC_RDS_SEEKOUT SeekOut;
	int outlen;
	int outexeclen;
	if (CradleFlag == TMC_CRADLE_OFF) {
		return -ENOTTY;
	}
	switch (cmd) {
	case FM_I2C_READ:
		{
		if (copy_from_user(&msg,(struct i2c_message __user *)arg,sizeof(msg)))
			return -EFAULT;
						
			tea5764_i2c_read(tea5764_fm_client,msg.buf,msg.len);
			break;
		}
	case FM_I2C_WRITE:
		{
		if (copy_from_user(&msg,(struct i2c_message __user *)arg,sizeof(msg)))
			return -EFAULT;

			tea5764_i2c_send(tea5764_fm_client,msg.buf,msg.len);
			break;
		}
	case RDS_I2C_READ:
		{
		if (copy_from_user(&msg,(struct i2c_message __user *)arg,sizeof(msg)))
			return -EFAULT;

			tea5764_i2c_read(tea5764_rds_client,msg.buf,msg.len);
			break;
		}
	case RDS_I2C_WRITE:
		{
		if (copy_from_user(&msg,(struct i2c_message __user *)arg,sizeof(msg)))
			return -EFAULT;

			tea5764_i2c_send(tea5764_rds_client,msg.buf,msg.len);
			break;
		}
	case RDS_CLEAR_BUFFER:
			return -EFAULT;
#if 0
		{
			spin_lock_irqsave(&tea5764_lock, flags);
	
				rdsin = 0;
				rdsout = 0;
	
			spin_unlock_irqrestore(&tea5764_lock, flags);

			break;
		}
#endif
	case RDS_GET_LEVEL:
		{
			if ((void *)arg == NULL) {
				return -EFAULT;
			}
			if (copy_from_user(&msg,(struct i2c_message __user *)arg,sizeof(msg))) {
				return -EFAULT;
			}
			ret = TMC_RDSGetLevel(NULL,0,msg.buf,4,&dwActualOut);
			DEBMSG_IOCTL(printk(KERN_INFO "tmc ioctl RDS_GET_LEVEL %d\n",(int)*(msg.buf));)
			break;
		}
	case RDS_GET_FREQ:
		{
			if ((void *)arg == NULL) {
				return -EFAULT;
			}
			if (copy_from_user(&msg,(struct i2c_message __user *)arg,sizeof(msg))) {
				return -EFAULT;
			}
			iFreqency = Ctea5764Driver_getFreq();
			DEBMSG_IOCTL(printk(KERN_INFO "tmc ioctl 10 Ctea5764Driver_getFreq end %d\n",iFreqency);)
			*((long *)(msg.buf)) = iFreqency;
			if (iFreqency == 0) {
				return -ENOTTY;
			}
			break;
		}
	case RDS_TUNE:
		{
			if ((void *)arg == NULL) {
				return -EFAULT;
			}
			if (copy_from_user(&msg,(struct i2c_message __user *)arg,sizeof(msg))) {
				return -EFAULT;
			}
			iFreqency = *(int *)(msg.buf);
			DEBMSG_IOCTL(printk(KERN_INFO "tmc ioctl RDS_TUNE1 %d\n",iFreqency);)
			printk(KERN_INFO "tmc ioctl RDS_TUNE %d\n",iFreqency);
			if(TMC_RDSPreset((char *)&iFreqency,(unsigned long)4,
								 (char *)0,(unsigned long )0,(unsigned long *)&dwActualOut) == false) {
				return -ENOTTY;
			}
			printk(KERN_INFO "tmc ioctl RDS_TUNE2 %d\n",iFreqency);
			break;
		}
	case RDS_GET_ALLREGREAD:
		{
			if ((void *)arg == NULL) {
				return -EFAULT;
			}
			if (copy_from_user(&msg,(struct i2c_message __user *)arg,sizeof(msg))) {
				return -EFAULT;
			}
			DEBMSG_IOCTL(printk(KERN_INFO "tmc ioctl RDS_GET_ALLREGREAD %d\n",msg.len);)
				Ctea5764Driver_get_allreg(msg.buf,msg.len,&dwActualOut);
			break;
		}
	case RDS_SEEKFREQENCY:
		{
			if ((void *)arg == NULL) {
				return -EFAULT;
			}
			if (copy_from_user(&msg,(struct i2c_message __user *)arg,sizeof(msg))){
				return -EFAULT;
			}
			DEBMSG_IOCTL(printk(KERN_INFO "tmc ioctl RDS_SEEKFREQENCY %d\n",msg.len);)
			seekin = (TMC_RDS_SEEKIN *)(msg.buf);
			outlen= sizeof(TMC_RDS_SEEKOUT); 
			ret = TMC_RDSSeekFreq((char *)seekin,(unsigned long )2,(char *)&SeekOut,
								  		(unsigned long )outlen,(unsigned long *)&outexeclen);
			DEBMSG_IOCTL(printk(KERN_INFO "tmc ioctl RDS_SEEKFREQENCY2 %d\n",msg.len);)
			if (ret == true) {
				seekin->iResult = SeekOut.iResult;
				seekin->dwBand = SeekOut.dwBand;
			} else {
				seekin->iResult = TMC_RDS_SEEK_NG;
				seekin->dwBand = 0;
				return -ENOTTY;
			}
			break;
		}
	case RDS_READCANCEL:
		break;
	case RDS_READCDATA:
		{
			if ((void *)arg == NULL) {
				return -EFAULT;
			}
			if (copy_from_user(&msg,(struct i2c_message __user *)arg,sizeof(msg))) {
				return -EFAULT;
			}
			DEBMSG_IOCTL(printk(KERN_INFO "tmc ioctl RDS_READDATA %d\n",msg.len);)
			seekin = (TMC_RDS_SEEKIN *)(msg.buf);
			outlen= sizeof(TMC_RDS_SEEKOUT); 
			if(TMC_RDSReadData(NULL,0,msg.buf,msg.len,(unsigned long *)&outexeclen) == false) {
			DEBMSG_IOCTL(printk(KERN_INFO "tmc ioctl RDS_READDATA end1 %d\n",msg.len);)
				return -ENOTTY;
			}
			DEBMSG_IOCTL(printk(KERN_INFO "tmc ioctl RDS_READDATA end2 %d\n",msg.len);)
			break;
		}
	case RDS_ALLCANCEL:
		TMC_RDSAllCancel();
		break;
	case RDS_NAVIATT:
		{
			if ((void *)arg == NULL) {
				return -EFAULT;
			}
			if (copy_from_user(&msg,(struct i2c_message __user *)arg,sizeof(msg))) {
				return -EFAULT;
			}
			DEBMSG_IOCTL(printk(KERN_INFO "tmc ioctl RDS_NAVIATT1 %d\n",msg.buf[0]);)
			if(TMC_RDSNaviAtt(msg.buf,msg.len,NULL,0,(unsigned long *)&outexeclen) == false) {
				return -ENOTTY;
			}
			break;
		}
	case RDS_NAVIATT_STS:
		{
			if ((void *)arg == NULL) {
				return -EFAULT;
			}
			DEBMSG_IOCTL(printk(KERN_INFO "tmc ioctl RDS_NAVIATT_STS %d\n",msg.len);)
			if (copy_from_user(&msg,(struct i2c_message __user *)arg,sizeof(msg))) {
				return -EFAULT;
			}
			TMC_RDSGetAttSts(NULL,0,msg.buf,msg.len,(unsigned long *)&outexeclen);
			break;
		}
	case RDS_TMC_INIT:
		{
			TMC_RDSInit();
			break;
		}
	default:
		return -EINVAL;
	}

	
	return 0;
}

static struct file_operations tea5764_fops = {
      .read    = tea5764_read,
      .ioctl   = tea5764_ioctl,
      .open    = tea5764_open,
      .release = tea5764_release,
};

/*-----------------------------------------------------------------------------------------------*/

/* tea5764_init(void) */


/*-----------------------------------------------------------------------------------------------*/
#define DEBMSG_INIT(a)
static int __init tea5764_init(void)
{

	navuGpioSetLevel(NAVU_OUT_RDS_POWER, 1);

	i2c_add_driver (&tea5764_fm_driver);
	i2c_add_driver (&tea5764_rds_driver);
	register_chrdev(XYZ3_MAJOR_TMC, TEA5764_DRV_NAME, &tea5764_fops);
	pte5764_FormPtr = NULL;
	pte5764_ListenerPtr = NULL;

	tea5764_intx_th = NULL;
	tea5764_Cradle_th = NULL;

	navuGpioSetLevel(NAVU_OUT_RDS_POWER, 0);
	mdelay(20);
	navuGpioSetLevel(NAVU_OUT_RDS_POWER, 1);
	mdelay(100);
#if 0
	tea5764_intx_th = kthread_run(tea5764_IntThread,(void *)&datadata,"tea5764_int");
	if(IS_ERR(tea5764_intx_th)) {
		printk(KERN_ERR "tmc make kthread error\n");
		return -EBUSY;
	}

	tea5764_Cradle_th = kthread_run(NavuCradleThread,(void *)&datadata,"tea5764Cradle_int");
	if(IS_ERR(tea5764_Cradle_th)) {
		printk(KERN_ERR "tmc make kthread cradle error\n");
		return -EBUSY;
	}
#endif
	/* 񻺤ͭ */
	pte5764_FormPtr = kmalloc(sizeof(II2CForm),GFP_KERNEL);
	if (pte5764_FormPtr == NULL) {
		return -EBUSY;
	}
	pte5764_FormPtr->i2c_send = I2CFormLinux_i2c_send;
	pte5764_FormPtr->i2c_recv = I2CFormLinux_i2c_recv;
	pte5764_FormPtr->i2c_init = I2CFormLinux_init;
	pte5764_FormPtr->i2c_PwSend = I2CFormLinux_i2c_send;
	pte5764_FormPtr->i2c_PwUp = I2CFormLinux_PowerUp;
	pte5764_FormPtr->i2c_PwDown = I2CFormLinux_PowerDown;

	pte5764_ListenerPtr = kmalloc(sizeof(IRDSListener),GFP_KERNEL);
	if (pte5764_ListenerPtr == NULL) {
		kfree(pte5764_FormPtr);
		pte5764_FormPtr = NULL;
		return -EBUSY;
	}
	pte5764_ListenerPtr->update_rds_data = tea5764_update_rds_data;

	TMC_ClearRingCount();
	if (navuGpioGetLevel(NAVU_IN_ON_CRADLE) == 0) {
		DEBMSG_INIT(printk(KERN_INFO "NAVU_IN_ON_CRADLE 0\n");)
		CradleFlag = TMC_CRADLE_ON;
		//ǥХν
		Ctea5764Driver_init(pte5764_ListenerPtr,pte5764_FormPtr);
	} else{
		DEBMSG_INIT(printk(KERN_INFO "NAVU_IN_ON_CRADLE 1\n");)
		CradleFlag = TMC_CRADLE_OFF;
		/* ݻƤ쥸Ͻ */
		Ctea5764Driver_MakeInitData();
	}
	return 0;
}

/*-----------------------------------------------------------------------------------------------*/

/* tea5764_exit(void) */


/*-----------------------------------------------------------------------------------------------*/
static void __exit tea5764_exit(void)
{
	i2c_del_driver (&tea5764_fm_driver);
	i2c_del_driver (&tea5764_rds_driver);
	unregister_chrdev(XYZ3_MAJOR_TMC, TEA5764_DRV_NAME);

	navuGpioSetLevel(NAVU_OUT_RDS_POWER, 0);
	if (pte5764_FormPtr != NULL) {
		kfree(pte5764_FormPtr);
		pte5764_FormPtr = NULL;
	}
	if (pte5764_ListenerPtr != NULL) {
		kfree(pte5764_ListenerPtr);
		pte5764_ListenerPtr = NULL;
	}

}
/*-----------------------------------------------------------------------------------------------*/

/* tea5764_suspend(struct device *dev, pm_message_t state) */


/*-----------------------------------------------------------------------------------------------*/
static int tea5764_suspend(struct device *dev, pm_message_t state)
{
//	printk( KERN_ERR "%s:&dev=0x%lx, state=0x%lx\n", __FUNCTION__, (ulong)dev, state );

//	printk( KERN_ERR "tea5764_suspend: called.\n");
	return 0;
}

/*-----------------------------------------------------------------------------------------------*/

/* tea5764_resume(struct device *dev) */


/*-----------------------------------------------------------------------------------------------*/
static int tea5764_resume(struct device *dev)
{
	printk( KERN_INFO "tea5764_resume: called.\n");
	/* ǥХŸƤ */
//	navuGpioSetLevel(NAVU_OUT_RDS_POWER, 1);
	/* 񻺤ͭ */
//	pte5764_FormPtr = kmalloc(sizeof(II2CForm),GFP_KERNEL);
#if 0
	if (pte5764_FormPtr == NULL) {
		pte5764_FormPtr->i2c_send = I2CFormLinux_i2c_send;
		pte5764_FormPtr->i2c_recv = I2CFormLinux_i2c_recv;
		pte5764_FormPtr->i2c_init = I2CFormLinux_init;
		pte5764_FormPtr->i2c_PwSend = I2CFormLinux_i2c_send;
		pte5764_FormPtr->i2c_PwUp = I2CFormLinux_PowerUp;
		pte5764_FormPtr->i2c_PwDown = I2CFormLinux_PowerDown;
	}
	if (pte5764_ListenerPtr == NULL) {
		pte5764_ListenerPtr = kmalloc(sizeof(IRDSListener),GFP_KERNEL);
		pte5764_ListenerPtr->update_rds_data = tea5764_update_rds_data;
	}
#endif
	if (pte5764_FormPtr != NULL) {
		if (navuGpioGetLevel(NAVU_IN_ON_CRADLE) == 0) {
			Ctea5764Driver_PowerUpInit();
		}
	}
	return 0;
}


module_init(tea5764_init);
module_exit(tea5764_exit);

MODULE_AUTHOR("Michael Hennerich <michael.hennerich@analog.com>");
MODULE_DESCRIPTION("Driver for TEA5764");
MODULE_LICENSE("GPL");

#define DEBMSG_I2CSEND(a)
/* ------------------------------add by T.oosumimoto------------------------------------------------- */
static int I2CFormLinux_i2c_send(unsigned char bSlaveAddress,char *pBuffer,unsigned long size,long *pExecWriteSize)
{
	int req;
	struct i2c_message tmc_message;
	tmc_message.buf = pBuffer;
	tmc_message.len = size;
	DEBMSG_I2CSEND(printk(KERN_INFO "I2CFormLinux_i2c_send 0x%X %d\n",bSlaveAddress,(int)size);)
	if (bSlaveAddress == TEA5764_TUNE_FM_WRITEMODE) {
		req = FM_I2C_WRITE;
		if(tea5764_i2c_send(tea5764_fm_client,tmc_message.buf,tmc_message.len) < 0) {
			*pExecWriteSize = 0;
			return false;
		} else {
			*pExecWriteSize = tmc_message.len;
		}
		DEBMSG_I2CSEND(printk(KERN_INFO "I2CFormLinux_i2c_send FM_I2C_SEND %d\n",tmc_message.len);)
	} else if (bSlaveAddress == TEA5764_TUNE_RDS_WRITEMODE) {
		req = RDS_I2C_WRITE;
		if(tea5764_i2c_send(tea5764_rds_client,tmc_message.buf,tmc_message.len) < 0) {
			*pExecWriteSize = 0;
			return false;
		} else {
			*pExecWriteSize = tmc_message.len;
		}
		DEBMSG_I2CSEND(printk(KERN_INFO "I2CFormLinux_i2c_send RDS_I2C_SEND %d\n",tmc_message.len);)
	} else {
		return false;
	}
	return true;
}

#define DEBMSG_I2CRECV(a)
static int I2CFormLinux_i2c_recv(unsigned char bSlaveAddress,char *pBuffer,unsigned long size,long *pExecReadSize)
{
	int req;
	struct i2c_message tmc_message;
	 DEBMSG_I2CRECV(printk(KERN_EMERG "I2CFormLinux_i2c_recv 0x%X %d 0x%X\n",bSlaveAddress,(int)size,pBuffer);)
	tmc_message.buf = pBuffer;
	tmc_message.len = size;
	if (bSlaveAddress == TEA5764_TUNE_FM_READMODE) {
		req = FM_I2C_READ;
		if(tea5764_i2c_read(tea5764_fm_client,tmc_message.buf,tmc_message.len) < 0) {
			*pExecReadSize = 0;
			return false;
		} else {
			*pExecReadSize = tmc_message.len;
		}
		 DEBMSG_I2CRECV(printk(KERN_EMERG "I2CFormLinux_i2c_recv FM_I2C_RECV %d\n",tmc_message.len);)
	} else if (bSlaveAddress == TEA5764_TUNE_RDS_READMODE) {
		req = RDS_I2C_READ;
		if(tea5764_i2c_read(tea5764_rds_client,tmc_message.buf,tmc_message.len) < 0) {
			*pExecReadSize = 0;
			return false;
		} else {
			*pExecReadSize = tmc_message.len;
		}
		 DEBMSG_I2CRECV(printk(KERN_EMERG "I2CFormLinux_i2c_recv RDS_I2C_RECV %d\n",tmc_message.len);)
	} else {
		return false;
	}
	return true;
}
#if 0
static int I2CFormLinux_i2c_sendPowerUp(unsigned char bSlaveAddress,char *pBuffer,unsigned long size,long *pExecReadSize)
{
	return 0;
}
static int I2CFormLinux_Deinit(void)
{
	return true;
}
static int I2CFormLinux_Loadinit(void)
{
	return true;
}
static int I2CFormLinux_DeviceRestFlag(void)
{
	return true;
}
#endif
static int I2CFormLinux_init(void)
{
	return true;
}
/* ʲϻȤʤΤǤϤʤ */
static void I2CFormLinux_PowerUp(void)
{
}
static void I2CFormLinux_PowerDown(void)
{
}

#define DEBMSG_RDS_UP(a)
#define DEBMSG_RDS_UP2(a)
#define DEBMSG_RDS_UP_DATA(a)

static void tea5764_update_rds_data(int Event,unsigned char *pBuffer,int size){
//	char oldwritecnt = 0;
//	char newwritecnt = 0;

	static int blk = -1;									// Υ֥åֹ
	int blid = -1;									// Υ֥åֹ
	char data[2];

	DEBMSG_RDS_UP(printk(KERN_EMERG "Litenner in %d\n",Event);)
	switch(Event){
	case TMC_RDS_DATA_UPDATE:
	DEBMSG_RDS_UP2(printk(KERN_EMERG "Litenner TMC_RDS_DATA_UPDATE\n");)
#if 0
		//		WaitForSingleObject(hRDSReadDataMutex, INFINITE);
		//ReadɤĤɤå
		newwritecnt = RdsBufferWriteCnt +1;
		if (newwritecnt >= RDS_READ_BUFFER_MAX) {
			newwritecnt = 0;
		}
		if (newwritecnt == RdsBufferReadCnt) {
			//ɤĤäΤǥߥǡͤù
			RDSReadData[(int)RdsBufferWriteCnt][0] = (unsigned char)RDS_DATA_FLAG_LSYN;
			RDSReadData[(int)RdsBufferWriteCnt][1] = (unsigned char)RDS_DATA_STAT1_NG;
			RDSReadData[(int)RdsBufferWriteCnt][2] = (unsigned char)RDS_DATA_STAT2_NG;
			RdsBufferReadCnt = RdsBufferWriteCnt;
			printk("R Over %d\r\n",RdsBufferReadCnt);
		} else {
			RDSReadData[(int)RdsBufferWriteCnt][0] = (unsigned char)pBuffer[0];
			memcpy(&RDSReadData[(int)RdsBufferWriteCnt][1],&(pBuffer[12]),4);
			DEBMSG_RDS_UP_DATA(printk(KERN_EMERG "copy data 0x%X 0x%X 0x%X 0x%X 0x%X\n",
										  RDSReadData[(int)RdsBufferWriteCnt][0],
										  RDSReadData[(int)RdsBufferWriteCnt][1],
										  RDSReadData[(int)RdsBufferWriteCnt][2],
										  RDSReadData[(int)RdsBufferWriteCnt][3],
										  RDSReadData[(int)RdsBufferWriteCnt][4]);)
		}
		oldwritecnt = RdsBufferWriteCnt;
		RdsBufferWriteCnt++;
		if (RdsBufferWriteCnt >= RDS_READ_BUFFER_MAX) {
			RdsBufferWriteCnt = 0;
		}
//			ReleaseMutex(hRDSReadDataMutex);
			tmc_dataread_flag = TMC_WAIT_RDS_UPDATE;
			wake_up_interruptible(&tmc_dataread_ack); 
//			SetEvent(ReadEventHandleList[RDS_READ_DAV]);
#else
/*
			printk(KERN_EMERG "readdata 0x%2X 0x%2X 0x%2X 0x%2X 0x%2X\n",
				pBuffer[0],pBuffer[12],pBuffer[13],pBuffer[14],pBuffer[15]);
*/
		TMC_ReadDataLock();
		if ((pBuffer[0] & 0x80) &&											 // DAVFLGͭ
				(pBuffer[12] & 0x03) < 1) {									 // 顼̵
			blid = (pBuffer[12] >> 4) & 0x03;							 // ֥åɣļ
//			memcpy(data, &(pBuffer[14]), 2);								 // ǡԡ
			data[0] = pBuffer[14];
			data[1] = pBuffer[15];

			blk++;																// Ԥ֥åֹ
//			memcpy(&(RDSReadData[(int)RdsBufferWriteCnt][0]) + (2 * blk), data, 2);		// ߤǥХåե˥ԡ
			RDSReadData[(int)RdsBufferWriteCnt][2 * blk] = data[0];
			RDSReadData[(int)RdsBufferWriteCnt][(2 * blk)+1] = data[1];
			DEBMSG_RDS_UP2(printk(KERN_EMERG "readdata 0x%2X%2X blk %d\r\n",data[0],data[1],blk);)
			if (blid != blk) {										// ֥åϢ³ʤ
				blk = -1;														// ϣ֥åܤ
			} else if (blk >= 3) {								// ֥å·ä
				DEBMSG_RDS_UP2(printk(KERN_EMERG "W countr %d\r\n",RdsBufferWriteCnt);)
				RdsBufferWriteCnt++;
				if (RdsBufferWriteCnt >= RDS_READ_BUFFER_MAX) {
					RdsBufferWriteCnt = 0;
				}
				/* ɤĤ줿ɤĴ٤ */
				if (RdsBufferWriteCnt == RdsBufferReadCnt) {
					RdsBufferReadCnt++;
					if (RdsBufferReadCnt >= RDS_READ_BUFFER_MAX) {
						RdsBufferReadCnt = 0;
					}
				}
				tmc_dataread_flag = TMC_WAIT_RDS_UPDATE;
				wake_up_interruptible(&tmc_dataread_ack); 
				blk = -1;														// ϣ֥åܤ
			}
		}
		TMC_ReadDataUnLock();
#endif
		break;
	case TMC_RDS_SEEK:
	DEBMSG_RDS_UP2(printk(KERN_EMERG "Litenner TMC_RDS_SEEK\n");)
		tmc_tune_flag = TMC_WAIT_SEEK_END;
		wake_up_interruptible(&tmc_tune_ack); 
//		SetEvent(SeekHandle[0]);
		break;
	case TMC_RDS_SEEKLIMIT:
	DEBMSG_RDS_UP2(printk(KERN_EMERG "Litenner TMC_RDS_SEEKLIMIT\n");)
		tmc_tune_flag = TMC_WAIT_SEEKLIMIT;
		wake_up_interruptible(&tmc_tune_ack); 
//		SetEvent(SeekHandle[1]);
		break;
	case TMC_RDS_PRESET_END:
	DEBMSG_RDS_UP2(printk(KERN_EMERG "Litenner TMC_RDS_PRESET_END\n");)
		tmc_preset_flag = TMC_WAIT_PRESET_END;
		wake_up_interruptible(&tmc_preset_ack); 
//		SetEvent(hRdsPresetHandle);
		break;
		
	}
	DEBMSG_RDS_UP(printk(KERN_EMERG "Litenner out\n");)

}


/*---------------------------------------------------------------------------------------
			TMC_RDSPreset
			ȿ
----------------------------------------------------------------------------------------*/
static int TMC_TuneModeCheck(void);
static int TmcWaitMode;
//DECLARE_MUTEX("TEA_MODE");
DECLARE_MUTEX(tea_Mode);
DECLARE_MUTEX(tea_Ring);

static void TMC_ModeLock(void)
{
//	spin_lock_bh(&tea5764_Modelock);
//	spin_lock_irq(&tea5764_Modelock);
//	DECLARE_MUTEX_LOCKED(tea_Mode);
	down_interruptible(&tea_Mode) ;
}
static void TMC_ModeUnLock(void)
{
//	spin_unlock_bh(&tea5764_Modelock);
//	spin_unlock_irq(&tea5764_Modelock);
	up(&tea_Mode);
//	DECLARE_MUTEX(tea_Mode);
}
//static unsigned long regiirqflag;
static void TMC_RegLock(unsigned long *regiirqflag)
{
//	spin_lock_bh(&tea5764_lock);
	spin_lock_irqsave(&tea5764_lock,*regiirqflag);
//	spin_lock_irq(&tea5764_lock);
}
static void TMC_RegUnLock(unsigned long *regiirqflag)
{
//	spin_unlock_bh(&tea5764_lock);
	spin_unlock_irqrestore(&tea5764_lock,*regiirqflag);
//	spin_unlock_irq(&tea5764_lock);
}
#define DEBMSG_TUNEMODE(a)
static int TMC_TuneModeCheck(void) {
	unsigned long flag;
	int WaitRet;
	if (TmcWaitMode == RDS_TMC_MODE_CANCEL_WAIT) {
		DEBMSG_TUNEMODE(printk(KERN_INFO "Cancel Wait Seqence NG\r\n");)
		return FALSE;
	} else if(TmcWaitMode == RDS_TMC_MODE_SEEK) {
		DEBMSG_TUNEMODE(printk(KERN_INFO "-- TMC_TM Seek Cancel start--\r\n");)
		if(TMC_CheckCradle()) {
			TMC_RegLock(&flag);
			Ctea5764Driver_SeekIntDisable();
			TMC_RegUnLock(&flag);
		}
		TmcWaitMode = RDS_TMC_MODE_CANCEL_WAIT;
//		SetEvent(SeekHandle[2]);
		tmc_tune_flag = TMC_WAIT_SEEK_CANCEL;
		wake_up_interruptible(&tmc_tune_ack); 
//		WaitForSingleObject(hRDSCancel, RDS_CANCEL_TIMEOUT);
		WaitRet = wait_event_interruptible_timeout(tmc_exec_cancel,
								tmc_cancel_flag == 1, msecs_to_jiffies(RDS_CANCEL_TIMEOUT));
		tmc_cancel_flag = 0;
		TMC_ChangeMode(RDS_TMC_MODE_IDLE);
		DEBMSG_TUNEMODE(printk(KERN_INFO "-- TMC_TM Seek Cancel end--\r\n");)
		return TRUE;
	} else if (TmcWaitMode == RDS_TMC_MODE_READ) {
		DEBMSG_TUNEMODE(printk(KERN_INFO "-- TMC_TM2 Read Cancel start--\r\n");)
		if(TMC_CheckCradle()) {
			TMC_RegLock(&flag);
			Ctea5764Driver_DavIntCtrl(FALSE);
			TMC_RegUnLock(&flag);

			//RingΥ󥿤򥯥ꥢ
			TMC_ClearRingCount();
		}
		TmcWaitMode = RDS_TMC_MODE_CANCEL_WAIT;
//		SetEvent(ReadEventHandleList[RDS_READ_CANCEL]);
//		WaitForSingleObject(hRDSCancel, RDS_CANCEL_TIMEOUT);
		tmc_dataread_flag = TMC_WAIT_RDS_DATAREAD_CANCEL;
		wake_up_interruptible(&tmc_dataread_ack); 
		WaitRet = wait_event_interruptible_timeout(tmc_exec_cancel,
								tmc_cancel_flag == 1, msecs_to_jiffies(RDS_CANCEL_TIMEOUT));
		tmc_cancel_flag = 0;
		TMC_ChangeMode(RDS_TMC_MODE_IDLE);
		DEBMSG_TUNEMODE(printk(KERN_INFO "-- TMC_TM2 Read Cancel end--\r\n");)
		return TRUE;
	}
	return TRUE;
}

static void TMC_ReadDataLock(void)
{
	down_interruptible(&tea_Ring) ;
//	spin_lock_bh(&tea5764_ReadDatalock);
}
static void TMC_ReadDataUnLock(void)
{
	up(&tea_Ring);
//	spin_unlock_bh(&tea5764_ReadDatalock);
}
/*---------------------------------------------------------------------------------------
			TMC_ClearRingCount
----------------------------------------------------------------------------------------*/
static void TMC_ClearRingCount(void)
{
//	WaitForSingleObject(hRDSReadDataMutex, INFINITE);
	TMC_ReadDataLock();
	RdsBufferReadCnt = 0;
	RdsBufferWriteCnt = 0;
	TMC_ReadDataUnLock();
//	ReleaseMutex(hRDSReadDataMutex);
}

#define DEBMSG_PRESET(a)
static BOOL TMC_RDSPreset(char * pBufIn,unsigned long dwLenIn,
						  char * pBufOut,unsigned long dwLenOut,unsigned long *pdwActualOut)
{
	unsigned long flag;
	int iFreqency = 0;
	int ret = 0;
	DEBMSG_PRESET(printk(KERN_EMERG "------- RDS_TUNE 1-------\r\n");)
	if (!TMC_TuneModeCheck()) {
		printk("Not Exec  Preset CancelWaitting\n");
		return TRUE;
	}
	TMC_RegLock(&flag);
	if (!Ctea5764Driver_DavIntCtrl(FALSE)) {
		TMC_RegUnLock(&flag);
		return FALSE;
//		TMC_ErrorInitDevice();
	}
	TMC_RegUnLock(&flag);

	//RingΥ󥿤򥯥ꥢ
	TMC_ClearRingCount();

	if (pBufIn != NULL){
		iFreqency = *(unsigned long *)pBufIn;
		TMC_RegLock(&flag);
		if (!Ctea5764Driver_tune(iFreqency)) {
//			TMC_ErrorInitDevice();
			TMC_RegUnLock(&flag);
			return FALSE;
		}

		TMC_RegUnLock(&flag);
//		WaitForSingleObject(hRdsPresetHandle,500);
		ret = wait_event_interruptible_timeout(tmc_preset_ack,
										tmc_preset_flag == TMC_WAIT_PRESET_END, msecs_to_jiffies(1000));
		if (ret == 0) {
			Ctea5764Driver_SetFRMaskOff();
		}
		tmc_preset_flag = TMC_WAIT_INIT;
		DEBMSG_PRESET(printk(KERN_EMERG "------- TMC_IOControl TUNE wait ret %d %d-------\r\n",ret,tmc_tune_flag);)
	} else {
//		*pdwActualOut = 0;
		DEBMSG_PRESET(printk(KERN_EMERG "------- TMC_IOControl End-------\r\n");)
		return FALSE;
	}
	mdelay(15);
	DEBMSG_PRESET(printk(KERN_EMERG "------- TMC_IOControl RDS_TUNE end-------\r\n");)
	return TRUE;
}


#define DEB_MSG_CHMODE(a)
static void TMC_ChangeMode(int iMode)
{
	TmcWaitMode = iMode;
	DEB_MSG_CHMODE(printk(KERN_EMERG "TMC_ChangeMode %d\n",TmcWaitMode);)
}
static bool TMC_CheckMode(int iMode)
{
	if (TmcWaitMode == iMode) {
		return true;
	} else {
		return false;
	}
}

static void SetCancelEvent(void);
static void SetCancelEvent(void)
{
	tmc_cancel_flag = 1;
	wake_up_interruptible(&tmc_exec_cancel); 
//	printk(KERN_EMERG "SetCancelEvent\n");
}
#define DEB_SEEK_MSG(a)
/*---------------------------------------------------------------------------------------
			TMC_RDSSeekFreq
			ȿ
----------------------------------------------------------------------------------------*/
static BOOL TMC_RDSSeekFreq(char * pBufIn,unsigned long dwLenIn,
							char * pBufOut,unsigned long dwLenOut,unsigned long *pdwActualOut)
{
	unsigned long flag;
	long SeekRet;
	TMC_RDS_SEEKIN *seekin = NULL;
	TMC_RDS_SEEKOUT *SeekOut = NULL;
	int iFreqency = 0;
	int iStartFreqency = 0;
	int SeekDir = TMC_RDS_SEEK_UP;
	char bLimitCnt = 0;
	unsigned char bLevelLimit = TMC_RDS_SEARCH_LEVEL_LIMIT5;

	if (!TMC_TuneModeCheck()) {
		DEB_SEEK_MSG(printk(KERN_EMERG "No Exec Seek CANCEL WAITING\r\n");)
		return TRUE;
	}
//---------------Seek֤ǥå----------------------
	TMC_ModeLock();
	//֥å

	TMC_RegLock(&flag);

	Ctea5764Driver_DavIntCtrl(FALSE);

	DEB_SEEK_MSG(printk(KERN_EMERG "------- TMC_IOControl RDS_SEEKFREQENCY-------\r\n");)
	if (pBufIn != NULL && pBufOut != NULL && pdwActualOut != NULL) {
		seekin = (TMC_RDS_SEEKIN*)pBufIn;
		if (seekin->bSeekFlag == TMC_RDS_SEARCH_UP) {
			SeekDir = TMC_RDS_SEEK_UP;
		} else {
			SeekDir = TMC_RDS_SEEK_DOWN;
		}
		switch(seekin->bLevelLimit) {
		case TMC_RDS_SEARCH_LEVEL_LIMIT3:
			bLevelLimit = RDS_SEARCH_LEVEL_LIMIT3;
			break;
		case TMC_RDS_SEARCH_LEVEL_LIMIT5:
			bLevelLimit = RDS_SEARCH_LEVEL_LIMIT5;
			break;
		case TMC_RDS_SEARCH_LEVEL_LIMIT7:
			bLevelLimit = RDS_SEARCH_LEVEL_LIMIT7;
			break;
		case TMC_RDS_SEARCH_LEVEL_LIMIT10:
			bLevelLimit = RDS_SEARCH_LEVEL_LIMIT10;
			break;
		default:
			bLevelLimit = RDS_SEARCH_LEVEL_LIMIT5;
			break;
		}
		//ȤȤμȿФƤ
		iStartFreqency = Ctea5764Driver_getFreq();
		if (iStartFreqency == -1) {
			TMC_RegUnLock(&flag);
			TMC_ChangeMode(RDS_TMC_MODE_IDLE);
			TMC_ModeUnLock();
			DEB_SEEK_MSG(printk(KERN_EMERG "Seek GetFreq Error1\r\n");)
			return FALSE;
		}

		while(bLimitCnt < RDS_SEEKLOOP_MAX){
			Ctea5764Driver_Seek(SeekDir,bLevelLimit);
			TMC_ChangeMode(RDS_TMC_MODE_SEEK);
			TMC_RegUnLock(&flag);

//--Ǿ֤Υå-----------------------
			TMC_ModeUnLock();
RE_WAIT_SEEK:			
			DEB_SEEK_MSG(printk(KERN_EMERG "RDS_SEEKFREQENCY step1\r\n");)
//			SeekRet = WaitForMultipleObjects(3,SeekHandle,FALSE,TMC_SEEK_TIMEOUT);
			SeekRet = wait_event_interruptible_timeout(tmc_tune_ack,
							tmc_tune_flag == TMC_WAIT_SEEK_END ||
							tmc_tune_flag == TMC_WAIT_SEEKLIMIT  ||
							tmc_tune_flag == TMC_WAIT_SEEK_CANCEL , msecs_to_jiffies(TMC_SEEK_TIMEOUT));
			if (SeekRet == 0) {
				DEB_SEEK_MSG(printk(KERN_EMERG "RDS_SEEKFREQENCY wait ret TIMEOUT %d\n",(int)SeekRet);)
				SeekRet = WAIT_TIMEOUT;
			} else if(SeekRet == -ERESTARTSYS){
				DEB_SEEK_MSG(printk(KERN_EMERG "RDS_SEEKFREQENCY wait ret ERESTARTSYS %d\n",(int)SeekRet);)
				SeekRet = 10;
			} else {
				DEB_SEEK_MSG(printk(KERN_EMERG "RDS_SEEKFREQENCY wait ret %d flag %d\n",(int)SeekRet,tmc_tune_flag);)
				SeekRet = tmc_tune_flag + WAIT_OBJECT_0 -1;
			}
			tmc_tune_flag = 0;
//			RETAILMSG(1, (TEXT("SeekW %X\r\n"),SeekRet));
//--Seekο֤ͤå
			TMC_ModeLock();
			TMC_RegLock(&flag);
			//ָʤΤǤǥ󥻥ԵɤåƤޤ
			if (((SeekRet - WAIT_OBJECT_0) == RDS_SEEK_TYPE_SEEKERROR)
				|| (TMC_CheckMode(RDS_TMC_MODE_CANCEL_WAIT))){
				//
				if (TMC_CheckMode(RDS_TMC_MODE_CANCEL_WAIT)) {
					DEB_SEEK_MSG(printk(KERN_EMERG " Seek Cancel\r\n");)
					Ctea5764Driver_SeekIntDisable();
					//ȿꤹ
					Ctea5764Driver_tune(iStartFreqency);
					TMC_RegUnLock(&flag);

//					WaitForSingleObject(hRdsPresetHandle,500);
					wait_event_interruptible_timeout(tmc_preset_ack,
														 tmc_preset_flag == TMC_WAIT_PRESET_END,msecs_to_jiffies(500));
					tmc_preset_flag = TMC_WAIT_INIT;
					TMC_RegLock(&flag);
					if (pBufOut != NULL){
						SeekOut = (TMC_RDS_SEEKOUT *)pBufOut;
						SeekOut->iResult = TMC_RDS_SEEK_NG;
						SeekOut->dwBand = (long)(iStartFreqency);
						*pdwActualOut =sizeof(TMC_RDS_SEEKOUT);
						DEB_SEEK_MSG(printk(KERN_EMERG "--- RDS_SEEKFREQENCY cancel--- %d\n",iStartFreqency);)
					}
					TMC_RegUnLock(&flag);
					//RingΥ󥿤򥯥ꥢ
					TMC_ClearRingCount();
//					SetEvent(hRDSCancel);
					SetCancelEvent();

//--Seek֤ȴ
					TMC_ModeUnLock();
//					ResetEvent(SeekHandle[0]);
//					ResetEvent(SeekHandle[1]);
					return TRUE;
				} else {
					//ְ㤤ǥʥˤʤäΤǤȤꤢ³׸Ƥ
					DEB_SEEK_MSG(printk(KERN_EMERG "continue? %d %d\n",(int)(SeekRet - WAIT_OBJECT_0),(int)TmcWaitMode);)
					TMC_RegUnLock(&flag);
					TMC_ModeUnLock();
					goto RE_WAIT_SEEK;
//					continue;
				}
			}
			//ƬǾ֤åƤΤǡʲϥ󥻥Եˤʤʤ
			//ȿĤޤ
			if ((SeekRet - WAIT_OBJECT_0) == RDS_SEEK_TYPE_SEEKEND) {
				//Ĥä
				iFreqency = Ctea5764Driver_getFreq();
				if (iFreqency == -1) {
					TMC_RegUnLock(&flag);
					TMC_ChangeMode(RDS_TMC_MODE_IDLE);
					TMC_ModeUnLock();
					tmc_tune_flag = TMC_WAIT_INIT;
					DEB_SEEK_MSG(printk(KERN_EMERG "Seek GetFreq Error2\r\n");)
					return FALSE;
				}
				//ȿꤹ
				Ctea5764Driver_tune(iFreqency);
#ifdef TEST_FREQ
				Ctea5764Driver_tune(-1);
				iFreqency = Ctea5764Driver_getFreq();
				DEB_SEEK_MSG(printk(KERN_EMERG "Seek Freq -1 %d\r\n",iFreqency);)				
#endif
				TMC_RegUnLock(&flag);
//				WaitForSingleObject(hRdsPresetHandle,500);
				wait_event_interruptible_timeout(tmc_preset_ack,
										tmc_preset_flag == TMC_WAIT_PRESET_END, msecs_to_jiffies(500));
				tmc_preset_flag = TMC_WAIT_INIT;
				if (pBufOut != NULL){
					SeekOut = (TMC_RDS_SEEKOUT *)pBufOut;
					SeekOut->iResult = TMC_RDS_SEEK_OK;
					SeekOut->dwBand = (long)(iFreqency);
					*pdwActualOut =sizeof(TMC_RDS_SEEKOUT);
				}
				//RingΥ󥿤򥯥ꥢ
				TMC_ClearRingCount();

//				RETAILMSG(1, (TEXT("FREQENCY Hit %d\r\n"),iFreqency));
				DEB_SEEK_MSG(printk(KERN_EMERG "--- RDS_SEEKFREQENCY Hit %d ---\r\n",iFreqency);)
				TMC_ChangeMode(RDS_TMC_MODE_IDLE);
//--Seek֤ȴ
				TMC_ModeUnLock();
				return TRUE;
			//ȿߥåȤˤʤޤ
			} else if ((SeekRet - WAIT_OBJECT_0) == RDS_SEEK_TYPE_LIMITEND) {
				//ߥåȤãʥ顼̤ˤΤ
#ifndef SEEK_LIMIT_END 
				iFreqency = Ctea5764Driver_getFreq();
				if (iFreqency == -1) {
					TMC_RegUnLock(&flag);
					TMC_ChangeMode(RDS_TMC_MODE_IDLE);
					TMC_ModeUnLock();
					DEB_SEEK_MSG(printk(KERN_EMERG "-- Seek Limit error %d --\r\n",iFreqency);)
					tmc_preset_flag = TMC_WAIT_INIT;
					return FALSE;
				}
				Ctea5764Driver_tune(iFreqency);
				TMC_RegUnLock(&flag);
//				WaitForSingleObject(hRdsPresetHandle,500);
				wait_event_interruptible_timeout(tmc_preset_ack,
										tmc_preset_flag == TMC_WAIT_PRESET_END, msecs_to_jiffies(500));
				tmc_preset_flag = TMC_WAIT_INIT;
				TMC_RegLock(&flag);
				if (pBufOut != NULL){
					SeekOut = (TMC_RDS_SEEKOUT *)pBufOut;
					SeekOut->iResult = TMC_RDS_SEEK_LIMIT;
					SeekOut->dwBand = (long)(iFreqency);
					*pdwActualOut =sizeof(TMC_RDS_SEEKOUT);
				}
				TMC_ChangeMode(RDS_TMC_MODE_IDLE);
				TMC_RegUnLock(&flag);
				//RingΥ󥿤򥯥ꥢ
				TMC_ClearRingCount();
//--Seek֤ȴ
				TMC_ModeUnLock();
				DEB_SEEK_MSG(printk(KERN_EMERG "-- Seek Limit %d --\r\n",iFreqency);)
				return TRUE;
#else
				//󥻥Ե桩
				bLimitCnt++;
				if (bLimitCnt < RDS_SEEKLOOP_MAX) {
					iFreqency = Ctea5764Driver_getFreq();
					if (iFreqency == -1) {
						TMC_RegUnLock(&flag);
						TMC_ChangeMode(RDS_TMC_MODE_IDLE);
						TMC_ModeUnLock();
						return FALSE;
					}
					//ȿꤹ
					Ctea5764Driver_tune(iFreqency);
					TMC_RegUnLock(&flag);
//					WaitForSingleObject(hRdsPresetHandle,500);
					wait_event_interruptible_timeout(tmc_preset_ack,
												tmc_preset_flag == TMC_WAIT_PRESET_END, msecs_to_jiffies(500));
					tmc_preset_flag = TMC_WAIT_INIT;
					TMC_RegLock(&flag);
				} else {
					Ctea5764Driver_tune(iStartFreqency);
					TMC_RegUnLock(&flag);
//					WaitForSingleObject(hRdsPresetHandle,500);
					wait_event_interruptible_timeout(tmc_preset_ack,
												 tmc_preset_flag == TMC_WAIT_PRESET_END, msecs_to_jiffies(500));
					tmc_preset_flag = TMC_WAIT_INIT;
					TMC_RegLock(&flag);
					if (pBufOut != NULL){
						SeekOut = (TMC_RDS_SEEKOUT *)pBufOut;
						SeekOut->iResult = TMC_RDS_SEEK_LIMIT;
						SeekOut->dwBand = (long)(iStartFreqency);
						*pdwActualOut =sizeof(TMC_RDS_SEEKOUT);
					}
					DEB_SEEK_MSG(printk(KERN_EMERG "--- RDS_SEEKFREQENCY Limit end %d---\r\n",iStartFreqency);)
					TMC_ChangeMode(RDS_TMC_MODE_IDLE);
					TMC_RegUnLock(&flag);
					//RingΥ󥿤򥯥ꥢ
					TMC_ClearRingCount();
//--Seek֤ȴ
					TMC_ModeUnLock();
					return TRUE;
				}
#endif
			//ԤäƤɥॢ
			} else if ((SeekRet >= WAIT_ABANDONED_0) && (SeekRet <=(WAIT_ABANDONED_0 + 2))) {
					DEB_SEEK_MSG(printk(KERN_EMERG " Seek Exit\n");)
				Ctea5764Driver_SeekIntDisable();
				//ȿꤹ
				Ctea5764Driver_tune(iStartFreqency);
				TMC_RegUnLock(&flag);
//				WaitForSingleObject(hRdsPresetHandle,500);
				wait_event_interruptible_timeout(tmc_preset_ack,
													 tmc_preset_flag == TMC_WAIT_PRESET_END, msecs_to_jiffies(500));
				tmc_preset_flag = TMC_WAIT_INIT;
				TMC_RegLock(&flag);
				if (pBufOut != NULL){
					SeekOut = (TMC_RDS_SEEKOUT *)pBufOut;
					SeekOut->iResult = TMC_RDS_SEEK_NG;
					SeekOut->dwBand = (long)(iStartFreqency);
					*pdwActualOut =sizeof(TMC_RDS_SEEKOUT);
					DEB_SEEK_MSG(printk(KERN_EMERG "--- RDS_SEEKFREQENCY cancel--- %d\n",iStartFreqency);)
				}
				TMC_RegUnLock(&flag);
				//RingΥ󥿤򥯥ꥢ
				TMC_ClearRingCount();
//				ResetEvent(SeekHandle[2]);
//				ResetEvent(SeekHandle[0]);
//				ResetEvent(SeekHandle[1]);
//--Seek֤ȴ
				TMC_ModeUnLock();
				return TRUE;
			} else if (SeekRet == WAIT_TIMEOUT) {
				DEB_SEEK_MSG(printk(KERN_EMERG "-- Seek timeout --\r\n");)
				//ȿꤹ
				Ctea5764Driver_SeekIntDisable();
				//startȤμȿꤷʤ
				Ctea5764Driver_tune(iStartFreqency);
				TMC_RegUnLock(&flag);
//				WaitForSingleObject(hRdsPresetHandle,500);
				wait_event_interruptible_timeout(tmc_preset_ack,
													 tmc_preset_flag == TMC_WAIT_PRESET_END, msecs_to_jiffies(500));
				tmc_preset_flag = TMC_WAIT_INIT;
				TMC_RegLock(&flag);
				if (pBufOut != NULL){
					SeekOut = (TMC_RDS_SEEKOUT *)pBufOut;
					SeekOut->iResult = TMC_RDS_SEEK_TIMEOUT;
					SeekOut->dwBand = (long)(iStartFreqency);
					*pdwActualOut =sizeof(TMC_RDS_SEEKOUT);
					DEB_SEEK_MSG(printk(KERN_EMERG "--- RDS_SEEKFREQENCY cancel--- %d\n",iStartFreqency);)
				}
				TMC_RegUnLock(&flag);
				//RingΥ󥿤򥯥ꥢ
				TMC_ClearRingCount();
//				ResetEvent(SeekHandle[0]);
//				ResetEvent(SeekHandle[1]);
//--Seek֤ȴ
				TMC_ModeUnLock();
				return TRUE;
			//褯狼ʤἺԤǤ
			} else {
				DEB_SEEK_MSG(printk(KERN_EMERG " unknown signal\r\n");)
				Ctea5764Driver_SeekIntDisable();
				//ȿꤹ
				Ctea5764Driver_tune(iStartFreqency);
				TMC_RegUnLock(&flag);
//				WaitForSingleObject(hRdsPresetHandle,500);
				wait_event_interruptible_timeout(tmc_preset_ack,
													 tmc_preset_flag == TMC_WAIT_PRESET_END, msecs_to_jiffies(500));
				tmc_preset_flag = TMC_WAIT_INIT;
				TMC_RegLock(&flag);
				if (pBufOut != NULL){
					SeekOut = (TMC_RDS_SEEKOUT *)pBufOut;
					SeekOut->iResult = TMC_RDS_SEEK_NG;
					SeekOut->dwBand = (long)(iStartFreqency);
					*pdwActualOut =sizeof(TMC_RDS_SEEKOUT);
					DEB_SEEK_MSG(printk(KERN_EMERG "--- RDS_SEEKFREQENCY cancel--- %d\n",iStartFreqency);)
				}
				TMC_RegUnLock(&flag);
				//RingΥ󥿤򥯥ꥢ
				TMC_ClearRingCount();
//				ResetEvent(SeekHandle[2]);
//				ResetEvent(SeekHandle[0]);
//				ResetEvent(SeekHandle[1]);
//--Seek֤ȴ
				TMC_ModeUnLock();
				return TRUE;
			}
		}
	} else {
		*pdwActualOut = 0;
		DEB_SEEK_MSG(printk(KERN_EMERG "------- TMC_IOControl End-------\r\n");)
		TMC_RegUnLock(&flag);
//--Ǿ֤Υå-----------------------
		TMC_ModeUnLock();
		tmc_preset_flag = TMC_WAIT_INIT;
		return FALSE;
	}
	TMC_RegUnLock(&flag);
	tmc_preset_flag = TMC_WAIT_INIT;
	return TRUE;
}
#define DEB_LEVEL_MSG(a)
/*---------------------------------------------------------------------------------------
			TMC_RDSGetLevel
			ټ
----------------------------------------------------------------------------------------*/
static BOOL TMC_RDSGetLevel(char * pBufIn,unsigned long dwLenIn,
					 char * pBufOut,unsigned long dwLenOut,unsigned long *pdwActualOut)
{
	unsigned long flag;
	short level;
	DEB_LEVEL_MSG(printk(KERN_EMERG "------- RDS_GET_LEVEL-------\n");)
	if ((pBufOut == NULL) || (pdwActualOut == NULL)) {
		if (pdwActualOut != NULL) {
			*pdwActualOut = 0;
		}
		DEB_LEVEL_MSG(printk(KERN_EMERG "------- End-------\n");)
		return FALSE;
	}
	TMC_ModeLock();
	TMC_RegLock(&flag);
	if ((TmcWaitMode == RDS_TMC_MODE_CANCEL_WAIT) || (TmcWaitMode == RDS_TMC_MODE_SEEK)) {
		level = 0;
	} else {
		DEB_LEVEL_MSG(printk(KERN_EMERG "Ctea5764Driver_getLevel\n");)
		level = Ctea5764Driver_getLevel();
	}
	TMC_RegUnLock(&flag);
	TMC_ModeUnLock();
	if (level == -1) {
		*(long *)pBufOut = (long)(0);
		*pdwActualOut = 4;
		DEB_LEVEL_MSG(printk(KERN_EMERG "------- End-------\n");)
		return FALSE;
	}
	*(long *)pBufOut = (long)(level);
	*pdwActualOut = 4;
	return TRUE;
}

static bool TMC_CheckCradle(void)
{
	if (CradleFlag == TMC_CRADLE_ON) {
		return true;
	} else {
		return false;
	}
}
#define DEBMSG_RM(a)
#define RDS_CANCEL_TIMEOUT 1000
/*---------------------------------------------------------------------------------------
			TMC_CheckRingCnt
----------------------------------------------------------------------------------------*/
static BOOL TMC_CheckRingCnt(void);
static BOOL TMC_CheckRingCnt(void)
{
	BOOL ret = FALSE;
//	WaitForSingleObject(hRDSReadDataMutex, INFINITE);
	TMC_ReadDataLock();
	if(RdsBufferReadCnt == RdsBufferWriteCnt) {
		ret = TRUE;
	} else {
		ret = FALSE;
	}
//	ReleaseMutex(hRDSReadDataMutex);
	TMC_ReadDataUnLock();
	return ret;
}
/*---------------------------------------------------------------------------------------
			TMC_RDSCommonReadCancel
----------------------------------------------------------------------------------------*/
static BOOL TMC_RDSCommonReadCancel(void);
static BOOL TMC_RDSCommonReadCancel(void)
{
	unsigned long flag;
	TMC_RegLock(&flag);
	Ctea5764Driver_DavIntCtrl(FALSE);
	TMC_RegUnLock(&flag);
	//RingΥ󥿤򥯥ꥢ
	TMC_ClearRingCount();
//	SetEvent(hRDSCancel);
	SetCancelEvent();
	return TRUE;
}

/*---------------------------------------------------------------------------------------
			TMC_ReadModeCheck
----------------------------------------------------------------------------------------*/
static int TMC_ReadModeCheck(void)
{
	unsigned long flag;
	int WaitRet = 0;
	if (TmcWaitMode == RDS_TMC_MODE_CANCEL_WAIT) {
		DEBMSG_RM(printk(KERN_EMERG "Cancel Wait Seqence NG\r\n");)
		return FALSE;
	} else if(TmcWaitMode == RDS_TMC_MODE_SEEK) {
		DEBMSG_RM(printk(KERN_EMERG "-- TMC_RM Now Seek Mode--\r\n");)
#if 0
		DEBMSG_RM(printk(KERN_EMERG "-- TMC_RM Seek Cancel start--\r\n");)
		if(TMC_CheckCradle()) {
			TMC_RegLock(&flag);
			Ctea5764Driver_SeekIntDisable();
			TMC_RegUnLock(&flag);
		}
		TmcWaitMode = RDS_TMC_MODE_CANCEL_WAIT;
//		SetEvent(SeekHandle[2]);
		tmc_tune_flag = TMC_WAIT_SEEK_CANCEL;
		wake_up_interruptible(&tmc_tune_ack); 
//		WaitForSingleObject(hRDSCancel, RDS_CANCEL_TIMEOUT);
		WaitRet = wait_event_interruptible_timeout(tmc_exec_cancel,
								tmc_cancel_flag == 1, msecs_to_jiffies(RDS_CANCEL_TIMEOUT));
		tmc_cancel_flag = 0;
		TMC_ChangeMode(RDS_TMC_MODE_IDLE);
		DEBMSG_RM(printk(KERN_EMERG "-- TMC_RM Seek Cancel end--\r\n");)
#endif
		return FALSE;
	} else if (TmcWaitMode == RDS_TMC_MODE_READ) {
		DEBMSG_RM(printk(KERN_EMERG "-- TMC_RM Seek Cancel start--\r\n");)
		if(TMC_CheckCradle()) {
			TMC_RegLock(&flag);
			Ctea5764Driver_DavIntCtrl(FALSE);
			TMC_RegUnLock(&flag);

			//RingΥ󥿤򥯥ꥢ
			TMC_ClearRingCount();
		}
		TmcWaitMode = RDS_TMC_MODE_CANCEL_WAIT;
//		SetEvent(ReadEventHandleList[RDS_READ_CANCEL]);
		tmc_dataread_flag = TMC_WAIT_RDS_DATAREAD_CANCEL;
		wake_up_interruptible(&tmc_dataread_ack); 
		DEBMSG_RM(printk(KERN_EMERG "-- TMC_RM Read Cancel SetEvent--\r\n");)
//		WaitForSingleObject(hRDSCancel, RDS_CANCEL_TIMEOUT);
		WaitRet = wait_event_interruptible_timeout(tmc_exec_cancel,
								tmc_cancel_flag == 1, msecs_to_jiffies(RDS_CANCEL_TIMEOUT));
		tmc_cancel_flag = 0;
		TMC_ChangeMode(RDS_TMC_MODE_IDLE);
		DEBMSG_RM(printk(KERN_EMERG "-- TMC_RM Seek Cancel end--\r\n");)
		return TRUE;
	}
	TmcWaitMode = RDS_TMC_MODE_IDLE;
	return TRUE;
}

#define DEBMSG_DR(a)
/*---------------------------------------------------------------------------------------
			TMC_RDSReadData
			ǡ
----------------------------------------------------------------------------------------*/
static BOOL TMC_RDSReadData(unsigned char * pBufIn,unsigned long dwLenIn,unsigned char * pBufOut,unsigned long dwLenOut,unsigned long *pdwActualOut)
{
	unsigned long flag;
	unsigned long WaitRet;
	unsigned long writecnt = dwLenOut;
	DEBMSG_DR(printk(KERN_EMERG "+-\r\n");)
//	DEBMSG_TIME(RETAILMSG(1, (TEXT("read start %d\r\n"),GetTickCount()));)
//-----------ReadModeǥå-------------------------//
	if ((pBufOut == NULL) || (pdwActualOut == NULL)) {
		return FALSE;
	}
	TMC_ModeLock();
	//¿Ť˸ƤФ줿׵򥭥󥻥뤹
	if (!TMC_ReadModeCheck()) {
		TMC_RDSCommonReadCancel();
		TMC_MakeNGData(pBufOut,dwLenOut);
		DEBMSG_DR(printk(KERN_EMERG " -- DR CANCELWait --\r\n");)
		TMC_ModeUnLock();
		DEBMSG_DR(printk(KERN_EMERG " -- DR CANCELWait ret --\r\n");)
		return TRUE;
	}
	//֤READѹ
	TMC_ChangeMode(RDS_TMC_MODE_READ);

	//ߤĤ
	TMC_RegLock(&flag);
	Ctea5764Driver_DavIntCtrl(TRUE);
	TMC_RegUnLock(&flag);

	while(TMC_CheckRingCnt()) {
		DEBMSG_DR(printk(KERN_EMERG "-- dataCheckRoop --\n");)
		TMC_ModeUnLock();
//-------------------------ޤReadMode--------------------------------------

//		WaitRet = WaitForMultipleObjects(2,ReadEventHandleList,FALSE, TMC_INTWAIT);
		WaitRet = wait_event_interruptible_timeout(tmc_dataread_ack,
						tmc_dataread_flag == TMC_WAIT_RDS_UPDATE ||
						tmc_dataread_flag == TMC_WAIT_RDS_DATAREAD_CANCEL,
						msecs_to_jiffies(1000));
		DEBMSG_DR(printk(KERN_EMERG "data_read wait%d %d\r\n",(int)WaitRet,(int)tmc_tune_flag);)
		if (WaitRet == 0) {
			//ॢ
			WaitRet = WAIT_TIMEOUT;
		} else if(WaitRet == -ERESTARTSYS) {
			WaitRet = WAIT_FAILED;
		} else {
			if (tmc_dataread_flag == TMC_WAIT_RDS_UPDATE) {
				WaitRet = TMC_RDS_DATA_UPDATE + WAIT_OBJECT_0;
			} else if (tmc_dataread_flag == TMC_WAIT_RDS_DATAREAD_CANCEL) {
				WaitRet = RDS_READ_CANCEL + WAIT_OBJECT_0;
			} else {
				WaitRet = WAIT_FAILED;
			}
		}
		tmc_dataread_flag = TMC_WAIT_INIT;
		DEBMSG_DR(printk(KERN_EMERG "@@@1\r\n");)
//----------------------ReadMode-------------------------------------------
		TMC_ModeLock();
		if ((WaitRet - WAIT_OBJECT_0)  == RDS_READ_CANCEL){
			//󥻥Ԥ֤ɤ
			DEBMSG_DR(printk(KERN_EMERG "@@@2\r\n");)
			if (TMC_CheckMode(RDS_TMC_MODE_CANCEL_WAIT)) {
				//ߤäƸäƤΤǳߤ
				TMC_RDSCommonReadCancel();
				//NGǡ
				TMC_MakeNGData(pBufOut,dwLenOut);
				DEBMSG_DR(printk(KERN_EMERG "-- RDS_READ Cancel end1 --\r\n");)
				//ƤʤΤǼ٥Ȥ󥷥ʥ
//				ResetEvent(ReadEventHandleList[RDS_READ_DAV]);
				
				TMC_ModeUnLock();
				return TRUE;
			} else {
				DEBMSG_DR(printk(KERN_EMERG "/@/\r\n");)
				continue;
			}
		} else if (WaitRet == WAIT_TIMEOUT) {
			DEBMSG_DR(printk(KERN_EMERG "DataRead TimeOut\r\n");)
			//λʤΤǥ顼ǡ֤Ȥ
			//NGǡ
			TMC_MakeNGData(pBufOut,dwLenOut);
//			ResetEvent(ReadEventHandleList[RDS_READ_DAV]);
//			TMC_ChangeMode(RDS_TMC_MODE_IDLE);
			if (TMC_CheckMode(RDS_TMC_MODE_CANCEL_WAIT)) {
//				SetEvent(hRDSCancel);
				SetCancelEvent();
//				ResetEvent(ReadEventHandleList[RDS_READ_CANCEL]);
			} else {
				TMC_ChangeMode(RDS_TMC_MODE_IDLE);
			}
			TMC_ModeUnLock();
			return TRUE;
		//󥻥HandleCloseǤˤäƤʤΤѹ
		} else if (WaitRet == WAIT_FAILED){
			DEBMSG_DR(printk(KERN_EMERG "@@@4\n");)
			//λʤΤǥ顼ǡ֤Ȥ
			//NGǡ
			TMC_MakeNGData(pBufOut,dwLenOut);
//			ResetEvent(ReadEventHandleList[RDS_READ_DAV]);
			if (TMC_CheckMode(RDS_TMC_MODE_CANCEL_WAIT)) {
//				SetEvent(hRDSCancel);
				SetCancelEvent();
//				ResetEvent(ReadEventHandleList[RDS_READ_CANCEL]);
			} else {
				TMC_ChangeMode(RDS_TMC_MODE_IDLE);
			}
			TMC_ModeUnLock();
			return TRUE;
		}
		//ԤΤΤʤΤǥå
		if (TMC_CheckRingCnt()) {
			DEBMSG_DR(printk(KERN_EMERG "-- RDS_READ XXX --\r\n");)
			TMC_RegLock(&flag);
			Ctea5764Driver_DavIntCtrl(TRUE);
			TMC_RegUnLock(&flag);
		} else {
			break;
		}
	}
	//Ԥ֤äƤɤ᤿ǡ֤
	if (writecnt > RDS_READ_BUFFER_SIZE) {
		writecnt = RDS_READ_BUFFER_SIZE;
	}
//	WaitForSingleObject(hRDSReadDataMutex, INFINITE);
	TMC_ReadDataLock();
	memcpy(pBufOut,&(RDSReadData[(int)RdsBufferReadCnt][0]),writecnt);
	RdsBufferReadCnt++;
	if (RdsBufferReadCnt >= RDS_READ_BUFFER_MAX) {
		RdsBufferReadCnt = 0;
	}
	TMC_ReadDataUnLock();
//	ReleaseMutex(hRDSReadDataMutex);

//	RETAILMSG(1, (TEXT("RDS DATA[%02d]   0x%X 0x%X 0x%X 0x%X\r\n"),RdsBufferReadCnt,pBufOut[0],pBufOut[1],pBufOut[3],pBufOut[4]));
	//ǡίޤä˥󥻥Ԥä饭󥻥뤹
	if (TMC_CheckMode(RDS_TMC_MODE_CANCEL_WAIT)) {
		DEBMSG_DR(printk(KERN_EMERG "@@@5\r\n");)
		//󥻥뤷äƸäƤƤΤǳߤʤ褦ˤƤޤ
		TMC_RDSCommonReadCancel();
		//NGǡ
		TMC_MakeNGData(pBufOut,dwLenOut);
		DEBMSG_DR(printk(KERN_EMERG "-- RDS_READ Cancel end2 --\r\n");)
		//󥻥ΥʥꥻåȤƤ
//		ResetEvent(ReadEventHandleList[RDS_READ_CANCEL]);
		TMC_ModeUnLock();
		DEBMSG_DR(printk(KERN_EMERG "W@W\r\n");)
		return TRUE;
	}
	TMC_ChangeMode(RDS_TMC_MODE_IDLE);
	TMC_ModeUnLock();
//------------------------------ReadMode-------------------------------------------
//	RETAILMSG(1, (TEXT("R:%02d 0x%X 0x%X %d\r\n"),RdsBufferReadCnt,pBufOut[0],pBufOut[1],GetTickCount()));
//	DEBMSG_TIME(RETAILMSG(1, (TEXT("read exit %d\r\n"),GetTickCount()));)
	DEBMSG_DR(printk(KERN_EMERG "e@e\r\n");)
	return TRUE;
}

/*---------------------------------------------------------------------------------------
			TMC_MakeNGData
			NGǡ
----------------------------------------------------------------------------------------*/
static BOOL TMC_MakeNGData(char * pBufOut,unsigned long dwLenOut)
{
	int i;
	if ((pBufOut != NULL) &&(dwLenOut > 3)) {
#if 0
		pBufOut[0] = RDS_DATA_FLAG_LSYN;
		pBufOut[1] = RDS_DATA_STAT1_NG;
		pBufOut[2] = RDS_DATA_STAT2_NG;
#else
		for(i = 0;i <dwLenOut;i++) {
			pBufOut[i] = 0;
		}
#endif
		return TRUE;
	} else {
		return FALSE;
	}

}
#define DEBMSG_CANCEL(a)
/*---------------------------------------------------------------------------------------
			TMC_RDSAllCancel
----------------------------------------------------------------------------------------*/
static void TMC_RDSAllCancel(void) {
	unsigned long flag;
	int WaitRet;
	//DAV DesableˤɤԤ򽪤餻
	DEBMSG_CANCEL(printk(KERN_EMERG "------- RDS_ALLCANCEL-------\n");)
	//Եʳϥ⡼ƤΤԤ
	TMC_ModeLock();
	//ƥ⡼ɤԵ
	if(TmcWaitMode == RDS_TMC_MODE_SEEK) {
		DEBMSG_CANCEL(printk(KERN_EMERG "-- TMC_AL Seek Cancel start--\n");)
		if(TMC_CheckCradle()) {
			TMC_RegLock(&flag);
			Ctea5764Driver_SeekIntDisable();
			TMC_RegUnLock(&flag);
		}
		TmcWaitMode = RDS_TMC_MODE_CANCEL_WAIT;
//		SetEvent(SeekHandle[2]);
		tmc_tune_flag = TMC_WAIT_SEEK_CANCEL;
		wake_up_interruptible(&tmc_tune_ack); 
		TMC_ModeUnLock();

		//ääƤΤԤ
//		WaitForSingleObject(hRDSCancel, RDS_CANCEL_TIMEOUT);
		WaitRet = wait_event_interruptible_timeout(tmc_exec_cancel,
								tmc_cancel_flag == 1, msecs_to_jiffies(RDS_CANCEL_TIMEOUT));
		tmc_cancel_flag = 0;
		TMC_ModeLock();
		TMC_ChangeMode(RDS_TMC_MODE_IDLE);
		DEBMSG_CANCEL(printk(KERN_EMERG "-- TMC_AL Seek Cancel end--\n");)
		TMC_ModeUnLock();
		return;
	} else if (TmcWaitMode == RDS_TMC_MODE_READ) {
		DEBMSG_CANCEL(printk(KERN_EMERG "-- TMC_AL Read Cancel start --\n");)
		if(TMC_CheckCradle()) {
			TMC_RegLock(&flag);
			Ctea5764Driver_DavIntCtrl(FALSE);
			TMC_RegUnLock(&flag);

			//RingΥ󥿤򥯥ꥢ
			TMC_ClearRingCount();
		}
		TmcWaitMode = RDS_TMC_MODE_CANCEL_WAIT;
//		SetEvent(ReadEventHandleList[RDS_READ_CANCEL]);
		tmc_dataread_flag = TMC_WAIT_RDS_DATAREAD_CANCEL;
		wake_up_interruptible(&tmc_dataread_ack); 
		TMC_ModeUnLock();
		//äƤΤԤ
//		WaitForSingleObject(hRDSCancel, RDS_CANCEL_TIMEOUT);
		WaitRet = wait_event_interruptible_timeout(tmc_exec_cancel,
								tmc_cancel_flag == 1, msecs_to_jiffies(RDS_CANCEL_TIMEOUT));
		tmc_cancel_flag = 0;
		TMC_ModeLock();
		TMC_ChangeMode(RDS_TMC_MODE_IDLE);
		TMC_ModeUnLock();
//		DEBMSG(RETAILMSG(1, (TEXT("-- TMC_Read Cancel end --\r\n")));)
		return;

	} else if (TmcWaitMode == RDS_TMC_MODE_CANCEL_WAIT) {
		//󥻥Ԥ˥󥻥ʤƤ
		DEBMSG_CANCEL(printk(KERN_EMERG "Cancel Wait Seqence NG\\n");)
		TMC_ModeUnLock();
		return;
	}
	//󥻥뤵줿IDLEǤDAVߤϻߤ
	TMC_RegLock(&flag);
	if(TMC_CheckCradle()) {
		Ctea5764Driver_DavIntCtrl(FALSE);
	}
	TMC_RegUnLock(&flag);

	TmcWaitMode = RDS_TMC_MODE_IDLE;
	TMC_ModeUnLock();

	DEBMSG_CANCEL(printk(KERN_EMERG "------- RDS_ALLCANCEL END-------\n");)
}

#define DEBMSG_ATT(a)
//static struct task_struct *pSwPortth = NULL; 

/*---------------------------------------------------------------------------------------
			TMC_RDSNaviAtt
			4W񤯤ΤǼȿϤʤ뤫⤷ʤΤ
			SeekʤSeekȤ
----------------------------------------------------------------------------------------*/
static BOOL TMC_RDSNaviAtt(char * pBufIn,unsigned long dwLenIn,char * pBufOut,unsigned long dwLenOut,unsigned long *pdwActualOut) {

	unsigned long flag;
	BOOL Flag = FALSE;
	BOOL retflag = FALSE;
//	long lpExitCode;
	int WaitRet;
	DEBMSG_ATT(printk(KERN_EMERG "-- TMC NaviAtt --\n");)
	if (pBufIn == NULL) {
		return FALSE;
	}
	//SeekʤǤ
	TMC_ModeLock();
	if(TmcWaitMode == RDS_TMC_MODE_SEEK) {
		DEBMSG_ATT(printk(KERN_EMERG " -- ATT TMC_Seek Cancel start--\n");)
		if(TMC_CheckCradle()) {
			TMC_RegLock(&flag);
			Ctea5764Driver_SeekIntDisable();
			TMC_RegUnLock(&flag);
		}
		TmcWaitMode = RDS_TMC_MODE_CANCEL_WAIT;
//		SetEvent(SeekHandle[2]);
		tmc_tune_flag = TMC_WAIT_SEEK_CANCEL;
		wake_up_interruptible(&tmc_tune_ack); 
		TMC_ModeUnLock();
//		WaitForSingleObject(hRDSCancel, RDS_CANCEL_TIMEOUT);
		WaitRet = wait_event_interruptible_timeout(tmc_exec_cancel,
								tmc_cancel_flag == 1, msecs_to_jiffies(RDS_CANCEL_TIMEOUT));
		tmc_cancel_flag = 0;
		TMC_ChangeMode(RDS_TMC_MODE_IDLE);
		TMC_ModeUnLock();
		DEBMSG_ATT(printk(KERN_EMERG "-- ATT TMC_Seek Cancel end--\n");)
	} else {
		DEBMSG_ATT(printk(KERN_EMERG "-- ATT 1--\n");)
		TMC_ModeUnLock();
	}
	if (*pBufIn == TMC_RDS_NAVIATT_ON) {
		Flag = TRUE;
	}
	TMC_ModeLock();
	TMC_RegLock(&flag);

	DEBMSG_ATT(printk(KERN_EMERG "-- ATT MakeThread --\n");)
	retflag = Ctea5764Driver_set_SWPort(Flag);
#if 1
//	tasklet_schedule(&tmc_att_tasklet);
//	DEBMSG_ATT(printk(KERN_EMERG "-- ATT MakeThread2 --\n");)
	Ctea5764CheckSwPort();
#else
	//λ
	if (SwPortEvent != NULL) {
		CloseHandle(SwPortEvent);
	}
	if (gTeaSwPortThread != NULL) {
		GetExitCodeThread(gTeaSwPortThread,&lpExitCode);
		if (lpExitCode == STILL_ACTIVE) {
			DEBMSG_ATT(printk(KERN_EMERG "-- ATT Thre ACTIVE--\n");)
		}
		CloseHandle(gTeaSwPortThread);
	}
	//SWPortƻ륹å
	SwPortEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
	gTeaSwPortThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE) TEASWPortThread, NULL, 0, NULL);
	if ( gTeaSwPortThread == NULL ) {
		return (FALSE);
	}
#endif
//	retflag = Ctea5764Driver_set_SWPort(Flag);
	TMC_RegUnLock(&flag);
	TMC_ModeUnLock();
	
	DEBMSG_ATT(printk(KERN_EMERG "-- TMC NaviAtt end --\n");)
	return retflag;
}

/*---------------------------------------------------------------------------------------
			TMC_RDSNaviAtt
----------------------------------------------------------------------------------------*/
static BOOL TMC_RDSGetAttSts(char * pBufIn,unsigned long dwLenIn,char * pBufOut,unsigned long dwLenOut,unsigned long *pdwActualOut) {
	unsigned long flag;
	BOOL mode = FALSE;
	TMC_RegLock(&flag);
	 mode = Ctea5764Driver_get_SWPort();
	 if((pBufOut != NULL) || (pdwActualOut != NULL)) {
		 if (mode == true) {
			 *pBufOut = TMC_RDS_NAVIATT_ON;
		 } else {
			 *pBufOut = TMC_RDS_NAVIATT_OFF;
		 }
		 *pdwActualOut = 1;
	 }
	TMC_RegUnLock(&flag);
	return mode;
}


#if 0
#define DEBMSG_ATT(a)
/*---------------------------------------------------------------------------------------
			TEASWPortThread
----------------------------------------------------------------------------------------*/
//static int TEASWPortThread(void *data)
static void TEASWPortThread(unsigned long data)
{
	unsigned long flag;
	long SwRet;
//	BOOL ret = FALSE;
//	SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_NORMAL);
	wait_queue_head_t tmc_swp_timer;
	DEBMSG_ATT(printk(KERN_EMERG "-- ATT Thread start --\n");)
	init_waitqueue_head(&tmc_swp_timer);
	DEBMSG_ATT(printk(KERN_EMERG "-- ATT Thread 2nd --\n");)
	SwRet = wait_event_interruptible_timeout(tmc_swp_timer,0, msecs_to_jiffies(50));
	DEBMSG_ATT(printk(KERN_EMERG "-- ATT Thread timeout %d --\n",(int)SwRet);)
	if (SwRet == 0) {
		//50msec˽񤭹ߤԤʤä
		TMC_RegLock(&flag);
		Ctea5764CheckSwPort();
		TMC_RegUnLock(&flag);
	} else {
		//λ
//		RETAILMSG(1, (TEXT("@!!@\r\n")));
	}
}
#endif
static void TMC_RDSInit(void) 
{
	navuGpioSetLevel(NAVU_OUT_RDS_POWER, 0);
	mdelay(20);
	navuGpioSetLevel(NAVU_OUT_RDS_POWER, 1);
	if (pte5764_FormPtr == NULL) {
		pte5764_FormPtr = kmalloc(sizeof(II2CForm),GFP_KERNEL);
		pte5764_FormPtr->i2c_send = I2CFormLinux_i2c_send;
		pte5764_FormPtr->i2c_recv = I2CFormLinux_i2c_recv;
		pte5764_FormPtr->i2c_init = I2CFormLinux_init;
		pte5764_FormPtr->i2c_PwSend = I2CFormLinux_i2c_send;
		pte5764_FormPtr->i2c_PwUp = I2CFormLinux_PowerUp;
		pte5764_FormPtr->i2c_PwDown = I2CFormLinux_PowerDown;
	}

	if (pte5764_ListenerPtr == NULL) {
		pte5764_ListenerPtr = kmalloc(sizeof(IRDSListener),GFP_KERNEL);
		pte5764_ListenerPtr->update_rds_data = tea5764_update_rds_data;
	}
	TMC_ClearRingCount();

	//ǥХν
	Ctea5764Driver_init(pte5764_ListenerPtr,pte5764_FormPtr);
}
#define DEBMSG_REINIT(a)
/*---------------------------------------------------------------------------------------
			æ줿ΤǥǥХκƽ
----------------------------------------------------------------------------------------*/
static BOOL TMC_ReInitRds(void)
{
	unsigned long flag;
	int retryyloop = 0;
	DEBMSG_REINIT(printk(KERN_EMERG "-- TMC_ReInitRds --\n");)
	if (tea5764_fm_client == NULL) {
		//åʤ
		DEBMSG_REINIT(printk(KERN_EMERG "-- TMC_ReInitRds tea5764_fm_client NULL--\n");)
		i2c_attach_client(tea5764_fm_client);
		return false;
	}
	if (pte5764_FormPtr == NULL) {
		DEBMSG_REINIT(printk(KERN_EMERG "-- pte5764_FormPtr NULL--\n");)
		pte5764_FormPtr = kmalloc(sizeof(II2CForm),GFP_KERNEL);
		if (pte5764_FormPtr == NULL) {
			return false;
		}
		pte5764_FormPtr->i2c_send = I2CFormLinux_i2c_send;
		pte5764_FormPtr->i2c_recv = I2CFormLinux_i2c_recv;
		pte5764_FormPtr->i2c_init = I2CFormLinux_init;
		pte5764_FormPtr->i2c_PwSend = I2CFormLinux_i2c_send;
		pte5764_FormPtr->i2c_PwUp = I2CFormLinux_PowerUp;
		pte5764_FormPtr->i2c_PwDown = I2CFormLinux_PowerDown;

		pte5764_ListenerPtr = kmalloc(sizeof(IRDSListener),GFP_KERNEL);
		if (pte5764_ListenerPtr == NULL) {
			kfree(pte5764_FormPtr);
			pte5764_FormPtr = NULL;
			return false;
		}
		pte5764_ListenerPtr->update_rds_data = tea5764_update_rds_data;
	}
	if(!(pte5764_FormPtr->i2c_init)()) {
		return FALSE;
	}
	DEBMSG_REINIT(printk(KERN_EMERG "-- TMC_ReInitRds 2--\n");)
	RdsBufferReadCnt = 0;
	RdsBufferWriteCnt = 0;
	navuGpioSetLevel(NAVU_OUT_RDS_POWER, 0);
	mdelay(20);
	navuGpioSetLevel(NAVU_OUT_RDS_POWER, 1);
	mdelay(300);
	//ǥХ
	TMC_RegLock(&flag);
//	Ctea5764Driver_init(pListenerPtr,pFormPtr);
	while(retryyloop < 2) {
	DEBMSG_REINIT(printk(KERN_EMERG "-- TMC_ReInitRds 3--\n");)
		if(!Ctea5764Driver_ReInit(pte5764_ListenerPtr,pte5764_FormPtr)) {
			retryyloop++;
		} else {
			break;
		}
	DEBMSG_REINIT(printk(KERN_EMERG "-- TMC_ReInitRds 4--\n");)
	}
	mdelay(10);
	TMC_RegUnLock(&flag);
	if (retryyloop >= 2) {
		DEBMSG_REINIT(printk(KERN_EMERG "-- ReInit NG end --\n");)
		return false;
	} else {
		DEBMSG_REINIT(printk(KERN_EMERG "-- ReInit OK end --\n");)
		return true;
	}

}

