/*
 * drivers/mtd/maps/xyz3-map.c
 *   MTD map driver for Sony XYZ3
 * 
 * Author:     Hirokazu ABE and Ryoichi KATO
 * Created:    Mar 30, 2006
 * Copyright:  Sony Corp.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/vmalloc.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/mtd/partitions.h>

#include <asm/io.h>
#include <asm/arch/xyz3-norheader.h>

#define ROM_ADDR 0x0
#define ROM_SIZE 0x00200000 /* MAX OF */

static struct mtd_info* mtdinfo = NULL;

static const struct map_info mapinfo_init = {
    .name = "xyz3-map",
    .size = ROM_SIZE,
    .bankwidth = 2,
    .phys = ROM_ADDR,
};
static struct map_info mapinfo;

static const struct mtd_partition partition_init[] = {
    {
        .name = "all area",
        .offset = 0,
        .size = 0,
        .mask_flags = 0,
    },
    {
        .name = "sdh",
        .size = 0x8000,
        .mask_flags = 0,
    },
    {
        .name = "log1",
        .size = 0x2000,
        .mask_flags = 0,
    },
    {
        .name = "log2",
        .size = 0x2000,
        .mask_flags = 0,
    },
    {
        .name = "check",
        .size = 0x4000,
        .mask_flags = 0,
    },
};
#define noof_mtd_partitions ((int)(sizeof(partition_init)/sizeof(partition_init[0])))
static struct mtd_partition partition[noof_mtd_partitions];

static void xyz3_map_exit(void)
{
    if(mtdinfo != NULL){
	del_mtd_partitions(mtdinfo);
	map_destroy(mtdinfo);
	mtdinfo = NULL;
    }

    if(mapinfo.virt != NULL){
	iounmap(mapinfo.virt);
	mapinfo.virt = NULL;
    }
}

static int __init xyz3_map_init(void)
{
    int i;
    uint32_t offset_last;

    memcpy(&mapinfo, &mapinfo_init, sizeof(mapinfo));
    memcpy(partition, partition_init, sizeof(partition));
  
    mapinfo.virt = ioremap(ROM_ADDR, ROM_SIZE);
    if(mapinfo.virt == NULL){
	return -EIO;
    }
    simple_map_init(&mapinfo);

    mtdinfo = do_map_probe("cfi_probe", &mapinfo);
    if(mtdinfo == NULL){
	xyz3_map_exit();
	return -ENXIO;
    }
    mtdinfo->owner = THIS_MODULE;

    /* etc. area offset, size */
    offset_last = mtdinfo->size;
    for(i = 1; i < noof_mtd_partitions; i++){
        if(partition[noof_mtd_partitions-i].size == 0) {
            partition[noof_mtd_partitions-i].size = mtdinfo->erasesize;
        }
        partition[noof_mtd_partitions-i].offset = offset_last - partition[noof_mtd_partitions-i].size;

        offset_last = partition[noof_mtd_partitions-i].offset;
    }
    partition[0].size = partition[1].offset - partition[0].offset;

    add_mtd_partitions(mtdinfo, partition, noof_mtd_partitions);

    return 0;
}

module_init(xyz3_map_init);
module_exit(xyz3_map_exit);

MODULE_AUTHOR("");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("MTD map driver for Sony XYZ3");
