/*
 * linux/drivers/pcmcia/pxa2xx_xyz3.c
 *
 * Mainstone PCMCIA specific routines.
 *
 * Created:     Apr 1, 2006
 * Author:      Takayoshi Sakai
 * Copyright:   Sony corp.
 *
 * Based on linux/drivers/pcmcia/pxa2xx_mainstone.c
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/platform_device.h>

#include <pcmcia/ss.h>

#include <asm/hardware.h>
#include <asm/irq.h>

#include <asm/arch/pxa-regs.h>

#include "soc_common.h"

#include <linux/delay.h>
#include <asm/arch/xyz3.h>


static int xyz3_pcmcia_hw_init(struct soc_pcmcia_socket *skt)
{

        /*
         * Setup default state of GPIO outputs
         * before we enable them as outputs.
         */
         GPSR(GPIO48_nPOE) =
                GPIO_bit(GPIO48_nPOE) |
                GPIO_bit(GPIO49_nPWE) |
                GPIO_bit(GPIO50_nPIOR) |
                GPIO_bit(GPIO51_nPIOW) ;
          GPSR3=
                GPIO_bit(102) |  //GPIO102_nPCE_1
                GPIO_bit(105); //GPIO105_nPCE_2

        pxa_gpio_mode(GPIO48_nPOE_MD);
        pxa_gpio_mode(GPIO49_nPWE_MD);
        pxa_gpio_mode(GPIO50_nPIOR_MD);
        pxa_gpio_mode(GPIO51_nPIOW_MD);

        pxa_gpio_mode(102 | GPIO_ALT_FN_1_OUT); //GPIO102_nPCE_1_MD
        pxa_gpio_mode(105 | GPIO_ALT_FN_1_OUT); //GPIO105_nPCE_2_MD

        pxa_gpio_mode(GPIO55_nPREG_MD);

        pxa_gpio_mode(GPIO56_nPWAIT_MD);

        pxa_gpio_mode(GPIO57_nIOIS16_MD);


        pxa_gpio_mode(97 | GPIO_OUT); //CF Power Enable
        pxa_gpio_mode(99 | GPIO_OUT); //CF Reset
        pxa_gpio_mode(100 | GPIO_OUT); //CF Addr Enable


        GPSR3=0x2;  //GPIO97 Power enable
        GPSR3=0x10; //GPIO100 Addr enable
        msleep(10);
        GPCR3=0x8;  //GPIO99 Reset


//      printk("xyz3_pcmcia_hw_init End\n");

  {
        skt->irq=XYZ3_CF_IRQ;
  }

        return 0;
}

static void xyz3_pcmcia_hw_shutdown(struct soc_pcmcia_socket *skt)
{

  /* disable IRQs */

  /* Disable CF bus: */
  GPCR3=0x2;       // Compact flash power off
  GPCR3=0x8;       // Compact flash disable power up reset
  GPCR3=0x10;      // Compact flash bus off
}

static void xyz3_pcmcia_socket_state(struct soc_pcmcia_socket *skt,
                                    struct pcmcia_state *state)
{
        unsigned long status;

//  printk("xyz3_pcmcia_socket_state \n");
        memset(state, 0,sizeof(struct pcmcia_state));

        status=GPLR0;
        state->ready  = (status & (1<<22)) ? 1 : 0;


        state->detect =  1; //Always detected

        state->bvd1   = 0;
        state->bvd2   = 0;
        state->vs_3v  = 1;
        state->vs_Xv  = 1;
        state->wrprot = 0;  /* not available */
}

static int xyz3_pcmcia_configure_socket(struct soc_pcmcia_socket *skt,
                                       const socket_state_t *state)
{

  unsigned long flags;

  local_irq_save(flags);


  switch(state->Vcc){
  case 0:
   // printk("0V CF Slot\n");
    GPCR3=0x2;
    break;

  case 50:
    printk("%s(): CS asked for 5V, applying 3.3V...\n",
           __FUNCTION__);

  case 33:  /* Can only apply 3.3V to the CF slot. */
   // printk("3.3V to the CF slot\n");
    GPSR3=0x2;
    break;

  default:
    printk("%s(): unrecognized Vcc %u\n", __FUNCTION__,
           state->Vcc);
    local_irq_restore(flags);
    return -1;
  }

  if (state->flags & SS_RESET){
    GPSR3=0x8; //printk("SS_RESET=1\n");
  }else{
    GPCR3=0x8;// printk("SS_RESET=0\n");
  }

  /* Silently ignore Vpp, output enable, speaker enable. */


  local_irq_restore(flags);

  return 0;
}

static void xyz3_pcmcia_socket_init(struct soc_pcmcia_socket *skt)
{
        // printk("xyz3_pcmcia_socket_init\n");
  /*
         * Setup default state of GPIO outputs
         * before we enable them as outputs.
         */
         GPSR(GPIO48_nPOE) =
                GPIO_bit(GPIO48_nPOE) |
                GPIO_bit(GPIO49_nPWE) |
                GPIO_bit(GPIO50_nPIOR) |
                GPIO_bit(GPIO51_nPIOW) ;
          GPSR3=
                GPIO_bit(102) |  //GPIO102_nPCE_1
                GPIO_bit(105); //GPIO105_nPCE_2

        pxa_gpio_mode(GPIO48_nPOE_MD);
        pxa_gpio_mode(GPIO49_nPWE_MD);
        pxa_gpio_mode(GPIO50_nPIOR_MD);
        pxa_gpio_mode(GPIO51_nPIOW_MD);

        pxa_gpio_mode(102 | GPIO_ALT_FN_1_OUT); //GPIO102_nPCE_1_MD
        pxa_gpio_mode(105 | GPIO_ALT_FN_1_OUT); //GPIO105_nPCE_2_MD

        pxa_gpio_mode(GPIO55_nPREG_MD);

        pxa_gpio_mode(GPIO56_nPWAIT_MD);

        pxa_gpio_mode(GPIO57_nIOIS16_MD);


        pxa_gpio_mode(97 | GPIO_OUT); //CF Power Enable
        pxa_gpio_mode(99 | GPIO_OUT); //CF Reset
        pxa_gpio_mode(100 | GPIO_OUT); //CF Addr Enable


        GPSR3=0x2;  //GPIO97 Power enable
        GPSR3=0x10; //GPIO100 Addr enable
        msleep(10);
        GPCR3=0x8;  //GPIO99 Reset
}

static void xyz3_pcmcia_socket_suspend(struct soc_pcmcia_socket *skt)
{
 // printk("xyz3_pcmcia_socket_suspend\n");
        GPCR3=0x2;  //GPIO97 Power enable
        GPCR3=0x10; //GPIO100 Addr enable
        GPSR3=0x8;  //GPIO99 Reset
}

static struct pcmcia_low_level xyz3_pcmcia_ops = {
        .owner                  = THIS_MODULE,
        .hw_init                = xyz3_pcmcia_hw_init,
        .hw_shutdown            = xyz3_pcmcia_hw_shutdown,
        .socket_state           = xyz3_pcmcia_socket_state,
        .configure_socket       = xyz3_pcmcia_configure_socket,
        .socket_init            = xyz3_pcmcia_socket_init,
        .socket_suspend         = xyz3_pcmcia_socket_suspend,
        .nr                     = 1,
};

static struct platform_device *xyz3_pcmcia_device;

static int __init xyz3_pcmcia_init(void)
{
        int ret;

      // printk("xyz3_pcmcia_init\n");

        xyz3_pcmcia_device = kmalloc(sizeof(*xyz3_pcmcia_device), GFP_KERNEL);
        if (!xyz3_pcmcia_device)
                return -ENOMEM;
        memset(xyz3_pcmcia_device, 0, sizeof(*xyz3_pcmcia_device));
        xyz3_pcmcia_device->name = "pxa2xx-pcmcia";
        xyz3_pcmcia_device->dev.platform_data = &xyz3_pcmcia_ops;

        ret = platform_device_register(xyz3_pcmcia_device);
        if (ret)
                kfree(xyz3_pcmcia_device);

        return ret;
}

static void __exit xyz3_pcmcia_exit(void)
{
        /*
         * This call is supposed to free our xyz3_pcmcia_device.
         * Unfortunately platform_device don't have a free method, and
         * we can't assume it's free of any reference at this point so we
         * can't free it either.
         */
        platform_device_unregister(xyz3_pcmcia_device);
}

module_init(xyz3_pcmcia_init);
module_exit(xyz3_pcmcia_exit);

MODULE_LICENSE("GPL");
