/*
 * linux/drivers/video/sonynavu_pxafb.c
 *
 *  Copyright (C) 2005 Sony. t-sakai@me.sony.co.jp
 *
 * Based on
 *  linux/drivers/video/pxafb.c
 *    -- Intel PXA250/210 LCD Controller Frame Buffer Device
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/interrupt.h>
#include <linux/slab.h>
#include <linux/fb.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <linux/cpufreq.h>
#include <linux/platform_device.h>
#include <linux/dma-mapping.h>
#include <linux/jiffies.h>

#include <asm/hardware.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <asm/div64.h>
#include <asm/arch/pxa-regs.h>
#include <asm/arch/bitfield.h>
#include <asm/arch/pxafb.h>
#include <asm/arch/dma.h>
/*
 * debugging?
 */
#define DEBUG 0
/*
 * Complain if VAR is out of range.
 */
#define DEBUG_VAR 0

#include "sonynavu_pxafb.h"

/* Bits which should not be set in machine configuration structures */
#define LCCR0_INVALID_CONFIG_MASK (LCCR0_OUM|LCCR0_BM|LCCR0_QDM|LCCR0_DIS|LCCR0_EFM|LCCR0_IUM|LCCR0_SFM|LCCR0_LDM|LCCR0_ENB)
#define LCCR3_INVALID_CONFIG_MASK (LCCR3_HSP|LCCR3_VSP|LCCR3_PCD|LCCR3_BPP)

static DECLARE_MUTEX(fcs_lcd_sem);


static int sonynavu_pxafb_activate_var(struct fb_var_screeninfo *var, struct sonynavu_pxafb_info *fbi);
static void sonynavu_pxafb_enable_controller(struct sonynavu_pxafb_info *fbi);
static void sonynavu_pxafb_disable_controller(struct sonynavu_pxafb_info *fbi);

#ifdef CONFIG_MACH_SONY_NAVU_XYZ4
extern int sonynavu_lcdssp_power_on(void);
extern int sonynavu_lcdssp_power_off(void);
#endif
//static void set_ctrlr_state(struct sonynavu_pxafb_info *fbi, u_int state);

//Display resolution
#ifndef LCD_XRES
static int LCD_XRES;
#endif
#ifndef LCD_YRES
static int LCD_YRES;
#endif
#ifndef LCD_LCCR1
static int LCD_LCCR1;
#endif
#ifndef LCD_LCCR2
static int LCD_LCCR2;
#endif
#ifndef LCD_LCCR3
static int LCD_LCCR3;
#endif
#ifndef LCD_LCCR4
static int LCD_LCCR4;
#endif
//
#ifdef CONFIG_FB_PXA_PARAMETERS
#define PXAFB_OPTIONS_SIZE 256
static char g_options[PXAFB_OPTIONS_SIZE] __initdata = "";
#endif

//QVGA or not
int QVGA_flag;

#define CLEAR_LCD_INTR(reg, intr) do {  \
	reg = (intr);			\
}while(0)

#define WAIT_FOR_LCD_INTR(reg,intr,timeout) ({	\
	int __done =0;				\
	int __t = timeout;			\
	while (__t) {				\
		__done = (reg) & (intr);	\
		if (__done) break;		\
		mdelay(10);			\
		__t--;				\
	}					\
	if (!__t) printk("wait " #intr " timeount\n");\
	__done;					\
})

//
static  sonynavu_pxafb_vsyncinfo_t	StartFrameInfo;

void sonynavu_pxafb_wait_startframe(void){
  DPRINTK("sonynavu_pxafb_wait_startframe:\n");
  atomic_inc(&StartFrameInfo.waitcnt);
  interruptible_sleep_on(&StartFrameInfo.waitq);
  atomic_dec(&StartFrameInfo.waitcnt);
  DPRINTK("wait_fin\n");
}

void sonynavu_pxafb_wait_cancel_startframe(void){
  DPRINTK("sonynavu_pxafb_wait_cancel_startframe:\n");
  local_irq_disable();
  //Wake up proccesses
  if(atomic_read(&StartFrameInfo.waitcnt)){
    wake_up_interruptible(&StartFrameInfo.waitq);
  }
  local_irq_enable();
}


/************************************************************************/
//DMAāEBaseFrame
int SetDMASetting0(struct sonynavu_pxafb_info *fbi,int mode){
	int i;
	dma_addr_t screen0_dma;

	if(mode==1){
		screen0_dma=fbi->screen0_dma1;
		fbi->Baseframe.Flip=1;
	}else{
		screen0_dma=fbi->screen0_dma0;
		fbi->Baseframe.Flip=0;
	}
	//žȥ
	DPRINTK("%d %d %d \n",fbi->Baseframe.FrameWidth,fbi->Baseframe.FrameX,fbi->Baseframe.FrameY);
	DPRINTK("%d %d \n",fbi->Baseframe.DispWidth,fbi->Baseframe.DispHeight);

	for(i=0;i<(fbi->Baseframe.DispHeight);i++){
		fbi->dmadesc_fbover0_cpu[i].fsadr 
			= screen0_dma
			+((fbi->Baseframe.FrameWidth*fbi->Baseframe.FrameY+(fbi->Baseframe.FrameX & 0xFFF8))*(LCD_BASE_BPP / 8))
			+((fbi->Baseframe.FrameWidth*LCD_BASE_BPP / 8)*i);
		fbi->dmadesc_fbover0_cpu[i].fidr = 0;
		fbi->dmadesc_fbover0_cpu[i].ldcmd 
			= ((fbi->Baseframe.DispWidth * LCD_BASE_BPP / 8));
	}
	fbi->dmadesc_fbover0_cpu[fbi->Baseframe.DispHeight-1].ldcmd |= LDCMD_EOFINT;
        fbi->dmadesc_fbover0_cpu[0].ldcmd |= LDCMD_SOFINT; //Start of frame interrupt
  
	fbi->dmadesc_palette0_cpu->fsadr = fbi->palette0_dma;
	fbi->dmadesc_palette0_cpu->fidr  = 0;
	fbi->dmadesc_palette0_cpu->ldcmd 
		= (fbi->palette0_size * sizeof(u32)) | LDCMD_PAL;
	//ΥǥEץΥɥE
	for(i=0;i<(fbi->Baseframe.DispHeight-1);i++){
		fbi->dmadesc_fbover0_cpu[i].fdadr
			=fbi->dmadesc_fbover0_dma+16*(i+1);
	}
	fbi->dmadesc_fbover0_cpu[fbi->Baseframe.DispHeight-1].fdadr 
		= fbi->dmadesc_palette0_dma;
	fbi->dmadesc_palette0_cpu->fdadr = fbi->dmadesc_fbover0_dma;

	fbi->fdadr0 = fbi->dmadesc_palette0_dma; /* flips back and forth between pal and fbhigh */
	return(mode);
}

//DMAāEOverlay1(GUI)
int SetDMASetting1(struct sonynavu_pxafb_info *fbi,int mode){

	int i;
	dma_addr_t screen1_dma;
  switch(mode){
  case 1:
		screen1_dma=fbi->screen1_dma1;
		fbi->Overlay1.Flip=1;
                break;
  case 0:
		screen1_dma=fbi->screen1_dma0;
		fbi->Overlay1.Flip=0;
                break;
  case 2:
  default:
                screen1_dma=fbi->screen2_dma;

        fbi->Overlay1.Flip=2;
                break;
	}

	//žȥ
	for(i=0;i<fbi->Overlay1.DispHeight;i++){
		fbi->dmadesc_fbover1_cpu[i].fsadr 
			= screen1_dma
			+((fbi->Overlay1.FrameWidth * fbi->Overlay1.FrameY + fbi->Overlay1.FrameX)*( LCD_OVERLAY1_BPP / 8))
			+((fbi->Overlay1.FrameWidth * LCD_OVERLAY1_BPP / 8)*i);
		fbi->dmadesc_fbover1_cpu[i].fidr = 0;
		fbi->dmadesc_fbover1_cpu[i].ldcmd 
			= ((fbi->Overlay1.DispWidth * LCD_OVERLAY1_BPP / 8));
	}

	fbi->dmadesc_palette1_cpu->fsadr = fbi->palette1_dma;
	fbi->dmadesc_palette1_cpu->fidr  = 0;
	fbi->dmadesc_palette1_cpu->ldcmd 
		= (fbi->palette1_size * sizeof(u32)) | LDCMD_PAL;
	//ΥǥEץΥɥE
	for(i=0;i<fbi->Overlay1.DispHeight-1;i++){
		fbi->dmadesc_fbover1_cpu[i].fdadr
			=fbi->dmadesc_fbover1_dma+16*(i+1);
	}
	if(LCD_OVERLAY1_BPP!=8){	
		fbi->dmadesc_fbover1_cpu[fbi->Overlay1.DispHeight-1].fdadr 
			= fbi->dmadesc_fbover1_dma;    
		fbi->dmadesc_palette1_cpu->fdadr = fbi->dmadesc_fbover1_dma;

		fbi->fdadr1 = fbi->dmadesc_fbover1_dma; /* flips back and forth between pal and fbhigh */
	}else{
		fbi->dmadesc_palette1_cpu->fdadr = fbi->dmadesc_fbover1_dma;
		fbi->dmadesc_fbover1_cpu[fbi->Overlay1.DispHeight-1].fdadr 
			= fbi->dmadesc_palette1_dma;
		fbi->fdadr1 = fbi->dmadesc_palette1_dma; /* flips back and forth between pal and fbhigh */
	}
	return(mode);
}

//DMAāEOverlay1(GUI)
#ifdef CONFIG_FB_SONYNAVU_START
static int StartUp_Count=0;
extern unsigned short NavU_StartUpImage[];

int SetDMASetting1_StartUp(struct sonynavu_pxafb_info *fbi,int mode){

	int i;
	unsigned short *pp;
	dma_addr_t screen1_dma;
	screen1_dma=(dma_addr_t)(&NavU_StartUpImage[0]);
	
	if(mode==1){
		screen1_dma=fbi->screen1_dma1;
		fbi->Overlay1.Flip=1;
		pp=(unsigned short *)fbi->screen1_cpu1;
	}else{
		screen1_dma=fbi->screen1_dma0;
		fbi->Overlay1.Flip=0;
		pp=(unsigned short *)fbi->screen1_cpu0;
	}

	for(i=0;i<(480*272);i++){
		pp[i]=NavU_StartUpImage[i];
	}
	//žȥ
	for(i=0;i<fbi->Overlay1.DispHeight;i++){
		fbi->dmadesc_fbover1_cpu[i].fsadr 
			= screen1_dma
			+((fbi->Overlay1.FrameWidth * fbi->Overlay1.FrameY + fbi->Overlay1.FrameX)*( LCD_OVERLAY1_BPP / 8))
			+((fbi->Overlay1.FrameWidth * LCD_OVERLAY1_BPP / 8)*i);
		fbi->dmadesc_fbover1_cpu[i].fidr = 0;
		fbi->dmadesc_fbover1_cpu[i].ldcmd 
			= ((fbi->Overlay1.DispWidth * LCD_OVERLAY1_BPP / 8));
	}

	fbi->dmadesc_palette1_cpu->fsadr = fbi->palette1_dma;
	fbi->dmadesc_palette1_cpu->fidr  = 0;
	fbi->dmadesc_palette1_cpu->ldcmd 
		= (fbi->palette1_size * sizeof(u32)) | LDCMD_PAL;
	//ΥǥEץΥɥE
	for(i=0;i<fbi->Overlay1.DispHeight-1;i++){
		fbi->dmadesc_fbover1_cpu[i].fdadr
			=fbi->dmadesc_fbover1_dma+16*(i+1);
	}
	
	fbi->dmadesc_fbover1_cpu[fbi->Overlay1.DispHeight-1].fdadr 
			= fbi->dmadesc_fbover1_dma;    
	fbi->dmadesc_palette1_cpu->fdadr = fbi->dmadesc_fbover1_dma;
	fbi->fdadr1 = fbi->dmadesc_fbover1_dma; /* flips back and forth between pal and fbhigh */
	
        printk("StartUp_Count\n");
        StartUp_Count=1 ; 

	return(mode);
}
#else

#define SetDMASetting1_StartUp(a,b) SetDMASetting1(a,b)

#endif

//DMAāEOverlay2_Linux
int SetDMASetting2_Console(struct sonynavu_pxafb_info *fbi,int mode){

	fbi->Overlay2_Console.Flip=0;
	//Overlay2 Linux
	fbi->dmadesc_fbover2_cpu->fsadr = fbi->screen2_dma;
	fbi->dmadesc_fbover2_cpu->fidr = 0;
	fbi->dmadesc_fbover2_cpu->ldcmd 
		= ((LCD_XRES * LCD_YRES * LCD_BPP / 8));

	fbi->dmadesc_palette2_cpu->fsadr = fbi->palette2_dma;
	fbi->dmadesc_palette2_cpu->fidr  = 0;
	fbi->dmadesc_palette2_cpu->ldcmd 
		= (fbi->palette2_size * sizeof(u32)) | LDCMD_PAL;
#if 0
	fbi->dmadesc_palette2_cpu->fdadr = fbi->dmadesc_fbover2_dma;
	fbi->dmadesc_fbover2_cpu->fdadr = fbi->dmadesc_palette2_dma;

	fbi->fdadr2 = fbi->dmadesc_palette2_dma; /* flips back and forth between pal and fbhigh */
#else
	fbi->dmadesc_palette2_cpu->fdadr = fbi->dmadesc_fbover2_dma;
	fbi->dmadesc_fbover2_cpu->fdadr = fbi->dmadesc_fbover2_dma;

	fbi->fdadr2 = fbi->dmadesc_fbover2_dma; /* flips back and forth between pal and fbhigh */
#endif
	return(0);
}
 
//DMAāEOverlay2_Video
int SetDMASetting2_Video(struct sonynavu_pxafb_info *fbi,int mode){

	//Overlay2 Video
	fbi->Overlay2_Video.Flip=0;
	//Memory->LCD
	fbi->dmadesc_fbvideoY_cpu->fsadr = fbi->screenvideo_y_dma;
	fbi->dmadesc_fbvideoY_cpu->fidr = 0;
	fbi->dmadesc_fbvideoY_cpu->ldcmd 
		= (VIDEO_XRES * VIDEO_YRES );
	fbi->dmadesc_fbvideoY_cpu->fdadr = fbi->dmadesc_fbvideoY_dma;

	fbi->dmadesc_fbvideoB_cpu->fsadr = fbi->screenvideo_b_dma;
	fbi->dmadesc_fbvideoB_cpu->fidr = 0;
	fbi->dmadesc_fbvideoB_cpu->ldcmd 
		= (VIDEO_XRES * VIDEO_YRES /2);
	fbi->dmadesc_fbvideoB_cpu->fdadr = fbi->dmadesc_fbvideoB_dma;

	fbi->dmadesc_fbvideoR_cpu->fsadr = fbi->screenvideo_r_dma;
	fbi->dmadesc_fbvideoR_cpu->fidr = 0;
	fbi->dmadesc_fbvideoR_cpu->ldcmd 
		= (VIDEO_XRES * VIDEO_YRES /2);
	fbi->dmadesc_fbvideoR_cpu->fdadr = fbi->dmadesc_fbvideoR_dma;

	fbi->fdadrY = fbi->dmadesc_fbvideoY_dma; 
	fbi->fdadrB = fbi->dmadesc_fbvideoB_dma; 
	fbi->fdadrR = fbi->dmadesc_fbvideoR_dma; 

	return(0);
}

//DMAāECursor
int SetDMASetting2_Cursor(struct sonynavu_pxafb_info *fbi,int mode){

	dma_addr_t screen3_dma;
	switch(mode){
	case 0:
		screen3_dma=fbi->screen3_dma0;
		fbi->Cursor.Flip=0;
		break;
	case 1:
		screen3_dma=fbi->screen3_dma1;
		fbi->Cursor.Flip=1;
		break;
	case 2:
		screen3_dma=fbi->screen3_dma2;
		fbi->Cursor.Flip=2;
		break;
	case 3:
		screen3_dma=fbi->screen3_dma3;
		fbi->Cursor.Flip=3;
		break;
	default :
		return(-1);
	}
	fbi->dmadesc_fbcursor_cpu->fsadr = screen3_dma;
	fbi->dmadesc_fbcursor_cpu->fidr = 0;
	fbi->dmadesc_fbcursor_cpu->ldcmd 
		= ((LCD_CURSOR_XRES * LCD_CURSOR_YRES * LCD_CURSOR_BPP / 8) );

	fbi->dmadesc_palette3_cpu->fsadr = fbi->palette3_dma;
	fbi->dmadesc_palette3_cpu->fidr  = 0;
	fbi->dmadesc_palette3_cpu->ldcmd 
		= (fbi->palette3_size * sizeof(u32)) | LDCMD_PAL;
	fbi->dmadesc_palette3_cpu->fdadr = fbi->dmadesc_fbcursor_dma;
	fbi->dmadesc_fbcursor_cpu->fdadr = fbi->dmadesc_palette3_dma;

	fbi->fdadr5 = fbi->dmadesc_palette3_dma; 
	return(0);
}
/************************************************************************/

static inline void sonynavu_pxafb_schedule_work(struct sonynavu_pxafb_info *fbi, u_int state)
{
	unsigned long flags;

	local_irq_save(flags);
	/*
	 * We need to handle two requests being made at the same time.
	 * There are two important cases:
	 *  1. When we are changing VT (C_REENABLE) while unblanking (C_ENABLE)
	 *     We must perform the unblanking, which will do our REENABLE for us.
	 *  2. When we are blanking, but immediately unblank before we have
	 *     blanked.  We do the "REENABLE" thing here as well, just to be sure.
	 */
	if (fbi->task_state == C_ENABLE && state == C_REENABLE)
		state = (u_int) -1;
	if (fbi->task_state == C_DISABLE && state == C_ENABLE)
		state = C_REENABLE;

	if (state != (u_int)-1) {
                fbi->task_state = state;
		schedule_work(&fbi->task);
	}
	local_irq_restore(flags);
}

static inline u_int chan_to_field(u_int chan, struct fb_bitfield *bf)
{
	chan &= 0xffff;
	chan >>= 16 - bf->length;
	return chan << bf->offset;
}


//Update SAKAI 051003
static int
sonynavu_pxafb_setpalettereg(u_int regno, u_int red, u_int green, u_int blue,
		       u_int trans, struct fb_info *info)
{
	struct sonynavu_pxafb_info *fbi = (struct sonynavu_pxafb_info *)info;
	u_int val, ret = 1;

	if (regno < fbi->palette2_size) {

		val = ((red << 8) & 0xFC0000);
		val |= ((green >> 0) & 0xFC00);
		val |= ((blue >> 8) & 0x00FC);
		val |= 0x0000000; //ƩӥåȤOFF

#if 0
		DPRINTK("palette_cpu2[%d]=0x%08X  red=0x%X green=0x%X blue=0x%X\n"
			,regno,val,red,green,blue);
#endif
		fbi->palette2_cpu[regno] = val;
		ret = 0;
	}
	return ret;
}

//Update SAKAI 051003
static int
sonynavu_pxafb_setcolreg(u_int regno, u_int red, u_int green, u_int blue,
		   u_int trans, struct fb_info *info)
{
	struct sonynavu_pxafb_info *fbi = (struct sonynavu_pxafb_info *)info;
	int ret = 1;

	/*
	 * If greyscale is true, then we convert the RGB value
	 * to greyscale no mater what visual we are using.
	 */
	if (fbi->fb.var.grayscale)
		red = green = blue = (19595 * red + 38470 * green +
					7471 * blue) >> 16;

	switch (fbi->fb.fix.visual) {
        case FB_VISUAL_TRUECOLOR:
		/*
		 * 16-bit True Colour.  We encode the RGB value
		 * according to the RGB bitfield information.
		 */
		if (regno < 16) {
                        unsigned int val;
			u32 *pal = fbi->fb.pseudo_palette;

			val  = chan_to_field(red, &fbi->fb.var.red);
			val |= chan_to_field(green, &fbi->fb.var.green);
			val |= chan_to_field(blue, &fbi->fb.var.blue);

			pal[regno] = val;
			ret = 0;
		}
		break;
	case FB_VISUAL_PSEUDOCOLOR:
		ret = sonynavu_pxafb_setpalettereg(regno, red, green, blue, trans, info);
		break;
          
	}

	return ret;
}


/*
 *  sonynavu_pxafb_check_var():
 *    Get the video params out of 'var'. If a value doesn't fit, round it up,
 *    if it's too big, return -EINVAL.
 *
 *    Round up in the following order: bits_per_pixel, xres,
 *    yres, xres_virtual, yres_virtual, xoffset, yoffset, grayscale,
 *    bitfields, horizontal timing, vertical timing.
 */
//Update SAKAI 051003
static int sonynavu_pxafb_check_var(struct fb_var_screeninfo *var, struct fb_info *info)
{
	struct sonynavu_pxafb_info *fbi = (struct sonynavu_pxafb_info *)info;

	if (var->xres < MIN_XRES)
		var->xres = MIN_XRES;
	if (var->yres < MIN_YRES)
		var->yres = MIN_YRES;
	if (var->xres > fbi->max_xres)
		var->xres = fbi->max_xres;
	if (var->yres > fbi->max_yres)
		var->yres = fbi->max_yres;
	var->xres_virtual =
		max(var->xres_virtual, var->xres);
	var->yres_virtual =
		max(var->yres_virtual, var->yres);

        /*
	 * Setup the RGB parameters for this display.
	 *
	 * The pixel packing format is described on page 7-11 of the
	 * PXA2XX Developer's Manual.
         */
	if (var->bits_per_pixel == 16) {
		var->red.offset   = 10; var->red.length   = 5;
		var->green.offset = 5;  var->green.length = 5;
		var->blue.offset  = 0;  var->blue.length  = 5;
		var->transp.offset = 0; var->transp.length = 0;
	} else {
		var->red.offset = var->green.offset = var->blue.offset = var->transp.offset = 0;
		var->red.length   = 8;
		var->green.length = 8;
		var->blue.length  = 8;
		var->transp.length = 0;
	}

	return 0;
}


/*
 * sonynavu_pxafb_set_par():
 *	Set the user defined part of the display for the specified console
 */
//Update SAKAI 051003
static int sonynavu_pxafb_set_par(struct fb_info *info)
{
	struct sonynavu_pxafb_info *fbi = (struct sonynavu_pxafb_info *)info;
	struct fb_var_screeninfo *var = &info->var;
	unsigned int palette_mem_size;
	unsigned int frame_mem_size;

	DPRINTK("sonynavu_pxafb_set_par\n");

         if (var->bits_per_pixel == 16){
		fbi->fb.fix.visual = FB_VISUAL_TRUECOLOR;
         }else{
		/*
		 * Some people have weird ideas about wanting static
		 * pseudocolor maps.  I suspect their user space
		 * applications are broken.
		 */
		fbi->fb.fix.visual = FB_VISUAL_PSEUDOCOLOR;
	}

	fbi->fb.fix.line_length = var->xres_virtual *
				  var->bits_per_pixel / 8;

        if (var->bits_per_pixel == 16)
		fbi->palette2_size = 0;
	else
		fbi->palette2_size = var->bits_per_pixel == 1 ? 4 : 1 << var->bits_per_pixel;
	

	palette_mem_size = fbi->palette2_size * sizeof(u32);

	DPRINTK("palette_mem_size = 0x%08lx\n", (u_long) palette_mem_size);

	frame_mem_size=(LCD_XRES * LCD_YRES * LCD_BPP / 8);
	fbi->palette2_cpu 
			= (u32 *)&fbi->screen2_cpu[frame_mem_size];
		fbi->palette2_dma 
			= fbi->screen2_dma + frame_mem_size ;

	/*
	 * Set (any) board control register to handle new color depth
	 */



	if (fbi->fb.var.bits_per_pixel == 16)
		fb_dealloc_cmap(&fbi->fb.cmap);
	else
		fb_alloc_cmap(&fbi->fb.cmap, 1<<fbi->fb.var.bits_per_pixel, 0);

	sonynavu_pxafb_activate_var(var, fbi);

	return 0;
}

/*
 * Formal definition of the VESA spec:
 *  On
 *  	This refers to the state of the display when it is in full operation
 *  Stand-By
 *  	This defines an optional operating state of minimal power reduction with
 *  	the shortest recovery time
 *  Suspend
 *  	This refers to a level of power management in which substantial power
 *  	reduction is achieved by the display.  The display can have a longer
 *  	recovery time from this state than from the Stand-by state
 *  Off
 *  	This indicates that the display is consuming the lowest level of power
 *  	and is non-operational. Recovery from this state may optionally require
 *  	the user to manually power on the monitor
 *
 *  Now, the fbdev driver adds an additional state, (blank), where they
 *  turn off the video (maybe by colormap tricks), but don't mess with the
 *  video itself: think of it semantically between on and Stand-By.
 *
 *  So here's what we should do in our fbdev blank routine:
 *
 *  	VESA_NO_BLANKING (mode 0)	Video on,  front/back light on
 *  	VESA_VSYNC_SUSPEND (mode 1)  	Video on,  front/back light off
 *  	VESA_HSYNC_SUSPEND (mode 2)  	Video on,  front/back light off
 *  	VESA_POWERDOWN (mode 3)		Video off, front/back light off
 *
 *  This will match the matrox implementation.
 */

/*
 * sonynavu_pxafb_blank():
 *	Blank the display by setting all palette values to zero.  Note, the
 * 	16 bpp mode does not really use the palette, so this will not
 *      blank the display in all modes.
 */
static int sonynavu_pxafb_blank(int blank, struct fb_info *info)
{
	struct sonynavu_pxafb_info *fbi = (struct sonynavu_pxafb_info *)info;
	int i;

	DPRINTK("sonynavu_pxafb_blank: blank=%d\n", blank);
	switch (blank) {
	case FB_BLANK_POWERDOWN:
	case FB_BLANK_VSYNC_SUSPEND:
	case FB_BLANK_HSYNC_SUSPEND:
	case FB_BLANK_NORMAL:
		if (fbi->fb.fix.visual == FB_VISUAL_PSEUDOCOLOR ||
		    fbi->fb.fix.visual == FB_VISUAL_STATIC_PSEUDOCOLOR)
			for (i = 0; i < fbi->palette2_size; i++)
				sonynavu_pxafb_setpalettereg(i, 0, 0, 0, 0, info);

		sonynavu_pxafb_schedule_work(fbi, C_DISABLE);
		//TODO if (sonynavu_pxafb_blank_helper) sonynavu_pxafb_blank_helper(blank);
		break;

	case FB_BLANK_UNBLANK:
		//TODO if (sonynavu_pxafb_blank_helper) sonynavu_pxafb_blank_helper(blank);
		if (fbi->fb.fix.visual == FB_VISUAL_PSEUDOCOLOR ||
		    fbi->fb.fix.visual == FB_VISUAL_STATIC_PSEUDOCOLOR)
			fb_set_cmap(&fbi->fb.cmap, info);
		sonynavu_pxafb_schedule_work(fbi, C_ENABLE);
	}
	return 0;
}

//Update SAKAI 051003
static int sonynavu_pxafb_ioctl(struct inode *inode, struct file *file, 
		    unsigned int cmd, unsigned long arg,
		    struct fb_info *info)
{
	
    unsigned int	*parg;
	struct sonynavu_pxafb_info *fbi = (struct sonynavu_pxafb_info *)info;

    int	err = 0;
	
    parg = (unsigned int*)arg;
	DPRINTK("ioctl:%d\n",cmd);
    switch (cmd){

	case   FBIO_PXA270_DISP: //LCD Controller Enable
		{
			unsigned int mode;
			get_user(mode, (unsigned int*)  parg);
			switch(mode){
			case 0:
				//Disable
				DPRINTK("FBIO_PXA270_DISP:Disable.....\n");
				sonynavu_pxafb_disable_controller(fbi);
				DPRINTK("Done\n");
				fbi->LCDStatus=0;
				err=0;
				break;
			case 1:
				//Enable
				DPRINTK("FBIO_PXA270_DISP:Enable.....\n");
				sonynavu_pxafb_enable_controller(fbi);
				DPRINTK("Done\n");
				fbi->LCDStatus=1;
				err=1;
				break;
			default:
				break;
			}
			err=fbi->LCDStatus;
		}
		break; 
	////
	case   FBIO_PXA270_LAYER2:  //Overlay 2 Enable (Linux)
		//⤷ ӥǥץ㤬ưƤEץߤ Linux򤹤
		{
			unsigned int mode;
			get_user(mode, (unsigned int*)  parg);
			switch(mode){
			case 0:
				//Disable
				DPRINTK("FBIO_PXA270_LAYER2:Disable...\n");
				if(fbi->Overlay2_Console.Status!=0){
                   int done;
					/* clear O2EN */
					fbi->ovl2c1 &= 0x7FFFFFFF;/* Disable Overlay2 Controller */
					OVL2C1 = fbi->ovl2c1;
                                        CLEAR_LCD_INTR(LCSR1, LCSR1_BS2);
                                  
					DFBR2  = fbi->fdadr2 | 3; //Update DMA
				        DPRINTK("0x%08X\n",OVL2C1);
                                        fbi->Overlay2_Console.Status=0;
 					done = WAIT_FOR_LCD_INTR(LCSR1, LCSR1_BS2, 100);
                                        if (!done) {
                                          DPRINTK(KERN_INFO "%s: timeout\n", __FUNCTION__);
                                          break;
                                        }
                               }
				
				break;
			case 1:
				//Enable
                                //if(fbi->Overlay1.Status==0) break;
				if(fbi->Overlay2_Console.Status==0){
				      /* disable branch/start/end of frame interrupt */
                                      LCCR5 |= (LCCR5_IUM4 | LCCR5_IUM3 | LCCR5_IUM2 | 
                                                LCCR5_BSM4 | LCCR5_BSM3 | LCCR5_BSM2 |
                                                LCCR5_EOFM4 | LCCR5_EOFM3 | LCCR5_EOFM2 |
                                                LCCR5_SOFM4 | LCCR5_SOFM3 | LCCR5_SOFM2);
                                      fbi->ovl2c1 |= 0x80000000;/* Enable Overlay2 Controller */
				      OVL2C1 = fbi->ovl2c1;
				      FDADR2 = fbi->fdadr2 ; //Update DMA
                                      fbi->Overlay2_Console.Status=1;
				}
				
				break;
			default:
				break;
			}
			err=fbi->Overlay2_Console.Status;
		}
		break;

#if 0 //Overlay2ԲĤȤE060312 t-sakai
	case   FBIO_PXA270_SETDISP2: //Set Overlay2 Position (Linux or Video)
		{
			int x,y;
			get_user(x, (unsigned int*)  parg);
			get_user(y, (unsigned int*)  (parg+1));
			DPRINTK("FBIO_PXA270_SETDISP2:(%d,%d)\n",x,y);
			//˥ԡ
			fbi->Overlay2_Console.DispX			=x;
			fbi->Overlay2_Console.DispY			=y;
			//E򹹿
			fbi->ovl2c1	
				= (fbi->ovl2c1 & 0xFFF00000) 
				| ((fbi->Overlay2_Console.DispHeight-1)<<10) 
				| ((fbi->Overlay2_Console.DispWidth-1))  ;
			fbi->ovl2c2	
				= (fbi->ovl2c2 & 0xFFF00000) 
				| ((fbi->Overlay2_Console.DispY)<<10) 
				| ((fbi->Overlay2_Console.DispX));
			OVL2C1=fbi->ovl2c1;
			DPRINTK("0x%08X\n",OVL2C1);
			OVL2C2=fbi->ovl2c2;
			DPRINTK("0x%08X\n",OVL2C2);
			DFBR2=fbi->fdadr2 | 0x1; //Update DMA
		}
		break;
#endif //060312 t-sakai
      case   FBIO_PXA270_GETDISP2: //Get Overlay2 Position (Linux or Video)
		{
			PXA270_DISPSETTING2 *disp_setting,disp_setting1;
 			disp_setting1.DispX=fbi->Overlay2_Console.DispX;
			disp_setting1.DispY=fbi->Overlay2_Console.DispY;
			disp_setting1.DispWidth=fbi->Overlay2_Console.DispWidth;
			disp_setting1.DispHeight=fbi->Overlay2_Console.DispHeight;
			DPRINTK("FBIO_PXA270_GETDISP2:\n");
			get_user(disp_setting, (PXA270_DISPSETTING2**)parg);
			copy_to_user(disp_setting,&disp_setting1,sizeof(PXA270_DISPSETTING2));
		}
		break;
	////
	case   FBIO_PXA270_GETBASE2: //Get Overlay2 BaseAddress
		DPRINTK("FBIO_PXA270_GETBASE2:0x%08X\n",OFFSET_OVERLAY2);
		return(OFFSET_OVERLAY2);
		break;
	////
	case   FBIO_PXA270_LAYER1:  //Overlay 1 Enable (GUI)
		{
			int mode;
			get_user(mode, (unsigned int*)  parg);
			switch(mode){
			case 0:
				//Disable
				DPRINTK("FBIO_PXA270_LAYER1:Disable...\n");
                                /* disable branch/start/end of frame interrupt */
				if(fbi->Overlay1.Status!=0){
                                        int done;
                                        /* clear O1EN */
	                                fbi->ovl1c1 &= 0x7FFFFFFF;/* Disable Overlay1 Controller */
					OVL1C1 = fbi->ovl1c1;
                                        CLEAR_LCD_INTR(LCSR1, LCSR1_BS1);
	                                DFBR1 = fbi->fdadr1 | 0x3;
				        fbi->Overlay1.Status=0;
				        DPRINTK("0x%08X\n",fbi->ovl1c1);
                                        
	                                done = WAIT_FOR_LCD_INTR(LCSR1, LCSR1_BS1, 100);
                                        if(!done){
                                           DPRINTK(KERN_INFO "%s: timeout\n", __FUNCTION__);
                                        }
                                        sonynavu_pxafb_enable_controller(fbi);
				}
				break;
			case 1:
				//Enable
                                //if(fbi->Overlay2_Console.Status!=0) break;
				if(fbi->Overlay1.Status==0){
                  fbi->lccr5 |= (LCCR5_IUM1 | LCCR5_BSM1 | LCCR5_EOFM1 | LCCR5_SOFM1);
                  LCCR5 = fbi->lccr5;
                  FDADR1 = fbi->fdadr1;
                  fbi->ovl1c1 |= 0x80000000;
                  OVL1C1 = fbi->ovl1c1 ;
                  fbi->Overlay1.Status=1;
                  sonynavu_pxafb_enable_controller(fbi);
				}
				DPRINTK("FBIO_PXA270_LAYER1:Enable 0x%08X\n",OVL1C1);
				break;
			}
			err=fbi->Overlay1.Status;
		}
		break;
	case   FBIO_PXA270_SETPAL1: //Set Overlay1 Palette (GUI)
		{
			unsigned int num,start,*palette;
			get_user(num, (unsigned int*)  parg);
			get_user(start, (unsigned int*)  (parg+1));
			get_user(palette, (unsigned int**)(parg+2));
			DPRINTK("FBIO_PXA270_SETPAL1:%u %u\n",num,start);
			if((num+start)<=256){
				//copy palette to buffer...
				copy_from_user(&fbi->palette1_cpu[start],palette,sizeof(unsigned int)*num);
			}
		}
		break;
	case   FBIO_PXA270_SELBUF1: //Select Overlay1 Buffer (GUI)
		{
			unsigned int mode;
			get_user(mode, (unsigned int*)  parg);
			DPRINTK("FBIO_PXA270_SELBUF1:%u \n",mode);
			switch(mode){
			case 0:
			case 1:
            case 2: //Add 060312 t-sakai
				local_irq_disable();
				fbi->RequestFlip=mode;
				local_irq_enable();
				break;
			default:
				if(fbi->Overlay1.Status==0){
					err=0xFFFFFFFF;
				}else{
					err=fbi->Overlay1.Flip;
				}
                err=fbi->Overlay1.Flip;
                err=0;
			}
		}
		break;
	case   FBIO_PXA270_SETTRANS1: //Multiplication Constant for Half Transparency (GUI)
		{
			unsigned int mode;
			get_user(mode, (unsigned int*)  parg);
			DPRINTK("FBIO_PXA270_SETTRANS1:%u \n",mode);
			if(mode<256){
				err=mode;
				mode=(mode >>5) & 0x7;
				fbi->lccr4=(LCCR4 & 0xFFFFFE00) | ( mode <<6 ) | ( mode <<3) | ( mode);
				LCCR4=fbi->lccr4;
			}else{
				err=(fbi->lccr4 & 0x7)<<5;
                                if(err==(0x7<<5)) err=255;
			}
		}
		break;
#if 0 //Overlay1ɽѹԲĤȤE060312 t-sakai
	case   FBIO_PXA270_SETDISP1: //Set Overlay1 Position (GUI)
		{
			int DispX,DispY,DispWidth,DispHeight,FrameX,FrameY;

			get_user(DispX,			(unsigned int*)  parg);
			get_user(DispY,			(unsigned int*)  (parg+1));
			get_user(DispWidth,		(unsigned int*)  (parg+2));
			get_user(DispHeight,	(unsigned int*)  (parg+3));
			get_user(FrameX,		(unsigned int*)  (parg+4));
			get_user(FrameY,		(unsigned int*)  (parg+5));

			DPRINTK("FBIO_PXA270_SETDISP1:%d %d %d %d %d %d\n"
				,DispX,DispY,DispWidth,DispHeight,FrameX,FrameY);
			//˥ԡ
			fbi->Overlay1.DispX			=DispX;
			fbi->Overlay1.DispY			=DispY;
			fbi->Overlay1.DispWidth		=DispWidth;
			fbi->Overlay1.DispHeight	=DispHeight;	
			fbi->Overlay1.FrameX		=FrameX;
                        fbi->Overlay1.FrameY		=FrameY;	
			//E򹹿
			fbi->ovl1c1	
				= (fbi->ovl1c1 & 0xFFF00000) 
				| ((fbi->Overlay1.DispHeight-1)<<10) 
				| ((fbi->Overlay1.DispWidth-1))  ;
			fbi->ovl1c2	
				= (fbi->ovl1c2 & 0xFFF00000) 
				| ((fbi->Overlay1.DispY)<<10) 
				| ((fbi->Overlay1.DispX));
			OVL1C1=fbi->ovl1c1;
			OVL1C2=fbi->ovl1c2;
			DFBR1=fbi->fdadr1 | 0x1;
		}
		break;
#endif //060312 t-sakai
      
	case   FBIO_PXA270_GETDISP1: //Get Overlay1 Position (GUI)
		{
			PXA270_DISPSETTING1 *disp_setting,disp_setting1;
			disp_setting1.DispX=fbi->Overlay1.DispX;
			disp_setting1.DispY=fbi->Overlay1.DispY;
			disp_setting1.DispWidth=fbi->Overlay1.DispWidth;
			disp_setting1.DispHeight=fbi->Overlay1.DispHeight;
			disp_setting1.FrameX=fbi->Overlay1.FrameX;
			disp_setting1.FrameY=fbi->Overlay1.FrameY;			
			DPRINTK("FBIO_PXA270_GETDISP1:\n");
			get_user(disp_setting, (PXA270_DISPSETTING1**)parg);
			copy_to_user(disp_setting,&disp_setting1,sizeof(PXA270_DISPSETTING1));
		}
		break;
	case   FBIO_PXA270_GETBASE1: //Get Overlay1 BaseAddress
		{
			unsigned int mode;

			get_user(mode, (unsigned int*)  parg);
			switch(mode){
			case 0:
				DPRINTK("FBIO_PXA270_GETBASE1:0x%08X\n",OFFSET_OVERLAY1_0);
				return(OFFSET_OVERLAY1_0);
			case 1:
				DPRINTK("FBIO_PXA270_GETBASE1:0x%08X\n",OFFSET_OVERLAY1_1);
				return(OFFSET_OVERLAY1_1);
                        case 2:
                               DPRINTK("FBIO_PXA270_GETBASE1:0x%08X\n",OFFSET_OVERLAY2);
				return(OFFSET_OVERLAY2);
			default :
				return(-1);
			}
		}
		
		break;

	////   
	case   FBIO_PXA270_SETPAL0:    //Set BaseFrame Palette (Map)
		{
			unsigned int num,start,*palette;
			int i,j;
			get_user(num, (unsigned int*)  parg);
			get_user(start, (unsigned int*)  (parg+1));
			get_user(palette, (unsigned int**)(parg+2));
			DPRINTK("FBIO_PXA270_SETPAL0:%u %u\n",num,start);
#if 1
			for(j=0,i=start;j<num && i<256;i++,j++){
				unsigned int a;
				copy_from_user(&a,&palette[j],sizeof(unsigned int));
				DPRINTK("FBIO_PXA270_SETPAL0:%u = 0x%08X\n",i,a);
				if(a&0x0F000000){
					fbi->palette0_work[i]= a & 0x0FFFFFFF;
				}else{
					fbi->palette0_work[i]= a & 0x00FFFFFF;
					fbi->palette0_cpu[i] = a & 0x00FFFFFF;
				}
			}
#else
			if((num+start)<=256){
				//copy palette to buffer...
				copy_from_user(&fbi->palette0_cpu[start],palette,sizeof(unsigned int)*num);
			}
#endif
		}
		break;
	case   FBIO_PXA270_GETPAL0:    //Set BaseFrame Palette (Map)
		{
			unsigned int start;
			get_user(start, (unsigned int*)  parg);
			DPRINTK("FBIO_PXA270_SETPAL0: %u\n",start);
			if((start)<=255){
				//copy palette to buffer...
				err=fbi->palette0_cpu[start];
			}else{
				err=0xFFFFFFFF;
			}
		}
		break;
	case   FBIO_PXA270_SETSCROLL0: //Set BaseFrame Position (Map)
		{
			int FrameX,FrameY,SpeedFrameX,SpeedFrameY;

			get_user(FrameX,		(unsigned int*)  (parg+0));
			get_user(FrameY,		(unsigned int*)  (parg+1));
			get_user(SpeedFrameX,	(unsigned int*)  (parg+2));
			get_user(SpeedFrameY,	(unsigned int*)  (parg+3));
			DPRINTK("FBIO_PXA270_SETSCROLL0:%d %d %d %d\n"
				,FrameX,FrameY,SpeedFrameX,SpeedFrameY);
			//˥ԡ
			local_irq_disable();
			fbi->ScrollTargetX		=FrameX;
			fbi->ScrollTargetY		=FrameY;
			fbi->ScrollSpeedX		=SpeedFrameX;
			fbi->ScrollSpeedY		=SpeedFrameY;
			fbi->ScrollStatus = 0x80000000;
			local_irq_enable();
		}
		break;
	case   FBIO_PXA270_SETSCROLL_FLIP0: //Set BaseFrame Position (Map)
		{
			int FrameX,FrameY,SpeedFrameX,SpeedFrameY;
			unsigned int mode;
			int StartX,StartY;

			get_user(FrameX,		(unsigned int*)  (parg+0));
			get_user(FrameY,		(unsigned int*)  (parg+1));
			get_user(SpeedFrameX,	(unsigned int*)  (parg+2));
			get_user(SpeedFrameY,	(unsigned int*)  (parg+3));
			get_user(StartX,		(unsigned int*)  (parg+4));
			get_user(StartY,		(unsigned int*)  (parg+5));
			get_user(mode,			(unsigned int*)  (parg+6));

			DPRINTK("FBIO_PXA270_SETSCROLL0:%d %d -> %d %d Speed:%d %d Mode:%d\n"
				,StartX,StartY,FrameX,FrameY,SpeedFrameX,SpeedFrameY,mode);
			//˥ԡ
			local_irq_disable();
			fbi->ScrollTargetX		=FrameX;
			fbi->ScrollTargetY		=FrameY;
			fbi->ScrollSpeedX		=SpeedFrameX;
			fbi->ScrollSpeedY		=SpeedFrameY;
			if(StartX!=0x80000000 
				|| StartY!=0x80000000){
				fbi->Baseframe.FrameX=StartX;
				fbi->Baseframe.FrameY=StartY;
			}
			switch(mode){
			case 0:
			case 1:
				fbi->ScrollFlip=mode;
				break;
			}
			fbi->ScrollStatus = 0x80000000;
			local_irq_enable();
		}
		break;
	case   FBIO_PXA270_GETSCROLL0: //Get BaseFrame Position (Map)
		{
			PXA270_GETSCROLL0 *disp_setting,disp_setting0;

			disp_setting0.FrameX=fbi->Baseframe.FrameX;
			disp_setting0.FrameY=fbi->Baseframe.FrameY;			
			disp_setting0.Flag=fbi->ScrollStatus;

			DPRINTK("FBIO_PXA270_GETSCROLL0:%d %d %d \n"
				,disp_setting0.FrameX,disp_setting0.FrameY,disp_setting0.Flag);
			get_user(disp_setting, (PXA270_GETSCROLL0**)parg);
			copy_to_user(disp_setting,&disp_setting0,sizeof(PXA270_GETSCROLL0));
		}
		break;
	case   FBIO_PXA270_SELBUF0:    //Select BaseFrame Buffer (Map)
		{
			unsigned int mode;
			get_user(mode, (unsigned int*)  parg);
			DPRINTK("FBIO_PXA270_SELBUF0:%d \n",mode);
			switch(mode){
			case 0:
			case 1:
				local_irq_disable();
				fbi->ScrollFlip=mode;
				fbi->ScrollStatus =0x80000000;
				local_irq_enable();
				break;
			default:
				err=fbi->Baseframe.Flip;
			}
		}
		break;
	case   FBIO_PXA270_SETFRAME0:  //Set BaseFrame Buffer Width & Heghit (Map)
		{
			int FrameWidth,FrameHeight;
			get_user(FrameWidth,		(unsigned int*)  (parg+0));
			get_user(FrameHeight,		(unsigned int*)  (parg+1));
			if(FrameWidth<LCD_BASE_MIN_XRES 
				|| FrameWidth>LCD_BASE_MAX_XRES
				|| FrameHeight<LCD_BASE_MIN_YRES
				|| FrameHeight>LCD_BASE_MAX_YRES){
				err=1;
			}
			fbi->Baseframe.FrameWidth=FrameWidth;
			fbi->Baseframe.FrameHeight=FrameHeight;
			fbi->Baseframe.FrameX=0;
			fbi->Baseframe.FrameY=0;
			DPRINTK("FBIO_PXA270_SETFRAME0:%d %d\n",FrameWidth,FrameHeight);
		}

		break;
	case   FBIO_PXA270_GETFRAME0:  //Get BaseFrame Buffer Width & Heghit (Map)
		{
			PXA270_DISPSETTING0 *disp_setting,disp_setting0;

			disp_setting0.FrameWidth=fbi->Baseframe.FrameWidth;
			disp_setting0.FrameHeight=fbi->Baseframe.FrameHeight;			
			get_user(disp_setting, (PXA270_DISPSETTING0**)parg);
			DPRINTK("FBIO_PXA270_GETFRAME0:%d %d \n"
				,disp_setting0.FrameWidth,disp_setting0.FrameHeight);
			copy_to_user(disp_setting,&disp_setting0,sizeof(PXA270_DISPSETTING0));
		}
		break;
	case   FBIO_PXA270_GETBASE0: //Get BaseFrame BaseAddress
		{
			unsigned int mode;

			get_user(mode, (unsigned int*)  parg);
			switch(mode){
			case 0:
				DPRINTK("FBIO_PXA270_GETBASE1:0x%08X\n",OFFSET_BASELAYER_0);
				return(OFFSET_BASELAYER_0);
			case 1:
				DPRINTK("FBIO_PXA270_GETBASE1:0x%08X\n",OFFSET_BASELAYER_1);
				return(OFFSET_BASELAYER_1);
			default :
				return(-1);
			}
		}
		break;
	///
	case   FBIO_PXA270_SETPALCRS:  //Set Cursor Palette (Cursor)
		{
			unsigned int num,start,*palette;

			get_user(num, (unsigned int*)  parg);
			get_user(start, (unsigned int*)  (parg+1));
			get_user(palette, (unsigned int**)(parg+2));
			DPRINTK("FBIO_PXA270_SETPALCRS:%u %u\n",num,start);
			if((num+start)<=4){
				//copy palette to buffer...
				copy_from_user(&fbi->palette3_cpu[start],palette,sizeof(unsigned int)*num);
			}
		}
		break;
	case   FBIO_PXA270_DISPPOSCRS: //Set Cursor Position (Cursor)
		{
			int x,y;
			get_user(x, (unsigned int*)  parg);
			get_user(y, (unsigned int*)  (parg+1));
			if(x<0 || x>479 || y<0 || y>271) break;
			fbi->cursor=((CURSOR & 0xFE00001F) | (y<<15) | ( x << 5 ));
			CURSOR = fbi->cursor;
			DFBR5=fbi->fdadr5 | 0x1;
			DPRINTK("FBIO_PXA270_DISPPOSCRS: 0x%08X (%d,%d)\n",CURSOR,x,y);
		}
		break; 
	case   FBIO_PXA270_SELCRS: //Select Cursor Pattern (Cursor)
		{
			unsigned int mode;
			get_user(mode, (unsigned int*)  parg);
			DPRINTK("FBIO_PXA270_SELBUF0:%d \n",mode);
			if(fbi->Cursor.Status==0){
				err=-1;
			}else{
				switch(mode){
				case 0:
				case 1:
				case 2:
				case 3:
					SetDMASetting2_Cursor(fbi,mode);
					break;
				default:
					err=fbi->Cursor.Flip;
				}
			}
		}
		break;
	case   FBIO_PXA270_CURSOR:     //Cursor Enable (Cursor)
		{
			unsigned int mode;
			get_user(mode, (unsigned int*)  parg);
			switch(mode){
			case 0://Disable
                                if(fbi->Cursor.Status!=0){
				   DPRINTK("FBIO_PXA270_CURSOR:Disable...\n");
                                   fbi->Cursor.Status	=0;
				   fbi->cursor &= 0x7FFFFFFF;/* Disable Overlay2 Controller */
                                    CURSOR=fbi->cursor;
                                   CLEAR_LCD_INTR(LCSR1, LCSR1_BS5);
                                   DFBR5 = fbi->fdadr5 | 3; //Cursor

                                  {
                                      int done = WAIT_FOR_LCD_INTR(LCSR1, LCSR1_BS5, 100);
                                      if (!done)  break;
                                   }
				   
				   DPRINTK("0x%08X\n",CURSOR);
                                }
				break;
			case 1://Enable
                                if(fbi->Cursor.Status==0){
				  fbi->Cursor.Status =1;
                                  /* disable branch/start/end of frame interrupt */
                                  fbi->lccr5 |=  (LCCR5_IUM5 | LCCR5_BSM5 | LCCR5_EOFM5 | LCCR5_SOFM5);
	                          LCCR5 =fbi->lccr5;
				  FDADR5 = fbi->fdadr5; //Cursor
                                  DFBR5 = fbi->fdadr5 | 1; //Cursor
                                  fbi->cursor |= 0x80000000;
				  CURSOR=fbi->cursor;
				  DPRINTK("FBIO_PXA270_CURSOR:Enable\n");
                                }
                                break;
			default:
				break;
			}
			err=fbi->Cursor.Status;
		}
		break;
	case   FBIO_PXA270_GETBASECURSOR: //Get Cursor BaseAddress
		{
			unsigned int mode;

			get_user(mode, (unsigned int*)  parg);
			switch(mode){
			case 0:
				DPRINTK("FBIO_PXA270_GETBASECURSOR 0:0x%08X\n",OFFSET_CURSOR_0);
				return(OFFSET_CURSOR_0);
			case 1:
				DPRINTK("FBIO_PXA270_GETBASECURSOR 1:0x%08X\n",OFFSET_CURSOR_1);
				return(OFFSET_CURSOR_1);
			case 2:
				DPRINTK("FBIO_PXA270_GETBASECURSOR 0:0x%08X\n",OFFSET_CURSOR_2);
				return(OFFSET_CURSOR_2);
			case 3:
				DPRINTK("FBIO_PXA270_GETBASECURSOR 0:0x%08X\n",OFFSET_CURSOR_3);
				return(OFFSET_CURSOR_3);
			default :
				return(-1);
			}
		}
		break;
	case   FBIO_PXA270_WAITVSYNC:
		{
			DPRINTK("FBIO_PXA270_WAITVSYNC:\n");
			atomic_inc(&fbi->vsyncinfo.waitcnt);
			interruptible_sleep_on(&fbi->vsyncinfo.waitq);
			atomic_dec(&fbi->vsyncinfo.waitcnt);
		}
		break;
	case   FBIO_PXA270_WAITSTARTFRAME:
		{
			DPRINTK("FBIO_PXA270_WAITSTARTFRAME:\n");
			sonynavu_pxafb_wait_startframe();
		}
		break;
#ifdef USE_CAPTURE
	case   FBIO_PXA270_STARTCAPTURE:
		{
			int DispX,DispY;
			get_user(DispX,			(unsigned int*)  parg);
			get_user(DispY,			(unsigned int*)  (parg+1));
			DPRINTK("FBIO_PXA270_STARTCAPTURE %d %d \n",DispX,DispY);
			//⤷Overlay2Linux󥽡EɽʤOFFν򤹤E
			if(fbi->Overlay2_Console.Status!=0){
				fbi->ovl2c1 &= 0x7FFFFFFF;/* Disable Overlay2 Controller */
				OVL2C1 = fbi->ovl2c1;
				DFBR2  = fbi->fdadr2 | 0x1; //Update DMA
				fbi->Overlay2_Console.Status=0;
			}
			
			fbi->Overlay2_Video.Status=1;
		}
		break;
	case   FBIO_PXA270_STOPCAPTURE:
		{
			DPRINTK("FBIO_PXA270_STOPCAPTURE:\n");
			if(fbi->Overlay2_Video.Status==0) break;
			//Overlay2OFFˤ
			fbi->ovl2c1 &= 0x7FFFFFFF;/* Disable Overlay2 Controller */
			OVL2C1 = fbi->ovl2c1;
			DFBR2  = fbi->fdadr2 | 0x1; //Update DMA
			fbi->Overlay2_Video.Status=0;
		}		
		break;
	case   FBIO_PXA270_GETFRAMEVIDEO:  //Get Video Size
		{
			PXA270_FRAMESETTINGVIDEO *disp_setting,disp_setting0;

			disp_setting0.FrameWidth=fbi->Overlay2_Video.FrameWidth;
			disp_setting0.FrameHeight=fbi->Overlay2_Video.FrameHeight;			
			get_user((int)disp_setting, (unsigned int*)  (parg));
			DPRINTK("FBIO_PXA270_GETFRAME0:%d %d \n"
				,disp_setting0.FrameWidth,disp_setting0.FrameHeight);
			copy_to_user(disp_setting,&disp_setting0,sizeof(PXA270_FRAMESETTINGVIDEO));
		}
		break;
	case   FBIO_PXA270_GETDISPVIDEO:
		{
			PXA270_DISPSETTINGVIDEO *disp_setting,disp_setting0;
			disp_setting0.DispX=fbi->Overlay2_Video.DispX;
			disp_setting0.DispY=fbi->Overlay2_Video.DispY;			
			disp_setting0.DispWidth=fbi->Overlay2_Video.DispWidth;
			disp_setting0.DispHeight=fbi->Overlay2_Video.DispHeight;			
			get_user((int)disp_setting, (unsigned int*)  (parg));
			DPRINTK("FBIO_PXA270_GETDISPVIDEO:%d %d \n"
				,disp_setting0.FrameWidth,disp_setting0.FrameHeight);
			copy_to_user(disp_setting,&disp_setting0,sizeof(PXA270_DISPSETTINGVIDEO));
			if(fbi->Overlay2_Video.Status==0){
				err=0;
			}else{
				err=1;
			}
		}
		break;

	case FBIO_PXA270_GETBASEVIDEO_Y:
		DPRINTK("FBIO_PXA270_GETBASEVIDEO_Y:0x%08X\n"
			,OFFSET_VIDEO);
		return(OFFSET_VIDEO);
		break;
	case FBIO_PXA270_GETBASEVIDEO_B:
		DPRINTK("FBIO_PXA270_GETBASEVIDEO_B:0x%08X\n"
			,OFFSET_VIDEO+(VIDEO_XRES * VIDEO_YRES));
		return(OFFSET_VIDEO+(VIDEO_XRES * VIDEO_YRES));
		break;
	case FBIO_PXA270_GETBASEVIDEO_R:
		DPRINTK("FBIO_PXA270_GETBASEVIDEO_R:0x%08X\n"
			,OFFSET_VIDEO+(VIDEO_XRES * VIDEO_YRES)+(VIDEO_XRES * VIDEO_YRES/2));
		return(OFFSET_VIDEO+(VIDEO_XRES * VIDEO_YRES)+(VIDEO_XRES * VIDEO_YRES/2));
		break;

#endif
    default:
		DPRINTK("Unknown IOCTL Command\n");
		return -ENOTTY;
	}
    return(err);

}

static int sonynavu_pxafb_mmap(struct fb_info *info, struct file *file,
		      struct vm_area_struct *vma)
{
	struct sonynavu_pxafb_info *fbi = (struct sonynavu_pxafb_info *)info;
	

#if 0
	   DPRINTK("vm_start = %x\n", (int)vma->vm_start);
	   DPRINTK("vm_end = %x\n", (int)vma->vm_end);
	   DPRINTK("vm_pgoff = %x\n", (int)vma->vm_pgoff);
	   DPRINTK("PAGE_OFFSET = %x\n", (int)PAGE_OFFSET);
		DPRINTK("off=%d len=%d\n",off,info->fix.smem_len);
#endif

	{
		unsigned int len;
		unsigned int start=0, off;

		if (vma->vm_pgoff > (~0UL >> PAGE_SHIFT)) {
			  return -EINVAL;
		}

		start = fbi->map_dma & PAGE_MASK;
		len = PAGE_ALIGN((start & ~PAGE_MASK) + fbi->map_size);

		off = vma->vm_pgoff << PAGE_SHIFT;

		if ((vma->vm_end - vma->vm_start + off) > len) {
			 return -EINVAL;
		}

		off += start;
		vma->vm_pgoff = off >> PAGE_SHIFT;

		//pgprot_val(vma->vm_page_prot) &= ~_CACHE_MASK;
		//pgprot_val(vma->vm_page_prot) |= _CACHE_CACHABLE_NONCOHERENT;
		//pgprot_val(vma->vm_page_prot) |= (6 << 9); //CCA=6

		/* This is an IO map - tell maydump to skip this VMA */
		vma->vm_flags |= VM_IO;

		if (io_remap_pfn_range(vma, vma->vm_start, off >> PAGE_SHIFT,
									vma->vm_end - vma->vm_start,
									vma->vm_page_prot)) {
		   return -EAGAIN;
		 }
	}
    return 0;

}

static struct fb_ops sonynavu_pxafb_ops = {
	.owner		= THIS_MODULE,
	.fb_check_var	= sonynavu_pxafb_check_var,
	.fb_set_par	= sonynavu_pxafb_set_par,
	.fb_setcolreg	= sonynavu_pxafb_setcolreg,
	.fb_fillrect	= cfb_fillrect,
	.fb_copyarea	= cfb_copyarea,
	.fb_imageblit	= cfb_imageblit,
	.fb_blank	= sonynavu_pxafb_blank,
	.fb_ioctl   = sonynavu_pxafb_ioctl,
	.fb_mmap	= sonynavu_pxafb_mmap,
};

/*
 * Calculate the PCD value from the clock rate (in picoseconds).
 * We take account of the PPCR clock setting.
 * From PXA Developer's Manual:
 *
 *   PixelClock =      LCLK
 *                -------------
 *                2 ( PCD + 1 )
 *
 *   PCD =      LCLK
 *         ------------- - 1
 *         2(PixelClock)
 *
 * Where:
 *   LCLK = LCD/Memory Clock
 *   PCD = LCCR3[7:0]
 *
 * PixelClock here is in Hz while the pixclock argument given is the
 * period in picoseconds. Hence PixelClock = 1 / ( pixclock * 10^-12 )
 *
 * The function get_lclk_frequency_10khz returns LCLK in units of
 * 10khz. Calling the result of this function lclk gives us the
 * following
 *
 *    PCD = (lclk * 10^4 ) * ( pixclock * 10^-12 )
 *          -------------------------------------- - 1
 *                          2
 *
 * Factoring the 10^4 and 10^-12 out gives 10^-8 == 1 / 100000000 as used below.
 */
static inline unsigned int get_pcd(unsigned int pixclock)
{
	unsigned int pcd;

	/* FIXME: Need to take into account Double Pixel Clock mode
         * (DPC) bit? or perhaps set it based on the various clock
         * speeds */

	pcd = (unsigned int )get_lcdclk_frequency_10khz() * pixclock;
	do_div(pcd, 100000000 * 2);
	/* no need for this, since we should subtract 1 anyway. they cancel */
	/* pcd += 1; */ /* make up for integer math truncations */
	return (unsigned int)pcd;
}

/*
 * sonynavu_pxafb_activate_var():
 *	Configures LCD Controller based on entries in var parameter.  Settings are
 *	only written to the controller if changes were made.
 */
//Update SAKAI 051003
static int sonynavu_pxafb_activate_var(struct fb_var_screeninfo *var, struct sonynavu_pxafb_info *fbi)
{
//	u_int pcd = get_pcd(var->pixclock);
	u_long flags;

	DPRINTK("Configuring PXA LCD\n");

	DPRINTK("var: xres=%d hslen=%d lm=%d rm=%d\n",
		var->xres, var->hsync_len,
		var->left_margin, var->right_margin);
	DPRINTK("var: yres=%d vslen=%d um=%d bm=%d\n",
		var->yres, var->vsync_len,
		var->upper_margin, var->lower_margin);

#if DEBUG_VAR
	if (var->xres < 16        || var->xres > 1024)
		printk(KERN_ERR "%s: invalid xres %d\n",
			fbi->fb.fix.id, var->xres);
	if (var->hsync_len < 1    || var->hsync_len > 64)
		printk(KERN_ERR "%s: invalid hsync_len %d\n",
			fbi->fb.fix.id, var->hsync_len);
	if (var->left_margin < 1  || var->left_margin > 255)
		printk(KERN_ERR "%s: invalid left_margin %d\n",
			fbi->fb.fix.id, var->left_margin);
	if (var->right_margin < 1 || var->right_margin > 255)
		printk(KERN_ERR "%s: invalid right_margin %d\n",
			fbi->fb.fix.id, var->right_margin);
	if (var->yres < 1         || var->yres > 1024)
		printk(KERN_ERR "%s: invalid yres %d\n",
			fbi->fb.fix.id, var->yres);
	if (var->vsync_len < 1    || var->vsync_len > 64)
		printk(KERN_ERR "%s: invalid vsync_len %d\n",
			fbi->fb.fix.id, var->vsync_len);
	if (var->upper_margin < 0 || var->upper_margin > 255)
		printk(KERN_ERR "%s: invalid upper_margin %d\n",
			fbi->fb.fix.id, var->upper_margin);
	if (var->lower_margin < 0 || var->lower_margin > 255)
		printk(KERN_ERR "%s: invalid lower_margin %d\n",
			fbi->fb.fix.id, var->lower_margin);
#endif



	DPRINTK("nlccr0 = 0x%08x\n", fbi->lccr0);
	DPRINTK("nlccr1 = 0x%08x\n", fbi->lccr1);
	DPRINTK("nlccr2 = 0x%08x\n", fbi->lccr2);
	DPRINTK("nlccr3 = 0x%08x\n", fbi->lccr3);
	DPRINTK("nlccr4 = 0x%08x\n", fbi->lccr4);

	/* Update shadow copy atomically */
	local_irq_save(flags);

	/* setup dma descriptors */
	//Overlay2 Video
	fbi->dmadesc_fbvideoY_cpu 
		= (struct sonynavu_pxafb_dma_descriptor *)((unsigned int)fbi->map_cpu2+DMA_DESC_OFFSET_VIDEO_Y);
	fbi->dmadesc_fbvideoY_dma = fbi->map_dma2 + DMA_DESC_OFFSET_VIDEO_Y;
	fbi->dmadesc_fbvideoB_cpu 
		= (struct sonynavu_pxafb_dma_descriptor *)((unsigned int)fbi->map_cpu2+DMA_DESC_OFFSET_VIDEO_B);
	fbi->dmadesc_fbvideoB_dma = fbi->map_dma2 + DMA_DESC_OFFSET_VIDEO_B;
	fbi->dmadesc_fbvideoR_cpu 
		= (struct sonynavu_pxafb_dma_descriptor *)((unsigned int)fbi->map_cpu2+DMA_DESC_OFFSET_VIDEO_R);
	fbi->dmadesc_fbvideoR_dma = fbi->map_dma2 + DMA_DESC_OFFSET_VIDEO_R;

	//Overlay2 Linux
	fbi->dmadesc_palette2_cpu 
		= (struct sonynavu_pxafb_dma_descriptor *)((unsigned int)fbi->map_cpu2+DMA_DESC_OFFSET_FRAME2_PAL);
	fbi->dmadesc_fbover2_cpu 
		=  (struct sonynavu_pxafb_dma_descriptor *)((unsigned int)fbi->map_cpu2+DMA_DESC_OFFSET_FRAME2_FB0);
	fbi->dmadesc_palette2_dma = fbi->map_dma2 + DMA_DESC_OFFSET_FRAME2_PAL;
	fbi->dmadesc_fbover2_dma = fbi->map_dma2  + DMA_DESC_OFFSET_FRAME2_FB0;
	//Cursor
	fbi->dmadesc_fbcursor_cpu 
		=  (struct sonynavu_pxafb_dma_descriptor *)((unsigned int)fbi->map_cpu2+DMA_DESC_OFFSET_CURSOR_FB0);
	fbi->dmadesc_palette3_cpu 
		= (struct sonynavu_pxafb_dma_descriptor *)((unsigned int)fbi->map_cpu2+DMA_DESC_OFFSET_CURSOR_PAL);
	fbi->dmadesc_fbcursor_dma = fbi->map_dma2  + DMA_DESC_OFFSET_CURSOR_FB0;
	fbi->dmadesc_palette3_dma = fbi->map_dma2 + DMA_DESC_OFFSET_CURSOR_PAL;
	//Overlay1 GUI
	fbi->dmadesc_fbover1_cpu 
		=  (struct sonynavu_pxafb_dma_descriptor *)((unsigned int)fbi->map_cpu2+DMA_DESC_OFFSET_FRAME1_FB0);
	fbi->dmadesc_palette1_cpu 
		= (struct sonynavu_pxafb_dma_descriptor *)((unsigned int)fbi->map_cpu2+DMA_DESC_OFFSET_FRAME1_PAL);
	fbi->dmadesc_fbover1_dma = fbi->map_dma2  + DMA_DESC_OFFSET_FRAME1_FB0;
	fbi->dmadesc_palette1_dma = fbi->map_dma2 + DMA_DESC_OFFSET_FRAME1_PAL;
	//Base layer
	fbi->dmadesc_fbover0_cpu 
		=  (struct sonynavu_pxafb_dma_descriptor *)((unsigned int)fbi->map_cpu2+DMA_DESC_OFFSET_FRAME0_FB0);
	fbi->dmadesc_palette0_cpu 
		= (struct sonynavu_pxafb_dma_descriptor *)((unsigned int)fbi->map_cpu2+DMA_DESC_OFFSET_FRAME0_PAL);
	fbi->dmadesc_fbover0_dma = fbi->map_dma2  + DMA_DESC_OFFSET_FRAME0_FB0;
	fbi->dmadesc_palette0_dma = fbi->map_dma2 + DMA_DESC_OFFSET_FRAME0_PAL;
	
	/* populate descriptors */
	//Overlay2 Linux
	SetDMASetting2_Console(fbi,fbi->Overlay2_Console.Flip);
	//Base layer Map
	SetDMASetting0(fbi,fbi->Baseframe.Flip);
	//Overlay1 GUI
	SetDMASetting1_StartUp(fbi,fbi->Overlay1.Flip);
	//Overlay2 Video
	SetDMASetting2_Video(fbi,fbi->Overlay2_Video.Flip);
	//Cursor
	SetDMASetting2_Cursor(fbi,fbi->Cursor.Flip);
	/***************************************************************************************/

	DPRINTK("fbi->dmadesc_fbover0_cpu = 0x%x\n", (unsigned int)fbi->dmadesc_fbover0_cpu);
	DPRINTK("fbi->dmadesc_palette0_cpu = 0x%x\n", (unsigned int)fbi->dmadesc_palette0_cpu);

	DPRINTK("fbi->dmadesc_fbover0_dma = 0x%x\n", fbi->dmadesc_fbover0_dma);
	DPRINTK("fbi->dmadesc_palette0_dma = 0x%x\n", fbi->dmadesc_palette0_dma);

	DPRINTK("fbi->dmadesc_fbover0_cpu->fdadr = 0x%x\n", fbi->dmadesc_fbover0_cpu->fdadr);
	DPRINTK("fbi->dmadesc_fbover0_cpu->fsadr = 0x%x\n", fbi->dmadesc_fbover0_cpu->fsadr);
	DPRINTK("fbi->dmadesc_fbover0_cpu->ldcmd = 0x%x\n", fbi->dmadesc_fbover0_cpu->ldcmd);
	
	DPRINTK("fbi->dmadesc_palette0_cpu->fdadr = 0x%x\n", fbi->dmadesc_palette0_cpu->fdadr);
	DPRINTK("fbi->dmadesc_palette0_cpu->fsadr = 0x%x\n", fbi->dmadesc_palette0_cpu->fsadr);
	DPRINTK("fbi->dmadesc_palette0_cpu->ldcmd = 0x%x\n", fbi->dmadesc_palette0_cpu->ldcmd);	
//
	DPRINTK("fbi->dmadesc_fbover1_cpu = 0x%x\n", (unsigned int)fbi->dmadesc_fbover1_cpu);
	DPRINTK("fbi->dmadesc_palette1_cpu = 0x%x\n", (unsigned int)fbi->dmadesc_palette1_cpu);

	DPRINTK("fbi->dmadesc_fbover1_dma = 0x%x\n", fbi->dmadesc_fbover1_dma);
	DPRINTK("fbi->dmadesc_palette1_dma = 0x%x\n", fbi->dmadesc_palette1_dma);

	DPRINTK("fbi->dmadesc_fbover1_cpu->fdadr = 0x%x\n", fbi->dmadesc_fbover1_cpu->fdadr);
	DPRINTK("fbi->dmadesc_fbover1_cpu->fsadr = 0x%x\n", fbi->dmadesc_fbover1_cpu->fsadr);
	DPRINTK("fbi->dmadesc_fbover1_cpu->ldcmd = 0x%x\n", fbi->dmadesc_fbover1_cpu->ldcmd);
	
	DPRINTK("fbi->dmadesc_palette1_cpu->fdadr = 0x%x\n", fbi->dmadesc_palette1_cpu->fdadr);
	DPRINTK("fbi->dmadesc_palette1_cpu->fsadr = 0x%x\n", fbi->dmadesc_palette1_cpu->fsadr);
	DPRINTK("fbi->dmadesc_palette1_cpu->ldcmd = 0x%x\n", fbi->dmadesc_palette1_cpu->ldcmd);
//
	DPRINTK("fbi->dmadesc_fbover2_cpu = 0x%x\n", (unsigned int)fbi->dmadesc_fbover2_cpu);
	DPRINTK("fbi->dmadesc_palette2_cpu = 0x%x\n", (unsigned int)fbi->dmadesc_palette2_cpu);

	DPRINTK("fbi->dmadesc_fbover2_dma = 0x%x\n", fbi->dmadesc_fbover2_dma);
	DPRINTK("fbi->dmadesc_palette2_dma = 0x%x\n", fbi->dmadesc_palette2_dma);
	DPRINTK("fbi->dmadesc_fbover2_cpu->fdadr = 0x%x\n", fbi->dmadesc_fbover2_cpu->fdadr);
	DPRINTK("fbi->dmadesc_fbover2_cpu->fsadr = 0x%x\n", fbi->dmadesc_fbover2_cpu->fsadr);
	DPRINTK("fbi->dmadesc_fbover2_cpu->ldcmd = 0x%x\n", fbi->dmadesc_fbover2_cpu->ldcmd);
	
	DPRINTK("fbi->dmadesc_palette2_cpu->fdadr = 0x%x\n", fbi->dmadesc_palette2_cpu->fdadr);
	DPRINTK("fbi->dmadesc_palette2_cpu->fsadr = 0x%x\n", fbi->dmadesc_palette2_cpu->fsadr);
	DPRINTK("fbi->dmadesc_palette2_cpu->ldcmd = 0x%x\n", fbi->dmadesc_palette2_cpu->ldcmd);


	local_irq_restore(flags);

	/*
	 * Only update the registers if the controller is enabled
	 * and something has changed.
	 */
#if 1
	if ((LCCR0 != fbi->lccr0)       || (LCCR1 != fbi->lccr1) ||
	    (LCCR2 != fbi->lccr2)       || (LCCR3 != fbi->lccr3) ||
	    (FDADR0 != fbi->fdadr0) )
		sonynavu_pxafb_schedule_work(fbi, C_REENABLE);
#endif
	return 0;
}

/*
 * NOTE!  The following functions are purely helpers for set_ctrlr_state.
 * Do not call them directly; set_ctrlr_state does the correct serialisation
 * to ensure that things happen in the right way 100% of time time.
 *	-- rmk
 */
static inline void __sonynavu_pxafb_backlight_power(struct sonynavu_pxafb_info *fbi, int on)
{
	DPRINTK("backlight o%s\n", on ? "n" : "ff");


}

static inline void __sonynavu_pxafb_lcd_power(struct sonynavu_pxafb_info *fbi, int on)
{
	DPRINTK("LCD power o%s\n", on ? "n" : "ff");


}

//Update SAKAI 051206
static void sonynavu_pxafb_setup_gpio(struct sonynavu_pxafb_info *fbi)
{

	/*
	 * setup is based on type of panel supported
	 */

	DPRINTK("sonynavu_pxafb_setup_gpio\n");

#ifdef CONFIG_FB_SONYNAVU_GPIO

	/* Set Out - High for GPIO16 */
	GPSR0 = 0x10000;
	GPDR0 = 0x10000;
	pxa_gpio_mode(GPIO16_PWM0_MD);
	pxa_set_cken(CKEN0_PWM0, 1);
	PWM_CTRL0 = 0;
	PWM_PWDUTY0 = 0x3ff;
	PWM_PERVAL0 = 0x3ff;
	if(1){
		// Backlight PWM ٤H
		//BITSET(PWMDCR0, 10, 1, 1); // FD
		PWM_PWDUTY0 |= (1<<10);
	}

	/* Set Out - Alt2 for GPIO58 - 77 */
	GPDR1 |= 0xfc000000;
	GPDR2 |= 0x3fff;
	GAFR1_U |= 0xaaa00000;
	GAFR2_L |= 0x0aaaaaaa;
	/* Set Out - Alt2 for GPIO86 - 87 */
	GPDR2 |= 0x00c00000;
	GAFR2_U |= 0x0000a000;
#else

#endif

}

//Update SAKAI 051003
static void sonynavu_pxafb_enable_controller(struct sonynavu_pxafb_info *fbi)
{
	DPRINTK("Enabling LCD controller....");
	/* FIXME:PXA27X E17: Overlay1 is not enabled intermittently after re-enabling LCD. */
	down(&fcs_lcd_sem);
        
        OVL1C1 = 0;
	OVL1C2 = 0;
	OVL2C1 = 0;
	OVL2C2 = 0;
        CURSOR=0;

        LCCR5 = fbi->lccr5;
	LCCR4 = fbi->lccr4;
	LCCR3 = fbi->lccr3;
	LCCR2 = fbi->lccr2;
	LCCR1 = fbi->lccr1;
	LCCR0 = fbi->lccr0 & ~LCCR0_ENB;

	LCCR0 |= LCCR0_ENB;
	LCCR0 |= LCCR0_DIS;

        LCCR5 = fbi->lccr5;
	LCCR4 = fbi->lccr4;
        LCCR3 = fbi->lccr3;
	LCCR2 = fbi->lccr2;
	LCCR1 = fbi->lccr1;
	LCCR0 = fbi->lccr0 & ~LCCR0_ENB;

	LCCR0 |= LCCR0_ENB;

	FDADR0 = fbi->fdadr0;
	FDADR1 = fbi->fdadr1; //Overlay1
	FDADR2 = fbi->fdadr2; //Overlay2

	FDADR5 = fbi->fdadr5; //Cursor
	CURSOR = fbi->cursor;

	LCCR0 |= LCCR0_ENB;
  
	OVL1C1 = fbi->ovl1c1;
	OVL1C2 = fbi->ovl1c2;
	OVL2C1 = fbi->ovl2c1;
	OVL2C2 = fbi->ovl2c2;
        up(&fcs_lcd_sem);
  
	DPRINTK("Done\n");
	DPRINTK("FDADR0 = 0x%08x\n", (unsigned int)FDADR0);
	DPRINTK("FDADR1 = 0x%08x\n", (unsigned int)FDADR1);
	DPRINTK("FDADR2 = 0x%08x\n", (unsigned int)FDADR2);
	DPRINTK("FDADR5 = 0x%08x\n", (unsigned int)FDADR5);
	DPRINTK("LCCR0 = 0x%08x\n", (unsigned int)LCCR0);
	DPRINTK("LCCR1 = 0x%08x\n", (unsigned int)LCCR1);
	DPRINTK("LCCR2 = 0x%08x\n", (unsigned int)LCCR2);
	DPRINTK("LCCR3 = 0x%08x\n", (unsigned int)LCCR3);
	DPRINTK("LCCR4 = 0x%08x\n", (unsigned int)LCCR4);
	DPRINTK("OVL1C1 = 0x%08x\n", (unsigned int)OVL1C1);
	DPRINTK("OVL1C2 = 0x%08x\n", (unsigned int)OVL1C2);
	DPRINTK("CURSOR = 0x%08x\n", (unsigned int)CURSOR);
}

static void sonynavu_pxafb_disable_controller(struct sonynavu_pxafb_info *fbi)
{
	DPRINTK("Disabling LCD controller....\n");

	down(&fcs_lcd_sem);
#if 0
	add_wait_queue(&fbi->ctrlr_wait, &wait);
	set_current_state(TASK_UNINTERRUPTIBLE);
	LCSR = 0xffffffff;	/* Clear LCD Status Register */
	LCCR0 &= ~LCCR0_LDM;	/* Enable LCD Disable Done Interrupt */
	LCCR0 |= LCCR0_DIS;	/* Disable LCD Controller */
	schedule_timeout(20 * HZ / 1000);
	remove_wait_queue(&fbi->ctrlr_wait, &wait);
#else
	LCCR0 &= (~LCCR0_ENB);	/* Disable LCD Controller */
#endif
	DPRINTK("Done\n");
	up(&fcs_lcd_sem);
}
/*
 *  sonynavu_pxafb_handle_irq: Handle 'LCD DONE' interrupts.
 */
//Update SAKAI 051003

static unsigned int irqcount=0;
static unsigned int irqcount2=0;
static irqreturn_t sonynavu_pxafb_handle_irq(int irq, void *dev_id, struct pt_regs *regs)
{
	struct sonynavu_pxafb_info *fbi = dev_id;
	unsigned int lcsr = LCSR;
        unsigned int lcsr1 = LCSR1;
        int handled=0;
	disable_irq(IRQ_LCD);
	if (lcsr & LCSR_LDD) {
                handled=1;
		//printk("LCCR0_LDM\n");
		LCCR0 |= LCCR0_LDM;
		// wake_up(&fbi->ctrlr_wait); //E Hang up
	}
        if (lcsr & LCSR_SOF) {
            handled=1;
            //Wake up proccesses
	    if(atomic_read(&StartFrameInfo.waitcnt)){
		wake_up_interruptible(&StartFrameInfo.waitq);
	    }
	    irqcount2++;
	    
        }
	if (lcsr & LCSR_EOF) {
		int i;
                handled=1;
		//The interupt for End of Frame
		irqcount++;
		
		//Wake up proccesses
		if(atomic_read(&fbi->vsyncinfo.waitcnt)){
			wake_up_interruptible(&fbi->vsyncinfo.waitq);
		}
		//Update Palette
		if((irqcount & 0x7)==0){
			//each 100msec
			for(i=0;i<256;i++){
				unsigned int a=fbi->palette0_work[i];
				if(a & 0x0F000000){
					unsigned int b=fbi->palette0_cpu[i];
					if((a & 0xF0000000)==0){
					   fbi->palette0_work[i]
						   =((a&0x0F000000)<<4) | (a&0x0F000000) | (b&0xFFFFFF);
					   fbi->palette0_cpu[i]
						   =(a & 0xFFFFFF);
					}else{
					   fbi->palette0_work[i]=a-0x10000000;
					}
				}
			}
		}

#if 1
		//Flip GUI
		{
			if(fbi->Overlay1.Flip!=fbi->RequestFlip){
				DPRINTK("Flip UI %d\n",fbi->RequestFlip);
				SetDMASetting1(fbi,fbi->RequestFlip);
				//fbi->Overlay1.Flip=fbi->RequestFlip;
			}
		}
		//AutoScroll
		{   
			int ScrollX,ScrollY;
			int FrameWidth,FrameHeight;
			int WindowWidth,WindowHeight;
			int NextX,NextY;
			int flag;

			//EԡɤꤵEƤE եEåפΥEȤäE
		    if((fbi->ScrollSpeedX!=0 
				|| fbi->ScrollSpeedY!=0)   
				|| fbi->ScrollFlip!=fbi->Baseframe.Flip){
				fbi->ScrollStatus |= 0x08000000;
				DPRINTK("Auto\n");
				//Eθߤΰ֤
				ScrollX=fbi->Baseframe.FrameX;
				ScrollY=fbi->Baseframe.FrameY;
		    	//եE
		    	FrameWidth =fbi->Baseframe.FrameWidth;
		    	FrameHeight=fbi->Baseframe.FrameHeight;
				//ɥ⤵
				WindowWidth =fbi->Baseframe.DispWidth;
				WindowHeight=fbi->Baseframe.DispHeight;
				DPRINTK("Frm:%d,%d Win:%d,%d Cur:%d,%d\n",FrameWidth,FrameHeight,WindowWidth,WindowHeight,ScrollX,ScrollY);
				if(fbi->ScrollSpeedX==0x80000000 
					&& fbi->ScrollSpeedY==0x80000000){
					//®٤̤ξE
					DPRINTK("NoSpeed\n");
					NextX= fbi->ScrollTargetX;
					NextY= fbi->ScrollTargetY;
				}else{
					NextX= ScrollX+fbi->ScrollSpeedX;
					NextY= ScrollY+fbi->ScrollSpeedY;
				}
				//ϰϳե饰Eå
				flag=0;
		        if(fbi->ScrollTargetX==0x80000000 
					&& fbi->ScrollTargetY==0x80000000){
					//Ū֤ꤵEƤʤE
			
				}else{
					//Ū֤ꤵEE
					//XɸˤĤƥå
					if(ScrollX<fbi->ScrollTargetX){
						if(NextX>fbi->ScrollTargetX) NextX=fbi->ScrollTargetX; //Ūã
					}else{
						if(NextX<fbi->ScrollTargetX) NextX=fbi->ScrollTargetX; //Ūã
					}
					if(ScrollX==fbi->ScrollTargetX) NextX=fbi->ScrollTargetX;  //Ūã
					//YɸˤĤƥå
					if(ScrollY<fbi->ScrollTargetY){
						if(NextY>fbi->ScrollTargetY) NextY=fbi->ScrollTargetY; //Ūã			
					}else{
						if(NextY<fbi->ScrollTargetY) NextY=fbi->ScrollTargetY; //Ūã
					}
					if(ScrollY==fbi->ScrollTargetY) NextY=fbi->ScrollTargetY; //Ūã
					// XY ȤŪϤã
					if(NextY==fbi->ScrollTargetY && NextX==fbi->ScrollTargetX) flag |=2;
				}
				//ϰϳˤʤäå
		        if(NextX<0 || NextX>(FrameWidth-WindowWidth)) flag|=1;
		        if(NextY<0 || NextY>(FrameHeight-WindowHeight)) flag|=1;
				DPRINTK("next:%d,%d flag=%d\n",NextX,NextY,flag);
				if(flag & 0x1){
					//ϰϳˤʤäƤEʤưʤ
					NextX=ScrollX; NextY=ScrollY;
				}
				
				ScrollX=NextX; ScrollY=NextY;
				//E̤Ǽ
				fbi->Baseframe.FrameX		=NextX;
				fbi->Baseframe.FrameY		=NextY;
				fbi->Baseframe.Flip			=fbi->ScrollFlip;
				//ɽ֤򹹿
				SetDMASetting0(fbi,fbi->Baseframe.Flip);
				//
				if(flag!=0){
					//ŪãϰϳȤʤäE®٤򣰤ˤE
					fbi->ScrollSpeedX=0;
					fbi->ScrollSpeedY=0;
				}
				fbi->ScrollStatus &= 0xFFFFFFFC; //ϰϳե饰Eå
				fbi->ScrollStatus |= flag; //ϰϳե饰򥻥å
			}else{
				fbi->ScrollStatus &=0xF;
					
			}//Scroll򤹤

		}
#endif
#ifdef CONFIG_FB_SONYNAVU_START
                if(StartUp_Count>0){
                   StartUp_Count--;
                  if(StartUp_Count==0){
                      printk("StartUp_Count==0\n");
                      fbi->lccr5 |= (LCCR5_IUM1 | LCCR5_BSM1 | LCCR5_EOFM1 | LCCR5_SOFM1);
                      LCCR5 = fbi->lccr5 ;
	              fbi->ovl1c1 |= 0x80000000;
	              fbi->Overlay1.Status=1;
                      OVL1C1=fbi->ovl1c1;
                  }
                }
#endif
	}
	LCSR = lcsr;
        //LCSR1= lcsr1;
  if(handled==0){
    printk("Not handled  0x%08X 0x%08X\n",lcsr,lcsr1);

  }
        enable_irq(IRQ_LCD);
	return IRQ_HANDLED;
}

/*
 * This function must be called from task context only, since it will
 * sleep when disabling the LCD controller, or if we get two contending
 * processes trying to alter state.
 */
static void set_ctrlr_state(struct sonynavu_pxafb_info *fbi, u_int state)
{
	u_int old_state;

	down(&fbi->ctrlr_sem);

	old_state = fbi->state;

	/*
	 * Hack around fbcon initialisation.
	 */
	if (old_state == C_STARTUP && state == C_REENABLE)
		state = C_ENABLE;

	switch (state) {
	case C_DISABLE_CLKCHANGE:
		DPRINTK("C_DISABLE_CLKCHANGE\n");
		/*
		 * Disable controller for clock change.  If the
		 * controller is already disabled, then do nothing.
		 */
		if (old_state != C_DISABLE && old_state != C_DISABLE_PM) {
			fbi->state = state;
			//TODO __sonynavu_pxafb_lcd_power(fbi, 0);
			sonynavu_pxafb_disable_controller(fbi);
		}
		break;

	case C_DISABLE_PM:
		DPRINTK("C_DISABLE_PM\n");
        fbi->state = state;
#ifdef CONFIG_MACH_SONY_NAVU_XYZ3
        //DISP(GPIO14)....disable
        GPCR0=1<<14;
        mdelay(51);
        //LCD5V_ON(GPIO35)...disable
        GPCR1=1<<(35-32);
        mdelay(101);
        //LCD I/F Signal..disable
        sonynavu_pxafb_disable_controller(fbi);
        mdelay(100);
        //LCD2.5(GPIO11)..disable
        //GPCR0=1<<11; //061106 This operation move to another place....
#endif
#ifdef CONFIG_MACH_SONY_NAVU_XYZ4
      if(QVGA_flag){


      }else{
        //LCD_ON(GPIO14)....disable
        GPCR0=1<<14;
        mdelay(81);
        //LCD I/F Signal..disable
        sonynavu_pxafb_disable_controller(fbi);
        mdelay(11);
        //LCD3.2(GPIO11)..disable
        GPCR0=1<<11;

      }
#endif
        break;
	case C_DISABLE:
		/*
		 * Disable controller
		 */
		if (old_state != C_DISABLE) {
			fbi->state = state;
			if (old_state != C_DISABLE_CLKCHANGE){
				sonynavu_pxafb_disable_controller(fbi);
			}
		}else{
			DPRINTK("C_DISABLE\n");
		}

		break;

	case C_ENABLE_CLKCHANGE:
		/*
		 * Enable the controller after clock change.  Only
		 * do this if we were disabled for the clock change.
		 */
		DPRINTK("C_DISABLE\n");
		if (old_state == C_DISABLE_CLKCHANGE) {
			fbi->state = C_ENABLE;
			sonynavu_pxafb_enable_controller(fbi);
			//TODO __sonynavu_pxafb_lcd_power(fbi, 1);
		}
		break;

	case C_REENABLE:
		/*
		 * Re-enable the controller only if it was already
		 * enabled.  This is so we reprogram the control
		 * registers.
		 */
		DPRINTK("C_REENABLE\n");
		if (old_state == C_ENABLE) {
			sonynavu_pxafb_disable_controller(fbi);
			sonynavu_pxafb_setup_gpio(fbi);
			sonynavu_pxafb_enable_controller(fbi);
		}
		break;

	case C_ENABLE_PM:
		/*
		 * Re-enable the controller after PM.  This is not
		 * perfect - think about the case where we were doing
		 * a clock change, and we suspended half-way through.
		 */
		DPRINTK("C_ENABLE_PM\n");
		if (old_state != C_DISABLE_PM)
			break;
		/* fall through */
        fbi->state = C_ENABLE;
#ifdef CONFIG_MACH_SONY_NAVU_XYZ3
        //LCD2.5(GPIO11)..enable
        GPSR0=1<<11;
        mdelay(11);
        //LCD I/F Signal..enable
        sonynavu_pxafb_enable_controller(fbi);
        mdelay(5);
        //LCD5V_ON(GPIO35)...enable
        GPSR1=1<<(35-32);
        mdelay(11);
        //DISP(GPIO14)....enable
        GPSR0=1<<14;
#endif
#ifdef CONFIG_MACH_SONY_NAVU_XYZ4
      if(QVGA_flag){


      }else{
        //LCD3.2(GPIO11)..enable
        GPSR0=1<<11;
        mdelay(21);
        //LCD I/F Signal..enable
        sonynavu_pxafb_enable_controller(fbi);
        mdelay(6);
        //LCD_ON(GPIO14)....enable
        GPSR0=1<<14;

      }
#endif
        break;
	case C_ENABLE:
		/*
		 * Power up the LCD screen, enable controller, and
		 * turn on the backlight.
		 */
		DPRINTK("C_ENABLE\n");

		if (old_state != C_ENABLE) {
			fbi->state = C_ENABLE;
			sonynavu_pxafb_setup_gpio(fbi);
			sonynavu_pxafb_enable_controller(fbi);
		}
		break;
	}
	up(&fbi->ctrlr_sem);
}

/*
 * Our LCD controller task (which is called when we blank or unblank)
 * via keventd.
 */
static void sonynavu_pxafb_task(void *dummy)
{
	struct sonynavu_pxafb_info *fbi = dummy;
	u_int state = xchg(&fbi->task_state, -1);

	set_ctrlr_state(fbi, state);
}

#ifdef CONFIG_CPU_FREQ
/*
 * CPU clock speed change handler.  We need to adjust the LCD timing
 * parameters when the CPU clock is adjusted by the power management
 * subsystem.
 *
 * TODO: Determine why f->new != 10*get_lclk_frequency_10khz()
 */
static int
sonynavu_pxafb_clkchg_notifier(struct notifier_block *nb, unsigned long val, void *data)
{
	struct pxafb_info *fbi = TO_INF(nb, freq_transition);
	//TODO struct cpufreq_freqs *f = data;
	u_int pcd;

	switch (val) {
	case CPUFREQ_PRECHANGE:
		set_ctrlr_state(fbi, C_DISABLE_CLKCHANGE);
		break;

	case CPUFREQ_POSTCHANGE:
		pcd = get_pcd(fbi->fb.var.pixclock);
		set_hsync_time(fbi, pcd);
		fbi->reg_lccr3 = (fbi->reg_lccr3 & ~0xff) | LCCR3_PixClkDiv(pcd);
		set_ctrlr_state(fbi, C_ENABLE_CLKCHANGE);
		break;
	}
	return 0;
}

static int
sonynavu_pxafb_freq_policy(struct notifier_block *nb, unsigned long val, void *data)
{
	struct pxafb_info *fbi = TO_INF(nb, freq_policy);
	struct fb_var_screeninfo *var = &fbi->fb.var;
	struct cpufreq_policy *policy = data;

	switch (val) {
	case CPUFREQ_ADJUST:
	case CPUFREQ_INCOMPATIBLE:
		printk(KERN_DEBUG "min dma period: %d ps, "
			"new clock %d kHz\n", pxafb_display_dma_period(var),
			policy->max);
		// TODO: fill in min/max values
		break;
#if 0
	case CPUFREQ_NOTIFY:
		printk(KERN_ERR "%s: got CPUFREQ_NOTIFY\n", __FUNCTION__);
		do {} while(0);
		/* todo: panic if min/max values aren't fulfilled
		 * [can't really happen unless there's a bug in the
		 * CPU policy verification process *
		 */
		break;
#endif
	}
	return 0;
}
#endif

#ifdef CONFIG_PM
/*
 * Power management hooks.  Note that we won't be called from IRQ context,
 * unlike the blank functions above, so we may sleep.
 */
static int sonynavu_pxafb_suspend(struct platform_device *dev, pm_message_t state)
{
	struct sonynavu_pxafb_info *fbi = platform_get_drvdata(dev);

	set_ctrlr_state(fbi, C_DISABLE_PM);
	return 0;
}

static int sonynavu_pxafb_resume(struct platform_device *dev)
{
	struct sonynavu_pxafb_info *fbi = platform_get_drvdata(dev);

	
		set_ctrlr_state(fbi, C_ENABLE_PM);
	return 0;
}
#else
#define sonynavu_pxafb_suspend	NULL
#define sonynavu_pxafb_resume	NULL
#endif

/*
 * sonynavu_pxafb_map_video_memory():
 *      Allocates the DRAM memory for the frame buffer.  This buffer is
 *	remapped into a non-cached, non-buffered, memory region to
 *      allow palette and pixel writes to occur without flushing the
 *      cache.  Once this area is remapped, all virtual memory
 *      access to the video memory should occur at the new region.
 */
static int __init sonynavu_pxafb_map_video_memory(struct sonynavu_pxafb_info *fbi)
{
	u_long frame_mem_size;
	int i;
	/*
	 * We reserve one page for the palette, plus the size
	 * of the framebuffer.
	 *
	 * layout of stuff in memory
	 *
	 *                fblow descriptor
	 *                fbhigh descriptor
	 *                palette descriptor
	 *                palette
	 *   page boundary->
	 *                frame buffer
	 */
	//DMA Descriptor
	fbi->map_size2 = PAGE_ALIGN(0x4000);

	fbi->map_cpu2 
		= dma_alloc_writecombine(NULL, fbi->map_size2,&fbi->map_dma2,GFP_KERNEL | GFP_DMA );

	DPRINTK("map_size2 = 0x%08lx\n", (u_long) fbi->map_size2 );
	DPRINTK("fbi->map_cpu2 = 0x%x\n", (unsigned int)fbi->map_cpu2);
	DPRINTK("fbi->map_dma2 = 0x%x\n", fbi->map_dma2);

	//Frame Buffer
	fbi->map_size = PAGE_ALIGN(fbi->fb.fix.smem_len );

	fbi->map_cpu 
		= dma_alloc_writecombine(NULL, fbi->map_size,&fbi->map_dma,GFP_KERNEL | GFP_DMA );	

	DPRINTK("map_size = 0x%08lx\n", (u_long) fbi->map_size );
	DPRINTK("fbi->map_cpu = 0x%x\n", (unsigned int)fbi->map_cpu);
	DPRINTK("fbi->map_dma = 0x%x\n", fbi->map_dma);

	if (fbi->map_cpu) {
		fbi->fb.fix.smem_start = fbi->map_dma;
		fbi->fb.screen_base = fbi->map_cpu;
		//Overlay2 Video
		frame_mem_size=(VIDEO_XRES * VIDEO_YRES);
		fbi->screenvideo_y_cpu = &fbi->map_cpu[OFFSET_VIDEO];
		fbi->screenvideo_y_dma = fbi->map_dma+OFFSET_VIDEO;
		fbi->screenvideo_b_cpu = &fbi->screenvideo_y_cpu[frame_mem_size];
		fbi->screenvideo_b_dma = fbi->screenvideo_y_dma+frame_mem_size;
		fbi->screenvideo_r_cpu = &fbi->screenvideo_b_cpu[frame_mem_size/2];
		fbi->screenvideo_r_dma = fbi->screenvideo_b_dma+frame_mem_size/2;

		//Overlay2 Linux
		frame_mem_size=(LCD_XRES * LCD_YRES * LCD_BPP / 8);
		fbi->screen2_cpu = &fbi->map_cpu[OFFSET_OVERLAY2_0];
		fbi->screen2_dma = fbi->map_dma+OFFSET_OVERLAY2_0 ;
		fbi->palette2_cpu 
			= (u32 *)&fbi->screen2_cpu[OFFSET_OVERLAY2];
		fbi->palette2_dma 
			= fbi->screen2_dma + OFFSET_OVERLAY2 ;
		
		//Cursor
		fbi->palette3_cpu 
			= (u32 *)&fbi->map_cpu[OFFSET_CURSOR_PAL];
		fbi->palette3_dma 
			= fbi->map_dma +OFFSET_CURSOR_PAL;
		fbi->screen3_cpu0 = (u_char *)&fbi->map_cpu[OFFSET_CURSOR_0];
		fbi->screen3_dma0 = fbi->map_dma +OFFSET_CURSOR_0 ;
		fbi->screen3_cpu1 = (u_char *)&fbi->map_cpu[OFFSET_CURSOR_1];
		fbi->screen3_dma1 = fbi->map_dma +OFFSET_CURSOR_1 ;
		fbi->screen3_cpu2 = (u_char *)&fbi->map_cpu[OFFSET_CURSOR_2];
		fbi->screen3_dma2 = fbi->map_dma +OFFSET_CURSOR_2 ;
		fbi->screen3_cpu3 = (u_char *)&fbi->map_cpu[OFFSET_CURSOR_3];
		fbi->screen3_dma3 = fbi->map_dma +OFFSET_CURSOR_3 ;

		//Base Layer Map
		fbi->palette0_cpu 
			= (u32 *)&fbi->map_cpu[OFFSET_BASELAYER];
		fbi->palette0_dma 
			= fbi->map_dma+OFFSET_BASELAYER;
		fbi->screen0_cpu0 = (u_char *)&fbi->map_cpu[OFFSET_BASELAYER_0];
		fbi->screen0_dma0 = fbi->map_dma+ OFFSET_BASELAYER_0  ;
		fbi->screen0_cpu1 = (u_char *)&fbi->map_cpu[OFFSET_BASELAYER_1];
		fbi->screen0_dma1 = fbi->map_dma+ OFFSET_BASELAYER_1  ;
		for(i=0;i<256;i++) fbi->palette0_cpu[i]=0x00000000;
		fbi->palette0_work 
			= (u32 *)&fbi->map_cpu[OFFSET_BASELAYER_PALWORK];
		for(i=0;i<256;i++) fbi->palette0_work[i]=0x00000000;
		//Overlay1 GUI
		fbi->palette1_cpu 
			= (u32 *)(fbi->map_cpu+OFFSET_OVERLAY1);
		fbi->palette1_dma 
			= fbi->map_dma + OFFSET_OVERLAY1 ;
		fbi->screen1_cpu0 = (u_char *)&fbi->map_cpu[OFFSET_OVERLAY1_0];
		fbi->screen1_dma0 = fbi->map_dma + OFFSET_OVERLAY1_0;
		fbi->screen1_cpu1 = (u_char *)&fbi->map_cpu[OFFSET_OVERLAY1_1];
		fbi->screen1_dma1 = fbi->map_dma+  OFFSET_OVERLAY1_1 ;

		//
		DPRINTK("fbi->screen0_cpu0 = 0x%x\n", (unsigned int)fbi->screen0_cpu0);
		DPRINTK("fbi->screen0_dma0 = 0x%x\n", fbi->screen0_dma0);
		DPRINTK("fbi->screen0_cpu1 = 0x%x\n", (unsigned int)fbi->screen0_cpu1);
		DPRINTK("fbi->screen0_dma1 = 0x%x\n", fbi->screen0_dma1);
                DPRINTK("fbi->palette0_cpu = 0x%x\n", (unsigned int)fbi->palette0_cpu);
		DPRINTK("fbi->palette0_dma = 0x%x\n", fbi->palette0_dma);

		DPRINTK("fbi->screen1_cpu0 = 0x%x\n", (unsigned int)fbi->screen1_cpu0);
		DPRINTK("fbi->screen1_dma0 = 0x%x\n", fbi->screen1_dma0);
		DPRINTK("fbi->screen1_cpu1 = 0x%x\n", (unsigned int)fbi->screen1_cpu1);
		DPRINTK("fbi->screen1_dma1 = 0x%x\n", fbi->screen1_dma1);
	        DPRINTK("fbi->palette1_cpu = 0x%x\n", (unsigned int)fbi->palette1_cpu);
		DPRINTK("fbi->palette1_dma = 0x%x\n", fbi->palette1_dma);

		DPRINTK("fbi->screen2_cpu = 0x%x\n", (unsigned int)fbi->screen2_cpu);
		DPRINTK("fbi->screen2_dma = 0x%x\n", fbi->screen2_dma);
	        DPRINTK("fbi->palette2_cpu = 0x%x\n", (unsigned int)fbi->palette2_cpu);
		DPRINTK("fbi->palette2_dma = 0x%x\n", fbi->palette2_dma);

	}

	return fbi->map_cpu ? 0 : -ENOMEM;
}

#ifdef CONFIG_FB_PXA_PARAMETERS
static int __init sonynavu_pxafb_parse_options(struct device *dev, char *options)
{
	char *this_opt;

        if (!options || !*options)
                return 0;

	printk("PXAFB:options are \"%s\"\n", options ? options : "null");

	/* could be made table driven or similar?... */
        while ((this_opt = strsep(&options, ",")) != NULL) {
                if (!strncmp(this_opt, "QVGA", 5)) {
                  LCD_XRES=QVGA_LCD_XRES;
                  LCD_YRES=QVGA_LCD_YRES;
                  LCD_LCCR1=QVGA_LCD_LCCR1;
                  LCD_LCCR2=QVGA_LCD_LCCR2;
                  LCD_LCCR3=QVGA_LCD_LCCR3;
                  LCD_LCCR4=QVGA_LCD_LCCR4;
                  QVGA_flag=1;
                } else {
                  dev_err(dev, "unknown option: %s\n", this_opt);
                  //ignore unknown option ...t-sakai@me.sony.co.jp
                  //return -EINVAL;
                }
        }
        return 0;

}
#endif

static struct sonynavu_pxafb_info * __init sonynavu_pxafb_init_fbinfo(struct device *dev)
{
	struct sonynavu_pxafb_info *fbi;
	void *addr;
	//struct sonynavu_pxafb_mach_info *inf = dev->platform_data;

	/* Alloc the sonynavu_pxafb_info and pseudo_palette in one step */
	fbi = kmalloc(sizeof(struct sonynavu_pxafb_info) + sizeof(u32) * 16, GFP_KERNEL);
	if (!fbi)
		return NULL;

	memset(fbi, 0, sizeof(struct sonynavu_pxafb_info));
	fbi->dev = dev;

	/**Eɽ֤򹹿**/
	fbi->LCDStatus=1;
	//BaseFrame
	fbi->Baseframe.Status		=1; //鴁E֤ɽ
	fbi->Baseframe.Flip			=0; //եEåף
	fbi->Baseframe.DispX		=0;
	fbi->Baseframe.DispY		=0;
	fbi->Baseframe.DispWidth	=LCD_XRES;
	fbi->Baseframe.DispHeight	=LCD_YRES;	
	fbi->Baseframe.FrameX		=0;
	fbi->Baseframe.FrameY		=0;	
	fbi->Baseframe.FrameWidth	=LCD_BASE_MAX_XRES;
	fbi->Baseframe.FrameHeight	=LCD_BASE_MAX_YRES;
	fbi->ScrollStatus=0;
	fbi->ScrollFlip=0;
	fbi->ScrollTargetX=0;
	fbi->ScrollTargetY=0;
	fbi->ScrollSpeedX=0;
	fbi->ScrollSpeedY=0;
	//Overlay1 (GUI)
	fbi->Overlay1.Status		=0; //鴁E֤ɽ
	fbi->Overlay1.Flip			=0; //եEåף
	fbi->Overlay1.DispX			=0;
	fbi->Overlay1.DispY			=0;
	fbi->Overlay1.DispWidth		=LCD_XRES;
	fbi->Overlay1.DispHeight	=LCD_YRES;
#if 0
	fbi->Overlay1.FrameX		=160;
	fbi->Overlay1.FrameY		=32;
#else
	fbi->Overlay1.FrameX		=0;
	fbi->Overlay1.FrameY		=0;
#endif
	fbi->Overlay1.FrameWidth	=LCD_OVERLAY1_MAX_XRES;
	fbi->Overlay1.FrameHeight	=LCD_OVERLAY1_MAX_YRES;
	fbi->RequestFlip=0;
	//Overlay2_Console
	fbi->Overlay2_Console.Status		=0; //鴁E֤ɽ
	fbi->Overlay2_Console.Flip					=0; //եEåף
	fbi->Overlay2_Console.DispX			=0;
	fbi->Overlay2_Console.DispY			=0;
	fbi->Overlay2_Console.DispWidth		=LCD_XRES;
	fbi->Overlay2_Console.DispHeight	=LCD_YRES;	
	fbi->Overlay2_Console.FrameX		=0;
	fbi->Overlay2_Console.FrameY		=0;	
	fbi->Overlay2_Console.FrameWidth	=LCD_XRES;
	fbi->Overlay2_Console.FrameHeight	=LCD_YRES;
	//Overlay2_Video
	fbi->Overlay2_Video.Status		=0; //鴁E֤ɽ
	fbi->Overlay2_Video.Flip		=0; //եEåף
	fbi->Overlay2_Video.DispX		=0;
	fbi->Overlay2_Video.DispY		=0;
	fbi->Overlay2_Video.DispWidth	=VIDEO_XRES;
	fbi->Overlay2_Video.DispHeight	=VIDEO_YRES;	
	fbi->Overlay2_Video.FrameX		=0;
	fbi->Overlay2_Video.FrameY		=0;	
	fbi->Overlay2_Video.FrameWidth	=VIDEO_XRES;
	fbi->Overlay2_Video.FrameHeight	=VIDEO_YRES;
	//Cursor
	fbi->Cursor.Status		=0; //鴁E֤ɽ
	fbi->Cursor.Flip		=0; //եEåף
	fbi->Cursor.DispX		=0;
	fbi->Cursor.DispY		=0;
	fbi->Cursor.DispWidth	=LCD_CURSOR_XRES;
	fbi->Cursor.DispHeight	=LCD_CURSOR_YRES;	
	fbi->Cursor.FrameX		=0;
	fbi->Cursor.FrameY		=0;	
	fbi->Cursor.FrameWidth	=LCD_CURSOR_XRES;
	fbi->Cursor.FrameHeight	=LCD_CURSOR_YRES;

	strcpy(fbi->fb.fix.id, NAVU_NAME);

	fbi->fb.fix.type	= FB_TYPE_PACKED_PIXELS;
	fbi->fb.fix.type_aux	= 0;
	fbi->fb.fix.xpanstep	= 0;
	fbi->fb.fix.ypanstep	= 0;
	fbi->fb.fix.ywrapstep	= 0;
	fbi->fb.fix.accel	= FB_ACCEL_NONE;

	fbi->fb.var.nonstd	= 0;
	fbi->fb.var.activate	= FB_ACTIVATE_NOW;
	fbi->fb.var.height	= -1;
	fbi->fb.var.width	= -1;
	fbi->fb.var.accel_flags	= 0;
	fbi->fb.var.vmode	= FB_VMODE_NONINTERLACED;

	fbi->fb.fbops		= &sonynavu_pxafb_ops;
	fbi->fb.flags		= FBINFO_DEFAULT;
	fbi->fb.node		= -1;

	addr = fbi;
	addr = addr + sizeof(struct sonynavu_pxafb_info);
	fbi->fb.pseudo_palette	= addr;

	fbi->max_xres			= LCD_XRES;
	fbi->fb.var.xres		= LCD_XRES;
	fbi->fb.var.xres_virtual	= LCD_XRES;
	fbi->max_yres			= LCD_YRES;
	fbi->fb.var.yres		= LCD_YRES;
	fbi->fb.var.yres_virtual	= LCD_YRES;
	fbi->max_bpp			= LCD_BPP;
	fbi->fb.var.bits_per_pixel	= LCD_BPP;
	fbi->fb.var.pixclock		= LCD_PIXCLOCK;
	fbi->fb.var.hsync_len		= LCD_HORIZONTAL_SYNC_PULSE_WIDTH;
	fbi->fb.var.left_margin		= LCD_BEGIN_OF_LINE_WAIT_COUNT;
	fbi->fb.var.right_margin	= LCD_END_OF_LINE_WAIT_COUNT;
	fbi->fb.var.vsync_len		= LCD_VERTICAL_SYNC_PULSE_WIDTH;
	fbi->fb.var.upper_margin	= LCD_BEGIN_FRAME_WAIT_COUNT;
	fbi->fb.var.lower_margin	= LCD_END_OF_FRAME_WAIT_COUNT;
	fbi->fb.var.sync		= LCD_SYNC;
	fbi->fb.var.grayscale		= 0;
	fbi->cmap_inverse		= 0;
	fbi->cmap_static		= 0;

	//
	fbi->lccr0			= LCD_LCCR0;
	fbi->lccr1			= LCD_LCCR1;
	fbi->lccr2			= LCD_LCCR2;
	fbi->lccr3			= LCD_LCCR3;
	fbi->lccr4			= LCD_LCCR4;
	fbi->lccr5			= LCD_LCCR5;

	fbi->ovl1c1			= LCD_OVL1C1(LCD_XRES,LCD_YRES);
	fbi->ovl1c2			= LCD_OVL1C2(0,0);
	fbi->ovl2c1			= LCD_OVL2C1(LCD_XRES,LCD_YRES);
	fbi->ovl2c2			= LCD_OVL2C2(0,0);
	
	fbi->cursor			= LCD_CURSOR;

	//
	fbi->state			= C_STARTUP;
	fbi->task_state		= (u_char)-1;

	fbi->palette0_size = LCD_BASE_BPP == 8 ? 256 : 16;
	fbi->palette1_size = LCD_OVERLAY1_BPP == 8 ? 256 : 16;
	fbi->palette2_size = LCD_OVERLAY2_BPP == 8 ? 256 : 16;
	fbi->palette3_size = 4; // 4color
	fbi->fb.fix.smem_len = VRAMSIZE;

	//Init Vsync Wait queue
	atomic_set(&fbi->vsyncinfo.waitcnt, 0);
        init_waitqueue_head(&fbi->vsyncinfo.waitq);

        //Init StartFrame Wait queue
	atomic_set(&StartFrameInfo.waitcnt, 0);
        init_waitqueue_head(&StartFrameInfo.waitq);
	//
	atomic_set(&fbi->ctrlr_waitcnt, 0);
	init_waitqueue_head(&fbi->ctrlr_wait);
  
	INIT_WORK(&fbi->task, sonynavu_pxafb_task, fbi);
	init_MUTEX(&fbi->ctrlr_sem);

	return fbi;
}

//ǥХΥץ
//Update SAKAI 060628
int __init sonynavu_pxafb_probe(struct platform_device *dev)
{
	struct sonynavu_pxafb_info *fbi;
	//struct sonynavu_pxafb_mach_info *inf;
	int ret;

	DPRINTK("sonynavu_pxafb_probe\n");

	//inf = dev->platform_data;
	ret = -ENOMEM;
	fbi = NULL;
	//if (!inf)goto failed;

    QVGA_flag=0;

#ifdef CONFIG_FB_PXA_PARAMETERS
    //Set default parameter
    LCD_XRES=WQVGA_LCD_XRES;
    LCD_YRES=WQVGA_LCD_YRES;
    LCD_LCCR1=WQVGA_LCD_LCCR1;
    LCD_LCCR2=WQVGA_LCD_LCCR2;
    LCD_LCCR3=WQVGA_LCD_LCCR3;
    LCD_LCCR4=WQVGA_LCD_LCCR4;

	ret = sonynavu_pxafb_parse_options(&dev->dev, g_options);
	if (ret < 0) goto failed;
#endif
	fbi = sonynavu_pxafb_init_fbinfo(&dev->dev);
	if (!fbi) {
		DPRINTK("Failed to initialize framebuffer device\n");
		ret = -ENOMEM; // only reason for sonynavu_pxafb_init_fbinfo to fail is kmalloc
		goto failed;
	}

	/* Initialize video memory */
	ret = sonynavu_pxafb_map_video_memory(fbi);
	if (ret) {
		DPRINTK("Failed to allocate video RAM: %d\n", ret);
		ret = -ENOMEM;
		goto failed;
	}

	ret = request_irq(IRQ_LCD, sonynavu_pxafb_handle_irq, SA_INTERRUPT, "LCD", fbi);
	if (ret) {
		DPRINTK("request_irq failed: %d\n", ret);
		ret = -EBUSY;
		goto failed;
	}

	/*
	 * This makes sure that our colour bitfield
	 * descriptors are correctly initialised.
	 */
	sonynavu_pxafb_check_var(&fbi->fb.var, &fbi->fb);
	sonynavu_pxafb_set_par(&fbi->fb);

	platform_set_drvdata(dev, fbi);

	ret = register_framebuffer(&fbi->fb);
	if (ret < 0) {
		DPRINTK("Failed to register framebuffer device: %d\n", ret);
		goto failed;
	}

#ifdef CONFIG_PM
	// TODO
#endif

#ifdef CONFIG_CPU_FREQ
	fbi->clockchg.notifier_call = sonynavu_pxafb_clkchg_notifier;
    fbi->freq_policy.notifier_call = sonynavu_pxafb_freq_policy;
	cpufreq_register_notifier(&fbi->clockchg);
    cpufreq_register_notifier(&fbi->freq_policy, CPUFREQ_POLICY_NOTIFIER);
#endif

	/*
	 * Ok, now enable the LCD controller
	 */
	set_ctrlr_state(fbi, C_ENABLE);
        LCCR4 |= (1 << 31);
	LCCR4 |= (5 << 17);
        fbi->lccr4 |= ((1 << 31) | (5 << 17)); 
	return 0;

failed:
	platform_set_drvdata(dev, NULL);
	kfree(fbi);
	return ret;
}

static struct platform_driver sonynavu_pxafb_driver = {
	.probe		= sonynavu_pxafb_probe,
#ifdef CONFIG_PM
	.suspend	= sonynavu_pxafb_suspend,
	.resume		= sonynavu_pxafb_resume,
#endif
	.driver		= {
        .name		= "pxa2xx-fb",  //This parameter must be "pxa2xx-fb"
    },
};

#ifndef MODULE
int __devinit sonynavu_pxafb_setup(char *options)
{
# ifdef CONFIG_FB_PXA_PARAMETERS
	if (options)
		strlcpy(g_options, options, sizeof(g_options));
# endif
	return 0;
}
#else
# ifdef CONFIG_FB_PXA_PARAMETERS
module_param_string(options, g_options, sizeof(g_options), 0);
MODULE_PARM_DESC(options, "LCD parameters (see Documentation/fb/pxafb.txt)");
# endif
#endif

int __devinit sonynavu_pxafb_init(void)
{
#ifndef MODULE
	char *option = NULL;
  
        if (fb_get_options("pxafb", &option)) return -ENODEV;
	sonynavu_pxafb_setup(option);
#endif

	printk("Sony Nav-U Frame Buffer Initialized......\n");
	return platform_driver_register(&sonynavu_pxafb_driver);
}

module_init(sonynavu_pxafb_init);

MODULE_DESCRIPTION("loadable framebuffer driver for SONY NAVU");
MODULE_LICENSE("GPL");


EXPORT_SYMBOL(sonynavu_pxafb_wait_startframe);
EXPORT_SYMBOL(sonynavu_pxafb_wait_cancel_startframe);
