#ifndef LEDS_XYZ3_H_INCLUDED
#define LEDS_XYZ3_H_INCLUDED

#include <linux/compiler.h>
#include <linux/init.h>

#include <asm/hardware.h>
#include <asm/leds.h>
#include <asm/system.h>

#include <asm/arch/pxa-regs.h>
#include <asm/arch/xyz3.h>
#include <asm/arch/leds-xyz3.h>

extern unsigned int __hw_led_state;

#define   XYZ3_DEBUG_LED_1  (1 << 0)
#define   XYZ3_DEBUG_LED_2  (1 << 1)
#define   XYZ3_DEBUG_LED_3  (1 << 2)
#define   XYZ3_DEBUG_LED_4  (1 << 3)

#define XYZ3_GET_LED_VALUE()  (__hw_led_state)

#define XYZ3_SET_LED_VALUE(val)                         \
{                                                       \
    __hw_led_state = (val);                             \
    XYZ3_LEDCTRL = (XYZ3_LEDCTRL|0xF) & ~(0xF & val);   \
}


typedef enum {
    xyz3_led_on,
    xyz3_led_off,
    xyz3_led_toggle,
    xyz3_led_set,
} xyz3_led_ctrl_t;


static inline void xyz3_led_ctrl(xyz3_led_ctrl_t ctrl, unsigned int led)
{
    unsigned long flags;
    unsigned int leds_value;

    local_irq_save(flags);
    leds_value = XYZ3_GET_LED_VALUE();

    switch (ctrl) {
        case xyz3_led_on:
            leds_value |=  led;
            break;

        case xyz3_led_off:
            leds_value &= ~led;
            break;

        case xyz3_led_toggle:
            leds_value ^=  led;
            break;

        case xyz3_led_set:
            leds_value  =  led;
            break;

        default:
            /*
            printk(KERN_ERR "%s[%d] Unhandled led control type: %d\n",
                            __FILE__, __LINE__, ctrl);
            */
            ;
    }

    XYZ3_SET_LED_VALUE( leds_value );
    local_irq_restore(flags);
    return;
}

#endif/*LEDS_XYZ3_H_INCLUDED*/
