/*
 * linux/include/asm-arm/arch-pxa/system.h
 *
 * Author:  Nicolas Pitre
 * Created: Jun 15, 2001
 * Copyright:   MontaVista Software Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include "hardware.h"
#include "pxa-regs.h"

static inline void arch_idle(void)
{
    cpu_do_idle();
}


/*
 *  Realize reset from software without using watchdog.
 */
static inline void arch_reset(char mode)
{
    unsigned long pmnc;

    /*
     * set scratchpad to zero, just in case it is used as a
     * restart address by the bootloader.
     */
    PSPR = 0;

    /*
     * We have to clear all wake-up detect status register before
     * reboot, because DeepSleep mode does not clear PEDR bits.
     * 0x80000000: Wake-up from RTC
     * 0x00000002: Wake-up from GPIO-1 (ACC-ON)
     * 0x00000001: Wake-up from GPIO-0 (POWER-button)
     */
    PEDR |= 0x80000003;

    /*
     * Set RTC Alarm
     */
    PWER |= 0x80000000; // Enable wake-up due to RTC Alarm.
    PSLR |= 0x00400000; // Ignore VDD Fault in deep sleep mode.
    RTAR  = RCNR + 2;   // Reboot after 1~2 sec
    mdelay(10); // wait for above settings to take effect.

    /*
     * going to deep sleep mode.
     */
    pmnc = 0x0f;
    asm ("mcr p14, 0, %0, c7, c0, 0" : : "r" (pmnc));
}

