/*
 * asm-arm/arch-pxa/xyz3-ostimer.h
 *     -- Operating System Timers driver on PXA270, header.
 *
 * Author:     Hirokazu ABE and Ryoichi KATO
 * Created:    Feb 16, 2006
 * Copyright:  Sony Corp.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#ifndef XYZ3_OSTIMER_H
#define XYZ3_OSTIMER_H

#define OSTIMER_RESOLUTION_STOP 0
#define OSTIMER_RESOLUTION_1us 4
#define OSTIMER_RESOLUTION_1ms 2
#define OSTIMER_RESOLUTION_1s 3

/* ioctl */
#define OSTIMER_IOC_MAGIC 'g'
#define OSTIMER_GET_COUNT _IOR(OSTIMER_IOC_MAGIC, 1, unsigned long*)
#define OSTIMER_START_FREERUN _IOW(OSTIMER_IOC_MAGIC, 2, int)
#define OSTIMER_BLOCK_BY_COUNT _IOW(OSTIMER_IOC_MAGIC, 3, int)

typedef struct {
    int resolution;
    int count;
} xyz3_ostimer_interval_setting;

#define OSTIMER_SET_INTERVAL _IOW(OSTIMER_IOC_MAGIC, 4, xyz3_ostimer_interval_setting*)
#define OSTIMER_LATCH_INTERVAL _IO(OSTIMER_IOC_MAGIC, 5)

#ifdef __KERNEL__

extern void xyz3_ostimer_setting(int ch, unsigned long count, int oneshot, int reset_on_match, int resolution);
extern unsigned long xyz3_ostimer_get_count(int ch);
extern int xyz3_ostimer_get_resolution(int ch);
extern int xyz3_ostimer_blockcall(int ch, unsigned long count, int resolution);
extern int xyz3_ostimer_register_callback(int ch, unsigned long count, int resolution, void (*callback)(void));
extern int xyz3_ostimer_remove_callback(int ch, void (*callback)(void));

#endif /* __KERNEL__ */

#define XYZ3_OSTIMER_CH_BOOT 4
#define XYZ3_OSTIMER_CH_AD_CYCTIMER 5
#define XYZ3_OSTIMER_CH_AD_TICK 6
#define XYZ3_OSTIMER_CH_BTUART 7

#endif /* XYZ3_OSTIMER_H */
