/*
 *  linux/include/asm-arm/arch-pxa/xyz3.h
 *
 *  Author:	Nicolas Pitre
 *  Created:	Nov 14, 2002
 *  Copyright:	MontaVista Software Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#ifndef ASM_ARCH_XYZ3_H
#define ASM_ARCH_XYZ3_H

#include <asm/irq.h>
#include <asm/arch/hardware.h>

#define XYZ3_ETH_PHYS		PXA_CS3_PHYS
#define XYZ3_FPGA_PHYS		PXA_CS2_PHYS
#define XYZ3_FPGA_VIRT		(0xf0000000)
#define XYZ3_P2V(x)		((x) - XYZ3_FPGA_PHYS + XYZ3_FPGA_VIRT)
#define XYZ3_V2P(x)		((x) - XYZ3_FPGA_VIRT + XYZ3_FPGA_PHYS)

#ifndef __ASSEMBLY__
# define __XYZ3_REG(x)           (*((volatile unsigned long *)XYZ3_P2V(x)))
#else
# define __XYZ3_REG(x)           XYZ3_P2V(x)
#endif

#define XYZ3_LEDCTRL             __XYZ3_REG(0x08000040)


/* A listing of interrupts used by external hardware devices */
#define	XYZ3_POWER			IRQ_GPIO(0)
#define	XYZ3_AC97_IRQ		IRQ_GPIO(13)
#define	XYZ3_CF_IRQ			IRQ_GPIO(22)
#define	XYZ3_GPS_PPS_IRQ	IRQ_GPIO(17)
#define	XYZ3_BATT_CHG_IRQ	IRQ_GPIO(20)
#define	XYZ3_USB_DET_IRQ	IRQ_GPIO(27)
#define	XYZ3_ON_CRADLE_IRQ	IRQ_GPIO(41)
#define	XYZ3_PARKING_IRQ	IRQ_GPIO(53)

#define OSCR4_ADDR	0x40A00040
#define OSCR5_ADDR	0x40A00044
#define OSCR6_ADDR	0x40A00048
#define OSCR7_ADDR	0x40A0004C
#define OSCR8_ADDR	0x40A00050
#define OSCR9_ADDR	0x40A00054
#define OSCR10_ADDR	0x40A00058
#define OSCR11_ADDR	0x40A0005C
#define OSCR5		__REG(OSCR5_ADDR)  /* OS Timer Counter Register ch 5 */
#define OSCR6		__REG(OSCR6_ADDR)  /* OS Timer Counter Register ch 6 */
#define OSCR7		__REG(OSCR7_ADDR)  /* OS Timer Counter Register ch 7 */
#define OSCR8		__REG(OSCR8_ADDR)  /* OS Timer Counter Register ch 8 */
#define OSCR9		__REG(OSCR9_ADDR)  /* OS Timer Counter Register ch 9 */
#define OSCR10		__REG(OSCR10_ADDR)  /* OS Timer Counter Register ch 10 */
#define OSCR11		__REG(OSCR11_ADDR)  /* OS Timer Counter Register ch 11 */

#define OSMR4_ADDR	0x40A00080
#define OSMR5_ADDR	0x40A00084
#define OSMR6_ADDR	0x40A00088
#define OSMR7_ADDR	0x40A0008C
#define OSMR8_ADDR	0x40A00090
#define OSMR9_ADDR	0x40A00094
#define OSMR10_ADDR	0x40A00098
#define OSMR11_ADDR	0x40A0009C
#define OSMR5		__REG(OSMR5_ADDR) /* OS Timer Match Register ch 5 */
#define OSMR6		__REG(OSMR6_ADDR) /* OS Timer Match Register ch 6 */
#define OSMR7		__REG(OSMR7_ADDR) /* OS Timer Match Register ch 7 */
#define OSMR8		__REG(OSMR8_ADDR) /* OS Timer Match Register ch 8 */
#define OSMR9		__REG(OSMR9_ADDR) /* OS Timer Match Register ch 9 */
#define OSMR10		__REG(OSMR10_ADDR) /* OS Timer Match Register ch 10 */
#define OSMR11		__REG(OSMR11_ADDR) /* OS Timer Match Register ch 11 */

#define OMCR4_ADDR	0x40A000C0
#define OMCR5_ADDR	0x40A000C4
#define OMCR6_ADDR	0x40A000C8
#define OMCR7_ADDR	0x40A000CC
#define OMCR8_ADDR	0x40A000D0
#define OMCR9_ADDR	0x40A000D4
#define OMCR10_ADDR	0x40A000D8
#define OMCR11_ADDR	0x40A000DC
#define OMCR5		__REG(OMCR5_ADDR) /* OS Match Control Register ch 5 */
#define OMCR6		__REG(OMCR6_ADDR) /* OS Match Control Register ch 6 */
#define OMCR7		__REG(OMCR7_ADDR) /* OS Match Control Register ch 7 */
#define OMCR8		__REG(OMCR8_ADDR) /* OS Match Control Register ch 8 */
#define OMCR9		__REG(OMCR9_ADDR) /* OS Match Control Register ch 9 */
#define OMCR10		__REG(OMCR10_ADDR) /* OS Match Control Register ch 10 */
#define OMCR11		__REG(OMCR11_ADDR) /* OS Match Control Register ch 11 */


#endif
