
#ifndef __XYZ3_AD7795_H
#define __XYZ3_AD7795_H

#ifdef __cplusplus
#define _EXTERN_
extern "C" {
#else
#define _EXTERN_	extern
#endif /* __cplusplus */

#include <linux/ioctl.h>

/*--- Private ---*/
/* Register difinition */
#define COMREG_WENX			(0x01 << 7)
#define COMREG_RD			(0x01 << 6)
#define COMREG_WR			(0x00 << 6)
#define COMREG_RS(r)		(((r) & 0x07) << 3)
#define COMRS_COMM			COMREG_RS(0)
#define COMRS_STATUS		COMREG_RS(0)
#define COMRS_MODE			COMREG_RS(1)
#define COMRS_CONFIG		COMREG_RS(2)
#define COMRS_DATA			COMREG_RS(3)
#define COMRS_ID			COMREG_RS(4)
#define COMRS_IO			COMREG_RS(5)
#define COMRS_OFFSET		COMREG_RS(6)
#define COMRS_F_SCALE		COMREG_RS(7)
#define COMREG_CREAD		(0x01<<2)

#define COM_REG(reg)		((reg) & 0xff)

#define MODEREG_MD(md)		(((md) & 0x07) << 13)
#define MODEMD_CONTINUE		MODEREG_MD(0)
#define MODEMD_SINGLE		MODEREG_MD(1)
#define MODEMD_IDLE			MODEREG_MD(2)
#define MODEMD_POWERDOWN	MODEREG_MD(3)
#define MODEMD_IN_ZERO_CAL	MODEREG_MD(4)
#define MODEMD_IN_FULL_CAL	MODEREG_MD(5)
#define MODEMD_SYS_ZERO_CAL	MODEREG_MD(6)
#define MODEMD_SYS_FULL_CAL	MODEREG_MD(7)
#define MODEREG_PSW			(0x01 << 12)
#define MODEREG_AMP_CM		(0x01 << 9)
#define MODEREG_CLK(x)		(((x) & 0x03) << 6)
#define MODEREG_CHOP_DIS	(0x01 << 4)
#define MODEREG_FS(x)		(((x) & 0x0f) << 0)

#define MODE_REG(reg)		((reg) & 0xffff)


#define CONFREG_BIAS(v)		(((v) & 0x03) << 14)
#define CONFREG_BO			(0x01 << 13)
#define CONFREG_UNI			(0x01 << 12)
#define CONFREG_BOOST		(0x01 << 11)
#define CONFREG_GAIN(g)		(((g) & 0x07) << 8)
#define CONFREG_REF(sel)	(((sel) & 0x03) << 6)
#define CONG_REF_REF1		CONFREG_REF(0)
#define CONG_REF_REF2		CONFREG_REF(1)
#define CONG_REF_INTER		CONFREG_REF(2)
#define CONFREG_REF_DET		(0x01 << 5)
#define CONFREG_BUF			(0x01 << 4)
#define CONFREG_CH(ch)		(((ch) & 0x0f) << 0)

#define CONF_REG(reg)		((reg) & 0xffff)



/*--- Public ---*/
#define AD_SAMPL_TIME	4		// Sensor sampling time
#define POS_BUF_NUM		(10000/(AD_SAMPL_TIME*5))		// size for 10sec


#define ADC_CH_G_X	0x0
#define ADC_CH_G_Y	0x1
#define ADC_CH_G_Z	0x2
#define ADC_CH_TEMP	0x5
#define ADC_CH_ATMO	0x4


typedef struct {
	unsigned long	tick;		// 1ms
	unsigned short	g_x;		// X axis
	unsigned short	g_y;		// Y axis
	unsigned short	g_z;		// X axis
	unsigned short	temp;		// Ambiance temperature
	unsigned short	atmo;		// Atmospheric pressure
	unsigned short	dummy;		// for 32bit alignment
} PosSensor;

typedef struct {
	int			maxReadCount;
	PosSensor	*buf;
} AdReadArg;

typedef struct {
	unsigned short	status;
	unsigned short	mode;
	unsigned short	config;
	unsigned short	id;
	unsigned short	io;
	unsigned short	offset[4];
	unsigned short	fullscale[4];
} Ad7795RegArg;



/*---[ IoControl code ]-------------------*/
#define XYZ3ADC_MAGIC 'Z'
#define XYZ3ADC_GET_CONFIG	(_IOR(XYZ3ADC_MAGIC, 1, sizeof(Ad7795RegArg)))
#define XYZ3ADC_READ		(_IOR(XYZ3ADC_MAGIC, 2, sizeof(AdReadArg)))
#define XYZ3ADC_STOP		(_IOR(XYZ3ADC_MAGIC, 3, sizeof(unsigned long)))
#define XYZ3ADC_RESTART		(_IOR(XYZ3ADC_MAGIC, 4, sizeof(unsigned long)))


/*---[ Prototype ]-------------------*/
int AdCallbackRegistry( void (*func)(unsigned short, unsigned short, unsigned short) );
void AdCallbackUnregistry( void );


#ifdef __cplusplus
};
#endif /* __cplusplus */
#endif /* __XYZ3_AD7795_H */

